/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifndef STREAM_HPP
#define STREAM_HPP

#include <string.h>

#define STRM_ERR_OPEN	-1
#define STRM_ERR_READ	-2
#define STRM_ERR_WRITE	-3
#define STRM_ERR_SEEK	-4

struct CStream {
	CStream( void ) { Init(); }

	void Init ( void );

	virtual void	Seek	( slong pos )	{ mPos = pos; }
	virtual slong	GetSize ( void ) 		{ return 0; }
	virtual slong	GetPos  ( void ) const { return mPos; }
	
	virtual void	Read    ( void * dst, SInt32 size );
	virtual void	Write   ( const void * dst, SInt32 size );
	virtual bool	NextTag ( UInt32 sign );
	virtual void	ReadBlk ( void * dst, SInt32 size );
	virtual void	WriteBlk( const void * dst, SInt32 size );

	void ReadChar  ( char * C ) 	{ Read(C, 1); }
	void ReadUChar ( uchar * C ) 	{ Read(C, 1); }
	void ReadShort ( short * S ) 	{ Read(S, 2); }
	void ReadUShort( ushort * S ) 	{ Read(S, 2); }
	void ReadLong  ( SInt32 * L ) 	{ Read(L, 4); }
	void ReadULong ( UInt32 * L )	{ Read(L, 4); }
	void ReadFloat ( float * F )	{ Read(F, 4); }
	void ReadDouble( double * D )	{ Read(D, 8); }
	void ReadBool  ( bool * B )		{ *B = ReadLong() != 0; }
	void ReadPStr  ( uchar * s );

	char   ReadChar  ( void ) 		{ char C; ReadChar(&C); return C; }
	uchar  ReadUChar ( void ) 		{ uchar C; ReadUChar(&C); return C; }
	short  ReadShort ( void ) 		{ short S; ReadShort(&S); return S; }
	ushort ReadUShort( void ) 		{ ushort S; ReadUShort(&S); return S; }
	SInt32 ReadLong  ( void ) 		{ SInt32 L; ReadLong(&L); return L; }
	UInt32 ReadULong ( void )		{ UInt32 L; ReadULong(&L); return L; }
	float  ReadFloat ( void )		{ float F; ReadFloat(&F); return F; }
	double ReadDouble( void )		{ double D; ReadDouble(&D); return D; }
	bool   ReadBool  ( void )		{ return ReadLong() != 0; }
		
	void WriteChar  ( char C ) 		{ WriteBlk(&C, 1); }
	void WriteUChar ( uchar C ) 	{ WriteBlk(&C, 1); }
	void WriteShort ( short S ) 	{ Write(&S, 2); }
	void WriteUShort( ushort S ) 	{ Write(&S, 2); }
	void WriteLong  ( SInt32 L )	{ Write(&L, 4); }
	void WriteULong ( UInt32 L )	{ Write(&L, 4); }
	void WriteFloat ( float F )		{ Write(&F, 4); }
	void WriteDouble( double D )	{ Write(&D, 8); }
	void WriteBool  ( bool B )		{ WriteLong(B ? 1 : 0); }
	void WritePStr  ( const uchar * s );
		
// data members
	slong	mPos;
	SInt32	mError;
};

struct CMemStream : public CStream {
	CMemStream( void * base, SInt32 size ) { Init((char *) base, size); }
	
	void Init( char * base, SInt32 limit );

	virtual slong GetSize( void ) 	{ return mEnd - mBeg; }

	virtual void Read 	( void * dst, SInt32 size );
	virtual void Write	( const void * dst, SInt32 size );
	virtual bool NextTag( UInt32 sign );
	virtual void ReadBlk( void * dst, SInt32 size );
	virtual void WriteBlk( const void * dst, SInt32 size );
	
// data members
	char * mBeg, * mEnd;
};

//--------------- CStream -------------

inline void CStream::Init( void )
{
	mPos = 0;
	mError = 0;
}

inline void CStream::Read( void *, SInt32 size )
{
	if (!mError) mPos += size; 
}

inline void CStream::Write( const void *, SInt32 size )
{
	if (!mError) mPos += size; 
}

inline bool CStream::NextTag( UInt32 )
{
	return false; 
}

inline void CStream::ReadBlk( void *, SInt32 size )
{
	if (!mError) mPos += size; 
}

inline void CStream::WriteBlk( const void *, SInt32 size )
{
	if (!mError) mPos += size; 
}

inline void CStream::ReadPStr( uchar * s )
{
	if (mError) return;
	UInt32 L = ReadULong();
	if (mError) return;
	L = MIN(L, 254);
	s[0] = (unsigned char) L;
	if (L) ReadBlk(s + 1, L);
	s[L + 1] = 0;
}

inline void CStream::WritePStr( const uchar * s )
{
	if (mError) return;
	UInt32 L = MIN(s[0], 254);
	WriteULong(L);
	if (!mError && L) WriteBlk(s + 1, L);
}


//--------------- CMemStream -------------

inline void CMemStream::Init( char * base, SInt32 limit )
{
	CStream::Init();
	mBeg = base;
	mEnd = base + limit;
}

inline void CMemStream::Read( void * dst, SInt32 size )
{
	if (!mError && mBeg + size > mEnd) 
	 mError	= STRM_ERR_READ;
	 
	if (!mError) {
	 memmove(dst, mBeg, size);
#ifdef __LITTLE_ENDIAN__
	 ReverseField(dst, size);
#endif
	 mBeg += size;
	 mPos += size;
	}
}

inline void CMemStream::Write( const void * src, SInt32 size )
{
	if (!mError && mBeg + size > mEnd) 
	 mError	= STRM_ERR_WRITE;

	if (!mError) {
	 memmove(mBeg, src, size);
#ifdef __LITTLE_ENDIAN__
	 ReverseField(mBeg, size);
#endif
	 mBeg += size;
	 mPos += size;
	}
}

inline bool CMemStream::NextTag( UInt32 sign )
{
	if (mError || mBeg + sizeof(UInt32) > mEnd) return false;
	if (ReadULong() == sign) return true;
	mBeg -= sizeof(UInt32);
	return false;
}

inline void CMemStream::ReadBlk( void * dst, SInt32 size )
{
	if (!mError && mBeg + size > mEnd) 
	 mError	= STRM_ERR_READ;
	 
	if (!mError) {
	 memmove(dst, mBeg, size);
	 mBeg += size;
	 mPos += size;
	}
}

inline void CMemStream::WriteBlk( const void * src, SInt32 size )
{
	if (!mError && mBeg + size > mEnd) 
	 mError	= STRM_ERR_WRITE;

	if (!mError) {
	 memmove(mBeg, src, size);
	 mBeg += size;
	 mPos += size;
	}
}

#endif // STREAM_HPP

