/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifdef _WINDOWS
#pragma warning(disable : 4244)
#pragma warning(disable : 4305)
#endif

#include "ColorSpline.h"
#include "EIShader.h"
#include "FPU.h"
#include "Matrix.h"
#include "Noise.h"
#include "PString.h"
#include <stddef.h>
#include "System.h"

#define MIX(a, b, f) ((a) + (f) * ((b) - (a)))

#define shaderType				'EROD'
#define shaderVersion			0x00000001
#define shaderRevVersion		0x01000000

#define defShaderDensity		10.0
#define defShaderBumpStrength	2.0
#define defShaderErodeHoles		0xFFFFFFFF
#define defShaderHoleMin		0.3
#define defShaderHoleMax		0.4

typedef struct {
	UInt32 shaderOwnerType;
	UInt32 shaderVersionNum;
	float shaderDensity;
	float shaderBumpStrength;
	SInt32 shaderErodeHoles;
	float shaderHoleMin;
	float shaderHoleMax;
	ARGB shaderHoleColor;
	ARGB shaderSurfaceColor;
} ShaderInterfaceRec, *ShaderInterfacePtr;

typedef struct {
	float shaderDensity;
	float shaderBumpStrength;
	SInt32 shaderErodeHoles;
	float shaderHoleMin;
	float shaderHoleMax;
	ARGB_Real shaderHoleColor;
	ARGB_Real shaderSurfaceColor;
} ShaderDataRec, *ShaderDataPtr;

/* Static functions */

static float CalcEroded(EIShaderShadeRec *theShadePtr)
{
	ShaderDataPtr theShaderData = (ShaderDataPtr)theShadePtr->shaderData;
	EIShaderMapPtr theMap = theShadePtr->shaderMap;
	float freq = theShaderData->shaderDensity;
	float x = freq * theMap->map_X0;
	float y = freq * theMap->map_Y0;
	float z = freq * theMap->map_Z0;
	float ampl = 2.0;
	float bumpA = 0.0;
	float bumpB = 0.0;
	float bumpC = 0.0;
	float magnitude = 0.0;
	SInt32 i;
	
    /* Calculate a simple fractal 1/f noise function */

	for (i = 0; i < 6.0; i += 1.0) {
		float a, b, c;
		float n = DNoise2(x, y, z, &a, &b, &c) - 0.5;
		
		magnitude += ampl * FABS(n);
		if (n > 0.0) {
			bumpA -= ampl * a;
			bumpB -= ampl * b;
			bumpC -= ampl * c;
		} else {
			bumpA += ampl * a;
			bumpB += ampl * b;
			bumpC += ampl * c;
		}
		x += x;
		y += y;
		z += z;
		ampl *= 0.5;
	}

	/* sharpen peaks */
	
    magnitude = magnitude * magnitude * magnitude;
	
	/* perterb the normal */
	
	ampl = magnitude * theShaderData->shaderBumpStrength;
	if (BTST(theShadePtr->shaderFeatureFlags, shaderBumpFlag)) {
		theShadePtr->shaderBumpA = bumpA * ampl;
		theShadePtr->shaderBumpB = bumpB * ampl;
		theShadePtr->shaderBumpC = bumpC * ampl;
	}
	
	/* make the really deep divots go clear through the material */

	return 1.0 - SmoothStep(theShaderData->shaderHoleMin, theShaderData->shaderHoleMax, magnitude);	
}  /* CalcEroded */

static void DefaultShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface->shaderOwnerType = shaderType;
	theShaderInterface->shaderDensity = defShaderDensity;
	theShaderInterface->shaderBumpStrength = defShaderBumpStrength;
	theShaderInterface->shaderErodeHoles = defShaderErodeHoles;
	theShaderInterface->shaderHoleMin = defShaderHoleMin;
	theShaderInterface->shaderHoleMax = defShaderHoleMax;
	theShaderInterface->shaderHoleColor.argb.a = 255;
	theShaderInterface->shaderHoleColor.argb.r = 0;
	theShaderInterface->shaderHoleColor.argb.g = 0;
	theShaderInterface->shaderHoleColor.argb.b = 0;
	theShaderInterface->shaderSurfaceColor.argb.a = 255;
	theShaderInterface->shaderSurfaceColor.argb.r = 179;
	theShaderInterface->shaderSurfaceColor.argb.g = 140;
	theShaderInterface->shaderSurfaceColor.argb.b = 0;
}  /* DefaultShaderInterface */

static void FixShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface;
	
	FIXFIELD(theShaderInterface->shaderOwnerType);
	FIXFIELD(theShaderInterface->shaderVersionNum);
	FIXFIELD(theShaderInterface->shaderDensity);
	FIXFIELD(theShaderInterface->shaderBumpStrength);
	FIXFIELD(theShaderInterface->shaderErodeHoles);
	FIXFIELD(theShaderInterface->shaderHoleMin);
	FIXFIELD(theShaderInterface->shaderHoleMax);
	
}  /* FixShaderInterface */

static void GetShaderInterface(void* theInterface, SInt32 theSize, ShaderInterfacePtr theShaderInterface)
{
	/* Copy the shader interface data */
	
	if (theSize == sizeof(ShaderInterfaceRec))
		*theShaderInterface = *(ShaderInterfacePtr)theInterface;
	else {
		theShaderInterface->shaderOwnerType = 0;
		theShaderInterface->shaderVersionNum = 0;
	}
	
	/* Validate the shader interface data */
	
	FixShaderInterface(theShaderInterface);
	if (theShaderInterface->shaderOwnerType != shaderType ||
		theShaderInterface->shaderVersionNum != shaderVersion)
		DefaultShaderInterface(theShaderInterface);
}  /* GetShaderInterface */

/* Export functions */

SInt32 EIShaderInformation(EIShaderInformationRec *theInformationPtr)
{
	ShaderInterfaceRec theShaderInterface;
		
	/* Get the shader interface parameters */
		
	GetShaderInterface(theInformationPtr->shaderInterface, theInformationPtr->shaderInterfaceSize, &theShaderInterface);
	
	/* Setup the shader attribute access flags */
	
	if (BTST(theShaderInterface.shaderErodeHoles, 31))
		BSET(theInformationPtr->shaderFeatureFlags, shaderClipFlag);
	
	/* Setup the shader feature flags */
	
	BSET(theInformationPtr->shaderFeatureFlags, shaderAntialiasFlag);
	BSET(theInformationPtr->shaderFeatureFlags, shaderColorFlag);
	SETFLAG(theInformationPtr->shaderFeatureFlags, shaderBumpFlag, (theShaderInterface.shaderBumpStrength != 0.0));
	return shaderNoError;
}  /* EIShaderInformation */

SInt32 EIShaderInitialize(EIShaderInitializeRec *theInitializePtr)
{
	ShaderDataPtr theShaderData;
	
	/* Initialize the noise library */
	
	if (!InitNoise((NoiseMallocFunction)theInitializePtr->HostMalloc, NULL))
		return shaderMemoryError;
	
	/* Allocate the shader's private data record */
	
	theShaderData = (ShaderDataPtr)theInitializePtr->HostMalloc(sizeof(ShaderDataRec));
	if (theShaderData) {
		ShaderInterfaceRec theShaderInterface;
		
		/* Get the shader interface parameters */
		
		GetShaderInterface(theInitializePtr->shaderInterface, theInitializePtr->shaderInterfaceSize, &theShaderInterface);

		/* Setup the shader's private data */
		
		theShaderData->shaderDensity = theShaderInterface.shaderDensity;
		theShaderData->shaderBumpStrength = theShaderInterface.shaderBumpStrength;
		theShaderData->shaderErodeHoles = BTST(theShaderInterface.shaderErodeHoles, 31);
		theShaderData->shaderHoleMin = theShaderInterface.shaderHoleMin;
		theShaderData->shaderHoleMax = theShaderInterface.shaderHoleMax;
		theShaderData->shaderHoleColor.a = theShaderInterface.shaderHoleColor.argb.a / 255.0;
		theShaderData->shaderHoleColor.r = theShaderInterface.shaderHoleColor.argb.r / 255.0;
		theShaderData->shaderHoleColor.g = theShaderInterface.shaderHoleColor.argb.g / 255.0;
		theShaderData->shaderHoleColor.b = theShaderInterface.shaderHoleColor.argb.b / 255.0;
		theShaderData->shaderSurfaceColor.a = theShaderInterface.shaderSurfaceColor.argb.a / 255.0;
		theShaderData->shaderSurfaceColor.r = theShaderInterface.shaderSurfaceColor.argb.r / 255.0;
		theShaderData->shaderSurfaceColor.g = theShaderInterface.shaderSurfaceColor.argb.g / 255.0;
		theShaderData->shaderSurfaceColor.b = theShaderInterface.shaderSurfaceColor.argb.b / 255.0;
		
		/* Return the pointer to the shader's private data to the host */
		
		theInitializePtr->shaderData = theShaderData;
		return shaderNoError;
	} else
		return shaderMemoryError;
}  /* EIShaderInitialize */

SInt32 EIShaderFinish(EIShaderFinishRec *theFinishPtr)
{
	/* Get rid of the shader's private data */
	
	theFinishPtr->HostFree(theFinishPtr->shaderData);
	
	/* Finish the noise library */
	
	FinishNoise((NoiseFreeFunction)theFinishPtr->HostFree);
	return shaderNoError;
}  /* EIShaderFinish */

SInt32 EIShaderShade(EIShaderShadeRec *theShadePtr)
{
	ShaderDataPtr theShaderData = (ShaderDataPtr)theShadePtr->shaderData;
	float factor = CalcEroded(theShadePtr);
	
	/* Calculate the shader transparency */
	
	if (theShaderData->shaderErodeHoles && BTST(theShadePtr->shaderFeatureFlags, shaderClipFlag))
		theShadePtr->shaderClip = factor;
	
	/* Calculate the shader color */
	
	if (BTST(theShadePtr->shaderFeatureFlags, shaderColorFlag))
		MixARGBColors(&theShaderData->shaderHoleColor, &theShaderData->shaderSurfaceColor, &theShadePtr->shaderColor, factor);

	return shaderNoError;
}  /* EIShaderShade */
