/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifdef _WINDOWS
#pragma warning(disable : 4244)
#endif

#include "EIShader.h"
#include "FPU.h"
#include "Matrix.h"
#include "PString.h"
#include <stddef.h>
#include "System.h"

#define MIX(a, b, f) ((a) + (f) * ((b) - (a)))

#define shaderType				'LSHD'
#define shaderVersion			0x00000001
#define shaderRevVersion		0x01000000

typedef struct {
	UInt32 shaderOwnerType;
	UInt32 shaderVersionNum;
	SInt32 dummy;
} ShaderInterfaceRec, *ShaderInterfacePtr;

typedef struct {
	SInt32 dummy;
} ShaderDataRec, *ShaderDataPtr;

/* Static functions */

static void DefaultShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface->shaderOwnerType = shaderType;
	theShaderInterface->shaderVersionNum = shaderVersion;
}  /* DefaultShaderInterface */

static void FixShaderInterface(ShaderInterfacePtr theShaderInterface)
{
	theShaderInterface;
	
	FIXFIELD(theShaderInterface->shaderOwnerType);
	FIXFIELD(theShaderInterface->shaderVersionNum);
}  /* FixShaderInterface */

static void GetShaderInterface(void* theInterface, SInt32 theSize, ShaderInterfacePtr theShaderInterface)
{
	/* Copy the shader interface data */
	
	if (theSize == sizeof(ShaderInterfaceRec))
		*theShaderInterface = *(ShaderInterfacePtr)theInterface;
	else {
		theShaderInterface->shaderOwnerType = 0;
		theShaderInterface->shaderVersionNum = 0;
	}
	
	/* Validate the shader interface data */
	
	FixShaderInterface(theShaderInterface);
	if (theShaderInterface->shaderOwnerType != shaderType ||
		theShaderInterface->shaderVersionNum != shaderVersion)
		DefaultShaderInterface(theShaderInterface);
}  /* GetShaderInterface */

/* Export functions */

SInt32 EIShaderInformation(EIShaderInformationRec *theInformationPtr)
{
	/* Setup the shader feature flags */
	
	BSET(theInformationPtr->shaderFeatureFlags, shaderIlluminationFlag);
	return shaderNoError;
}  /* EIShaderInformation */

SInt32 EIShaderInitialize(EIShaderInitializeRec *theInitializePtr)
{
	ShaderDataPtr theShaderData;
	
	/* Allocate the shader's private data record */
	
	theShaderData = (ShaderDataPtr)theInitializePtr->HostMalloc(sizeof(ShaderDataRec));
	if (theShaderData) {
		ShaderInterfaceRec theShaderInterface;
		
		/* Get the shader interface parameters */
		
		GetShaderInterface(theInitializePtr->shaderInterface, theInitializePtr->shaderInterfaceSize, &theShaderInterface);
		
		/* Return the pointer to the shader's private data to the host */
		
		theInitializePtr->shaderData = theShaderData;
		return shaderNoError;
	} else
		return shaderMemoryError;
}  /* EIShaderInitialize */

SInt32 EIShaderFinish(EIShaderFinishRec *theFinishPtr)
{
	/* Get rid of the shader's private data */
	
	theFinishPtr->HostFree(theFinishPtr->shaderData);
	return shaderNoError;
}  /* EIShaderFinish */

SInt32 EIShaderShade(EIShaderShadeRec *theShadePtr)
{
	EIShaderAttributesPtr theAttributes = theShadePtr->shaderAttributes;
	EIShaderPixelPtr thePixel = theShadePtr->shaderPixel;
	float temp;
	float rNum, gNum, bNum;	/* The calculated color values */
	float sRed = 0.0;		/* The total light falling upon the surface */
	float sGreen = 0.0;
	float sBlue = 0.0;
	float hRed = 0.0;		/* The total highlight falling upon the surface */
	float hGreen = 0.0;
	float hBlue = 0.0;
	boolean transmissionFlag = (theAttributes->transmission.r > 0.0) || (theAttributes->transmission.g > 0.0) || (theAttributes->transmission.b > 0.0);
	char *theLight = NULL;
		
	/* Calculate the total illumnination color values */
	
	do {
		float dot;
		float normA, normB, normC;			/* The current lightsource's normal */
		float curRed, curGreen, curBlue;	/* The current lightsource's color */
		UInt32 calcSpecular;
		UInt32 calcDiffuse;
		
		/* Calculate the lightsource illumination */
		
		theShadePtr->HostLight(&theLight, &calcDiffuse, &calcSpecular, &dot, &normA, &normB, &normC, &curRed, &curGreen, &curBlue);
		
		/* Calculate the diffuse and specular shading values */
			
		if (calcDiffuse)
			if (dot > 0.0) {
				if (calcSpecular) {
					float highVal = dot + dot;
					
					/* Adjust the highlight value if the group has a non-linear highlight factor */
					
					highVal = (highVal * thePixel->shadeNormalA - normA) * thePixel->shadeXN + (highVal * thePixel->shadeNormalB - normB) * thePixel->shadeYN + (highVal * thePixel->shadeNormalC - normC) * thePixel->shadeZN;
					if (highVal > 0.0) {
						highVal = theAttributes->highlight * POW(highVal, theAttributes->spread);
						if (highVal > 1.0)
							highVal = 1.0;
						if (curRed > 0.0)	/* Components must be compared to prevent highlights from dark lights */
							hRed += curRed * highVal;
						if (curGreen > 0.0)
							hGreen += curGreen * highVal;
						if (curBlue > 0.0)
							hBlue += curBlue * highVal;
					}
				}
				
				/* Adjust the dot if the group has a non-linear terminator factor */
				
				if (theAttributes->terminator != 1.0)
					dot = POW(dot, theAttributes->terminator);
				sRed += curRed * dot;
				sGreen += curGreen * dot;
				sBlue += curBlue * dot;
			} else 	if (transmissionFlag) {
				
				/* Adjust the dot if the group has a non-linear terminator factor */
				
				if (theAttributes->terminator != 1.0)
					dot = POW(dot, theAttributes->terminator);
				sRed += curRed * dot * theAttributes->transmission.r;
				sGreen += curGreen * dot * theAttributes->transmission.g;
				sBlue += curBlue * dot * theAttributes->transmission.b;
			}
	} while (theLight);
	
	/* The global ambient will insure that illumination does not fall below a minimum level */
	
	sRed += thePixel->shadeAmbientR * theAttributes->ambient.r;
	sGreen += thePixel->shadeAmbientG * theAttributes->ambient.g;
	sBlue += thePixel->shadeAmbientB * theAttributes->ambient.b;

	temp = thePixel->shadeMinAmbientR * theAttributes->ambient.r;
	if (sRed < temp)
		sRed = temp;
	temp = thePixel->shadeMinAmbientG * theAttributes->ambient.g;
	if (sGreen < temp)
		sGreen = temp;
	temp = thePixel->shadeMinAmbientB * theAttributes->ambient.b;
	if (sBlue < temp)
		sBlue = temp;
	
	/* Start with the surface color */
	
	rNum = theAttributes->reference.r;
	gNum = theAttributes->reference.g;
	bNum = theAttributes->reference.b;
	
	/* Calculate the diffuse color */
	
	rNum *= theAttributes->diffuse.r;
	gNum *= theAttributes->diffuse.g;
	bNum *= theAttributes->diffuse.b;
	
	/* Calculate the color values by using the light source illumination values.  The glossy calculation is a simple image */
	/* beautification algorithm which adds white as the light intensity increases past 75%. */
	
	temp = theAttributes->gloss;
	if (sRed > temp)
		rNum = rNum * temp + sRed - temp;
	else
		rNum *= sRed;
	if (sGreen > temp)
		gNum = gNum * temp + sGreen - temp;
	else
		gNum *= sGreen;
	if (sBlue > temp)
		bNum = bNum * temp + sBlue - temp;
	else
		bNum *= sBlue;
	
	/* Factor the luminance into the color values */

	temp = theAttributes->shade;
	if (temp > 0.0) {
		rNum += temp * (theAttributes->reference.r - rNum);
		gNum += temp * (theAttributes->reference.g - gNum);
		bNum += temp * (theAttributes->reference.b - bNum);
	}
	
	/* The luminance color will be added to the shaded color */
	
	rNum += theAttributes->luminance.r;
	gNum += theAttributes->luminance.g;
	bNum += theAttributes->luminance.b;
	
	/* Add the highlight to the color values if any */
	
	if (BTST(theAttributes->shadeFlags2, groupShadeBiasSpecularFlag)) {
		float theSpecularValue = theAttributes->specular.r * 0.3 + theAttributes->specular.g * 0.59 + theAttributes->specular.b * 0.11;
		
		rNum += hRed * theSpecularValue * theAttributes->reference.r;
		gNum += hGreen * theSpecularValue * theAttributes->reference.g;
		bNum += hBlue * theSpecularValue * theAttributes->reference.b;
	} else {
		rNum += hRed * theAttributes->specular.r;
		gNum += hGreen * theAttributes->specular.g;
		bNum += hBlue * theAttributes->specular.b;
	}
	
	/* Add the reflection to the color values if any */
	
	if (BTST(theAttributes->shadeFlags2, groupShadeBiasReflectionFlag)) {
		rNum += theAttributes->reflection.r * thePixel->shadeReflection.r * theAttributes->reference.r;
		gNum += theAttributes->reflection.g * thePixel->shadeReflection.g * theAttributes->reference.g;
		bNum += theAttributes->reflection.b * thePixel->shadeReflection.b * theAttributes->reference.b;
	} else {
		rNum += theAttributes->reflection.r * thePixel->shadeReflection.r;
		gNum += theAttributes->reflection.g * thePixel->shadeReflection.g;
		bNum += theAttributes->reflection.b * thePixel->shadeReflection.b;
	}
	
	/* Limit color values into the range of 0..1 */
	
	if (rNum > 1.0)
		rNum = 1.0;
	else if (rNum < 0.0)
		rNum = 0.0;
	if (gNum > 1.0)
		gNum = 1.0;
	else if (gNum < 0.0)
		gNum = 0.0;
	if (bNum > 1.0)
		bNum = 1.0;
	else if (bNum < 0.0)
		bNum = 0.0;
	
	/* Return the final color value */
	
	thePixel->shadeColor.a = theAttributes->alpha;
	thePixel->shadeColor.r = rNum;
	thePixel->shadeColor.g = gNum;
	thePixel->shadeColor.b = bNum;

	return shaderNoError;
}  /* EIShaderShade */
