/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifndef EI_LAYERS_H
#define EI_LAYERS_H

#include "EIShader.h"

#ifndef __System_h__
 #include "System.h"
#endif 

#ifdef __cplusplus
extern "C" {
#endif

/* Layers Data Types */

#define LDATA_NONE				1	// none					(0 bytes)
#define LDATA_ARGB8				2	// ARGB color 8-bits		(4 bytes)
#define LDATA_RGB8_				3	// RGB color 8-bits	 	(3 bytes)
#define LDATA_GRAY8				4	// grayscale 8-bits	  	(1 byte)
#define LDATA_CHAR				5	// 1 character 			(1 byte) 
#define LDATA_SHORT				6	// 1 short/ushort 		(2 bytes) 
#define LDATA_LONG				7	// 1 long/ulong			(4 bytes) 
#define LDATA_FLOAT				8	// 1 float 				(4 bytes) 
#define LDATA_XYZ				9	// 3 floats x, y, z		(12 bytes)
#define LDATA_BIT_MASK			10	// variable size supported by host
#define LDATA_USER				100	// user-defined types start here

/* Layers Flags */

#define LFLAG_LAYER 			0	// layer is a Photoshop layer
#define LFLAG_MASK		 		1	// layer has a Photoshop layer's mask
#define LFLAG_CHANNEL 			2	// layer is a Photoshop channel
#define LFLAG_FOLDER 	 		3	// layer is a Photoshop folder
#define LFLAG_NO_MERGE	 	 	4	// layer's pixel should NOT be merged with previous (transparent) value 
#define LFLAG_PROCESS 	 		5	// layer should be passed for post-processing 
#define LFLAG_ADDITIVE	 		6	// layer is additive
#define LFLAG_NO_ALPHA	 		7 	// layer's alpha is ignored
#define LFLAG_GRP_MASK	 		8 	// layer is used as a mask of Photoshop folder
#define LFLAG_EDGE_ALPHA 		9	// layer's alpha is unpremultiplied with edge mask

/* Layers Contents */

#define LCONTENT_CUSTOM			-1	// custom (unknown) content
#define LCONTENT_DIFFUSE		1	// material's diffuse 
#define LCONTENT_LIGHTS			2	// sum of all lights values
#define LCONTENT_LUMINANCE		3	// luminance
#define LCONTENT_SPECULAR		4	// specular
#define LCONTENT_AMBIENT		5	// ambient
#define LCONTENT_REFLECTION		6	// material's reflections summary
#define LCONTENT_REFL_BM		7	// bitmap reflections 
#define LCONTENT_REFRACTION		8	// RT refractions
#define LCONTENT_TRANSPARENCY	9	// transparency
#define LCONTENT_FOG			10	// fog
#define LCONTENT_SHADE_COLOR	11	// final shaded (non-clamped) color
#define LCONTENT_COVERAGE		12	// sub-pixel's coverage (ARGB_Real)
#define LCONTENT_GROUP_ID		13	// group ID assigned at render time
#define LCONTENT_UV				14	// UV coordinates as ARGB_Real (a = 1 indicates "natural" UV's)
#define LCONTENT_PIXEL_MASK		15	// a packed array of bits representing the curent facet's visibility at each sub-pixel
#define LCONTENT_MOTION_BLUR	16	// motion blur x and y vectors
#define LCONTENT_LAST			LCONTENT_MOTION_BLUR

/* Error codes */

#define layerError		-1	

/* Commands passed in EILayerShade callback */

#define LSHADE_CMD_GET_LAYER		1	// get the actual layer's content
#define LSHADE_CMD_GET_PROPERTY		2	// get the property color
#define LSHADE_CMD_SET_LAYER		3	// set the layer's content
#define LSHADE_CMD_SET_PROPERTY		4	// set the property color
#define LSHADE_CMD_GET_PREV			5	// get the previous layer's content
			
#ifdef __MACINTOSH__
#if __LP64__
	#pragma pack(4)
#else
	#pragma options align = native
#endif
#endif

/* EI_LayerInfo structure */

typedef struct {
	UInt32 	layerID;				// unique layer's ID
	UInt32 	layerFlag;				// layer's flag (see "Layers Flags" above) 
	UInt32 	layerDataType;			// layer's data type (see "Layers Data Types" above)
	UInt32 	layerDataSize;			// sub-pixel's data size in bytes (used for custom or user-defined layers)
	SInt32  layerContentID;			// layer's content ID (see "Layers Contents" above) 
	uchar	layerDefault[64];		// default sub-pixel's value block
	UInt32  layerParentID;  		// ID of layers's parent (0 = none)
	UInt32  layerModePS; 			// layer's Photoshop mode (add, multiply etc.)
	PStr63	layerName;				// layer's name 
	UInt32	layerCreator; 			// signature of layer's creator
	void *	layerUserData;			// pointer to user-defined data
	SInt32	layerTransRange[2];		// range of phong transparency passes ("0" means all further passes)
	UInt32 	layerReserved[8]; 		// reserved, must be zero
} EI_LayerInfoRec, *EI_LayerInfoPtr;

#ifdef __MACINTOSH__
#pragma options align = reset
#endif


/* Layers creation and filling */

typedef SInt32 (*EILayerCreateProc)( EI_LayerInfoPtr );

/* Layers enumaration */

typedef SInt32 (*EILayerEnumProc)( EI_LayerInfoPtr theLayerInfo );

/* Layers reading/writing callbacks */

#ifndef DotColor
	#ifdef __LITTLE_ENDIAN__
	#define DotColor ARGB
	#else
	#define DotColor RGBA
	#endif
#endif
	
typedef SInt32 (*EILayersReadProc) ( L_Rect * theRect, UInt32 theLayersCount, UInt32 * theLayersID, uchar * theData );
typedef SInt32 (*EILayersWriteProc)( L_Rect * theRect, UInt32 theLayersCount, UInt32 * theLayersID, uchar * theData );
typedef SInt32 (*EILayersWriteSlabLineProc)( SInt32 theLine, DotColor * theColor, DotColor * theAlpha );

/* Prototypes for shader functions */

typedef struct {
	void * shaderData;								/* The pointer to the shader's private data */
	UInt32 theRenderPassID;							/* Not used, reserved for future expansion */
	EILayerCreateProc   HostLayerCreateProc;		/* The layer's create callback */
	EILayerEnumProc 	HostLayerEnumProc;			/* The layers' enumarating start callback */
} EIShaderInitLayersRec, *EIShaderInitLayersPtr;

typedef struct {
	void * shaderData;								/* The pointer to the shader's private data */

	L_Rect shadeRect;								/* The rendering rectangle */
	L_Rect shadeSlabRect;							/* The rendering slab rectangle */
	L_Rect shadeImageRect;							/* The whole rendering image rectangle */
	L_Rect shadeSlabImageRect;						/* The slab rendering image rectangle */
	float  shadePixelWidth;							/* The antialiasing pixel width */
	float  shadePixelScale;							/* The zoom factor */
	float  shadePixelRatio;							/* The y/x pixel ratio */
	float  shadeOffsetX, shadeOffsetY;				/* The x and y image offset */
	UInt32  shadePerspectiveFlag;					/* 1 for perspective rendering, 0 for isometric rendering */
	
	UInt32  shaderPassStartFlag;					/* 1 = pass start, 0 = pass end */
	UInt32  shaderPassNomer;						/* The nomer of phong transparency pass, last pass = bit 31 is set */

	char * shaderFullPSDName;						/* The full name of output PSD file (C string) */
	char * shaderSetName;							/* The name of layer's set the shader is applied to (C string) */
	char * shaderFileName;							/* The full name of the shader file (C string) */
		
	EILayersReadProc	HostLayersReadProc;			/* The layers read callback */
	EILayersWriteProc	HostLayersWriteProc;		/* The layers write callback */
	
} EIShaderSlabLayersRec, *EIShaderSlabLayersPtr;	

typedef struct {
	void *	shaderData;								/* The pointer to the shader's private data */

	L_Rect	shadeRect;								/* The rendering rectangle */
	L_Rect	shadeSlabRect;							/* The rendering slab rectangle */
	L_Rect	shadeImageRect;							/* The whole rendering image rectangle */
	L_Rect	shadeSlabImageRect;						/* The slab rendering image rectangle */
	float	shadePixelWidth;						/* The antialiasing pixel width */
	float	shadePixelScale;						/* The zoom factor */
	float	shadePixelRatio;						/* The y/x pixel ratio */
	float	shadeOffsetX, shadeOffsetY;				/* The x and y image offset */
	UInt32	shadePerspectiveFlag;					/* 1 for perspective rendering, 0 for isometric rendering */
	
	SInt32  shaderLayerID;							/* ID of post-processing layer */
	SInt32  shaderTransparency;						/* 1 if transparency is ON */

	PStr255 shaderStatusString;						/* The status string */
		
	EILayersWriteSlabLineProc HostLayersWriteSlabLineProc;		/* The layers write callback */
	
} EIShaderProcLayersRec, *EIShaderProcLayersPtr;	

/* Export function prototypes */

SInt32 EIShaderInitLayers( EIShaderInitLayersRec *theHostInitLayersPtr );
SInt32 EIShaderFillLayers( EIShaderShadeRec *theShadePtr );
SInt32 EIShaderSlabLayers( EIShaderSlabLayersRec *theHostSlabPtr );
SInt32 EIShaderProcLayers( EIShaderProcLayersRec *theHostProcPtr );

#define EIShaderInitLayersName		PS("EIShaderInitLayers")
#define EIShaderFillLayersName		PS("EIShaderFillLayers")
#define EIShaderSlabLayersName		PS("EIShaderSlabLayers")
#define EIShaderProcLayersName		PS("EIShaderProcLayers")

#ifdef __cplusplus
}
#endif

#endif // EI_LAYERS_H