/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifndef __FACTStuff_h__
#define __FACTStuff_h__

#ifndef __System_h__
#include "System.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* FACTStuff.h Copyright (c) 1988-1998 Electric Image Inc. */
/* FACT 3D Model File Format Copyright (c) 1988-1998 Electric Image Inc. */

/* ElectricImage(tm) 3D Model File Specification */
/* Version 2.0 */

/* This document describes the ElectricImage model file format.  The file type is 'FACT' and */
/* the creator is 'EIAM'.  All of the information is stored in the data fork.  The format is best */
/* suited for permanent disk based storage and is not intended to be used as a RAM based data */
/* structure.  The format is expandable and is both forward and backward compatible as */
/* described below.  There are no limits to a model's complexity or consistency.  The file */
/* type 'FACT' is short for 'Facet'. */

/* The purpose of FACT is to provide a single file format for the storage and exchange of 3D */
/* models for use in rendering and animation.  The format was designed to be both forward */
/* and backward compatible, memory efficient, capable of storing most forms of 3D data, and */
/* relatively simple to support for both import and export. */

/* IFF (Interchange File Format) Description */

/* For our 3D model data file, we use the IFF format.  IFF files are composed of Blocks */
/* of data.  Each Block consists of a four character Type, a long integer Size, and a List of */
/* bytes.  A special Block of Type 'FORM' is used to contain other Blocks.  In this way, a */
/* hierarchical structure can be created within an IFF file.  The List of a FORM Block consists */
/* of a Type followed by zero or more Blocks of data. */

/* Example #1:  A Block containing TEXT information could be created as follows: */
/* 	Entry					Kind		Bytes		Explanation */
/* 	'TEXT'					Type		4			This is the Block Type identifier */
/* 	12						Size		4			This is the Size of the List */
/* 	'Hello World<CR>'		List		12			This is the data contained in the Block's List */

/* Example #2:  A FORM Block TEST that contains two TEXT blocks could be created as follows: */
/* 	Entry					Kind		Bytes		Explanation */
/* 	'FORM'					Type		4			This is the FORM Block Type identifier */
/* 	31						Size		4			This is the Size of the List */
/* 	'TEST'					Type		4			This is the FORM Type identifier */
/* 	(TEXT 'Hello' $0)		Block		14			This is a TEXT Block containing 'Hello_' */
/* 	(TEXT 'There!')			Block		14			This is a TEXT Block containing 'There!' */

/* Any FORM can contain any Block, including other FORMs and even FORMs of its own Type. */
/* Of course, in the interest of programming sanity, an IFF file can be completely defined */
/* so that a program that must read the file need only know about the Blocks that it must handle. */
/* If the program encounters a Block it does not recognize or a Block that is in the wrong place */
/* in the file, it can simply skip it.  When a Block is read in, however, the program should not */
/* count on the block to be the proper size as defined in a file format.  This will allow newer */
/* versions of the file format to be created while maintaining compatibility with earlier versions. */
/* To implement this, the program should first initialize a data structure in memory and then read */
/* the Block from the file overwriting the structure.  If fewer than the desired bytes are read, */
/* the remaining parts of the structure will contain the default information.  If the Block is larger */
/* than the desired size, the extra bytes should be skipped.  In this way it is even possible to have */
/* older version programs read newer version files.  The program which reads the file should */
/* also not insist that a non essential Type is included.  Whenever possible, such blocks should */
/* be assigned default values before the file is read. */

/* One issue left unresolved in IFF file formats is whether a Size must always be even.  In some */
/* software which reads IFF files, the file is read into memory entirely before parsing occurs.  If */
/* the software is running on a 68000 based computer system, it may not be able to access an odd */
/* address easily.  Some implementations of IFF, therefore, always pad odd Size Blocks with a zero */
/* byte without changing the size to an even number.  The pad byte is simply assumed to follow the */
/* block.  In order to avoid this issue, all of the Blocks used in FACT will have an even Size.  In */
/* Example #2 above, the 'Hello' is followed by a zero character to denote the end of the string.  The */
/* Size of the block would be 6 bytes rather than 5 since the zero character is part of the block. */

/* Generally, an IFF file's chunks are contained within one FORM.  The name of this form defines */
/* the IFF file type.  This Type can be used to quickly verify the format of a file before it is read. */

/* For our data file (File Type 'FACT'), we will use the following Types: */

/* 	FORM	3DFL	3D model data FiLe */
/* 			FHDR		File header */
/* 			GRUP		Group information (1 or more in a 3DFL) */

/* 	FORM	FHDR	File header */
/* 			FINF		File information */

/* 	FORM	GRUP	Group data */
/*			GHDR		Group header */
/* 			CORD		Coordinate list -- 3 floats (4 byte), one Coordinate structure for each vertex */
/* 			DCOR		DCoordinate list -- 3 Doubles (8 byte), one DCoordinate structure for each vertex */
/*			CVRT		Color vertex list --  ARGB, one color structure for each vertex, follows coordinates */
/*			NVRT		Normal vertex list -- 3 floats, one normal structure for each vertex, follows coordinates */
/*			TVRT		Texture vertex list -- 3 floats, one structure for each vertex, follows coordinates */
/*			BVRT		Bump alignment vertex list -- 9 floats, one structure for each vertex, follows coordinates */
/* 			ELEM		Element list (contains both simple and complex polygons) */

/* 	FORM	GHDR	Group header */
/* 			GINF		Group information */
/* 			GLNK		Group hierarchy and linkage information */
/* 			GSTR		Group starting position */
/* 			GBLR		Group motion blur motion */
/* 			GATR		Group shading attributes */
/* 			TMAP		Texture map  information */
/* 			TMEI 		Texture map user interface information */
/* 			TATR 		Texture map shading attributes */
/* 			BMAP		Secondary texture map information (obsolete) */
/* 			BMEI 		Secondary texture map user interface information (obsolete) */
/* 			BATR 		Secondary texture map shading attributes (obsolete) */
/* 			RMAP		Reflectance map information */
/* 			RMEI		Reflectance map user interface information */

/* The list above represents the recommended order in which the data blocks should appear. */
/* None of the blocks are required unless they are specifically referenced by some other block. */
/* A CORD or DCOR block is required, for example, if it is referenced by an ELEM block.  The */
/* same rule applies for the order in which the blocks must appear.  A TMAP need not precede */
/* a TMEI or TATR block, but a CORD or DCOR block must precede an ELEM block.  Each block */
/* type should appear no more than once in a single FORM. */

/* Vertex colors (CVRT), normals (NVRT), texture positions (TVRT) must ben enabled by setting their */
/* respecive flags in the GINF structure's flags field.  There must be the same number of coordinates as */
/* there are enabled vertex values (number of CORD records=number of NVRT records, etc.)  The CORD */
/* block must preceed the CVRT, NVRT, TVRT and BVRT blocks (which can appear in any order) and the ELEM */
/* block must follow these. */

/* Notation Note:  In the description of flag bits, bit zero is the least significant bit.  For */
/* example, bit 7 is 128.  One is 1 and Zero is 0 unless otherwise noted.  Undefined */
/* variables and flag bits are always reserved and should be set to Zero. */

/* These are some predefined default values */

#ifdef __LITTLE_ENDIAN__
#define fieldARGB			0xFFFFFFFFL
#define fieldAlpha			0x000000FFL
#define fieldRed			0x0000FF00L
#define fieldGreen			0x00FF0000L
#define fieldBlue			0xFF000000L
#define fieldRGB			0xFFFFFF00L
#else
#define fieldARGB			0xFFFFFFFFL
#define fieldAlpha			0xFF000000L
#define fieldRed			0x00FF0000L
#define fieldGreen			0x0000FF00L
#define fieldBlue			0x000000FFL
#define fieldRGB			0x00FFFFFFL
#endif

/* FACT IFF Types */

#define kFORM	QUADBYTE('F', 'O', 'R', 'M')
#define k3DFL	QUADBYTE('3', 'D', 'F', 'L')
#define kFACT	QUADBYTE('F', 'A', 'C', 'T')
#define kGRUP	QUADBYTE('G', 'R', 'U', 'P')
#define kGINF	QUADBYTE('G', 'I', 'N', 'F')
#define kGBLR	QUADBYTE('G', 'B', 'L', 'R')
#define kGATR	QUADBYTE('G', 'A', 'T', 'R')
#define kTMAP	QUADBYTE('T', 'M', 'A', 'P')
#define kTATR	QUADBYTE('T', 'A', 'T', 'R')
#define kBMAP	QUADBYTE('B', 'M', 'A', 'P')
#define kBATR	QUADBYTE('B', 'A', 'T', 'R')
#define kRMAP	QUADBYTE('R', 'M', 'A', 'P')
#define kGHDR	QUADBYTE('G', 'H', 'D', 'R')
#define kCORD	QUADBYTE('C', 'O', 'R', 'D')
#define kDCOR	QUADBYTE('D', 'C', 'O', 'R')
#define kXCOR	QUADBYTE('X', 'C', 'O', 'R')
#define kCVRT	QUADBYTE('C', 'V', 'R', 'T')
#define kNVRT	QUADBYTE('N', 'V', 'R', 'T')
#define kTVRT	QUADBYTE('T', 'V', 'R', 'T')
#define kBVRT	QUADBYTE('B', 'V', 'R', 'T')
#define kELEM	QUADBYTE('E', 'L', 'E', 'M')
#define kSURF	QUADBYTE('S', 'U', 'R', 'F')

/* GATR Default Values */

#define defShadeFlags		0x2200		/* Phong shading, smoothing on, faster anti-aliasing, filter transparency enabled, */
										/* cast and receive shadows enabled */
#define defReferenceA		255
#define defReferenceR		255
#define defReferenceG		255
#define defReferenceB		255

#define defAmbientA			255
#define defAmbientR			255
#define defAmbientG			255
#define defAmbientB			255

#define defSpecularA		255
#define defSpecularR		255
#define defSpecularG		255
#define defSpecularB		255

#define defSpread			200.0

#define defTransparencyA	255
#define defTransparencyR	0
#define defTransparencyG	0
#define defTransparencyB	0

#define defRefraction		1.0

#define defReflectionA		255
#define defReflectionR		0
#define defReflectionG		0
#define defReflectionB		0

#define defLuminanceA		0
#define defLuminanceR		0
#define defLuminanceG		0
#define defLuminanceB		0

#define defLineWeight		0.0

#define defEdgeDropoff		0.0
#define defEdgeValue		1.0

#define defDiffuseA			255
#define defDiffuseR			255
#define defDiffuseG			255
#define defDiffuseB			255

#define defShadeFlags2		0x0000

#define defTransmissionA	255
#define defTransmissionR	0
#define defTransmissionG	0
#define defTransmissionB	0

#define defGlowA			0
#define defGlowR			0
#define defGlowG			0
#define defGlowB			0

#define defReflectionBlur	0.0
#define defRefractionBlur	0.0

#define defReflectionBlurType	  0
#define defReflectionBlurSamples  8	
#define defReflectionBlurAngle	  5.0

#define defRefractionBlurType	  0
#define defRefractionBlurSamples  8
#define defRefractionBlurAngle	  5.0

#define defCausticsAmount		  0.0
#define defCausticsFactor		  1.0
#define defCausticsDropoff		  1.0
#define defCausticsReceive		  1.0

#define defTranslucentAmount	  0.0	
#define defTranslucentThickness	  10.0
#define defTranslucentA			  255
#define defTranslucentR			  128
#define defTranslucentG			  128
#define defTranslucentB			  128
#define defTranslucentScatter	  1.0
#define defTranslucentFocus		  100.0
#define defTransparentThickness	  10.0

/* GINF Group Information Flags */

#define groupInfoInvisibleFlag				31	/* Invisible group */
#define groupInfoBreakFacetFlag				30	/* Break nonflat polygons */
#define groupInfoCorrectNormalFlag			29	/* Correct normals (either all clockwise or all counter-clockwise vertices) */
#define groupInfoDetailFirstFlag			28	/* Detail polygons first in list (reverse detail priority) */
#define groupInfoCycleFlag					27	/* Cycle child groups */
#define groupInfoFlattenComplexFlag			26	/* Flatten complex polygons */
#define groupInfoOutwardNormalFlag			25	/* The group normals all point outwards */
#define groupInfoHairLinesFlag				24	/* Lines reperesent hair segments */
#define groupInfoTextureVerticesFlag		23	/* The group contains texture vertex records */
#define groupInfoNormalVerticesFlag			22	/* The group contains normal vertex records */
#define groupInfoColorVerticesFlag			21	/* The group contains color vertex records */
#define groupInfoBumpVerticesFlag			20	/* The group contains bump vertex records */
#define groupInfoFogFlag					19	/* Disable camera fog for the group when this bit is set */
#define groupInfoCullBackFacetsFlag			18	/* Cull back facing facets */
#define groupInfoReverseBackFacetsFlag		17	/* Reverse the surface normal before culling back facing facets */
#define groupInfoColorIlluminationFlag		16	/* Use the color vertices as illumination values */
#define groupInfoShadeLinesFlag				15	/* Lightsource shade point and line facets */
#define groupInfoTraceReservedFlag			14	/* Unused -- set to zero */
#define groupInfoTraceReflectionFlag		13	/* Cast reflection rays */
#define groupInfoTraceRefractionFlag		12	/* Cast refraction rays */
#define groupInfoCameraRayFlag				11	/* The group is invisible to rays cast directly from the camera */
#define groupInfoReflectionRayFlag			10	/* The group is invisible to reflection rays */
#define groupInfoRefractionRayFlag			9	/* The group is invisible to refraction rays */
#define groupInfoShadowRayFlag				8	/* The group is invisible to shadow rays */
#define groupInfoReflectionObscurationFlag	7	/* The group's reflections are obscured by one bounce reflection rays */
#define groupInfoCylindricalUFlag			6	/* The group's texture u vertices are in 0..1 cylindrical coordinates */
#define groupInfoCylindricalVFlag			5	/* The group's texture v vertices are in 0..1 cylindrical coordinates */

#define groupInfoSampleLevel2				2	/* The group's sampling level is stored in the lowest three bits of the flags: */
#define groupInfoSampleLevel1				1	/* 0 = 1x1; 1 = 2x2; 2 = 4x4; 3 = 8x8; 4 = 16x16; 5 = 32x32; 6 = 64x64; */
#define groupInfoSampleLevel0				0	/* 7 = automatic (the same as the frame's maximum sampling level) */

/* GLNK Hierarchical Linking Types */

#define linkFree						0 	/* x,y,z rotation and x,y,z translations */
#define linkBallPlanar					1	/* x,y,z rotation and x,z translations only */
#define linkCylPlanar					2	/* y,z rotation and x,z translations only */
#define linkPlanar						3	/* y rotation and x,z translations only */
#define linkSocket						4	/* x,y,z rotations only */
#define linkCylinder					5	/* z rotations and z translations only */
#define linkUniversal					6	/* y,z rotations only */
#define linkSlide						7	/* z translation only */
#define linkPin							8	/* z rotation only */
#define linkLock						9	/* No transformations */

#define linkCustomBit					15	/* Set to 1 if custom movement limitations are defined in the other bits */
#define linkCustomLockXPosBit			0	/* Set to 1 if x translation is locked */
#define linkCustomLockYPosBit			1	/* Set to 1 if y translation is locked */
#define linkCustomLockZPosBit			2	/* Set to 1 if z translation is locked */
#define linkCustomLockXRotBit			3	/* Set to 1 if x rotation is locked */
#define linkCustomLockYRotBit			4	/* Set to 1 if y rotation is locked */
#define linkCustomLockZRotBit			5	/* Set to 1 if z rotation is locked */

/* GBLR Group Motion Blur Flags */

#define blurEnableFlag					31	/* 1 to enable motion blur for a group */
#define blurIntensityFlag				30	/* 1 to calculate the intensity for point/line motion blur */
#define blurFacetFlag					28	/* 1 for facet motion blur, 0 for point-line motion blur */

/* GOUT Group Outline Flags */

#define outlinePolygonFlag				31	/* 1 to enable polygon outlines */
#define outlineEdgeFlag					30	/* 1 to enable polygon outlines */
#define outlineSilhouetteFlag			29	/* 1 to enable polygon outlines */

/* GATR Group Shading Flags */

#define groupShadeColorBlendFlag		15	/* Blend average vertex colors across facets */
#define groupShadeGlossyFlag			14	/* Add white to brightly lit surfaces (obsolete, set to 0) */
#define groupShadeSmoothFlag			13	/* Enable smooth interpolation of facet normals for the group */
#define groupShadeSuperSampleFlag		12	/* Set the group's sampling level to 2x2 (obsolete - now set in the GINF) */
#define groupShadeMode3					11	/* Bits 8-11 are the shading mode (See GATR Shading Modes below) */
#define groupShadeMode2					10
#define groupShadeMode1					9
#define groupShadeMode0					8
#define groupShadeOverrideColorFlag		7	/* Override surface color with GATR reference color */
#define groupShadeTransparencyFlag		6	/* Disable filter filter transparency for the group */
#define groupShadeTransparencyAlphaFlag	5	/* Use facet and color vertex alpha channel as transparency */

#define groupShadeShadowMaskFlag		3	/* Set to cause the group to generate a shadow mask */
#define groupShadeShadowCastFlag		2	/* Set to prevent the group from casting shadows */
#define groupShadeShadowReceiveFlag		1	/* Set to prevent the group from receiving shadows */
#define groupShadeShadowPrecisionFlag	0	/* Set to use a high precision shadow buffer algorithm for the group */

/* GATR Group Shading Flags 2 */

#define groupShadeDisableGlobalIllumFlag	15	/* Disable Global Illumination */
#define groupShadeVolumeTransparencyFlag	14	/* Use volume transparency */
#define groupShadeDisableBackSpecFlag		13	/* Disable backside specular */
#define groupShadeEdgeShadowFlag			12	/* Use transparency edge control for shadows only */
#define groupShadeFresnelReflectionFlag		11	/* Change the reflectivity attributes based on the refraction index */
#define groupShadeTranslucentTransFlag		10	/* Use transparency for translucency calculation */					
#define groupShadeVolumeCausticsFlag		9	/* Calculate volume reflections for refraction caustics */					
#define groupShadeBiasCausticsFlag			8	/* Bias the reflective caustics color with the diffuse color */					
#define groupShadeSolidTransparencyFlag		7	/* Adjust the transparency for objects with front and back sides */
#define groupShadeBiasLuminanceFlag			6	/* Bias the luminance color with the diffuse color */
#define groupShadeAddTransparencyFlag		5	/* Use additive transparency mode (vs subtractive mode) */
#define groupShadeNewTransparencyFlag		4	/* Use the new transparency shading function */
#define groupShadeBiasTransparencyFlag		3	/* Bias the transparency color with the diffuse color */
#define groupShadeBiasSpecularFlag			2	/* Bias the specular color with the diffuse color */
#define groupShadeBiasReflectionFlag		1	/* Bias the reflection color with the diffuse color */
#define groupShadeFresnelRefractionFlag		0	/* Change the transparency attributes based on the refraction index */

/* GATR Group Shading Modes */

#define groupShadeWire						15	/* Only render points, lines and polygon edges */
#define groupShadeFlat						0	/* Compute a constant shading color for each facet */
#define groupShadeGouraud					1	/* Interpolate shading colors across facets */
#define groupShadePhong						2	/* Compute a shading color for each pixel */
#define groupShadeRayTrace					3	/* Ray trace a color for each pixel */

/* TATR Texture Attribute shadeFlags (obsolete) */

#define attributeSurface					15	/* Replace the surface color */
#define attributeAmbient					14	/* Replace the ambient color */
#define attributeSpecular					13	/* Replace the specular color */
#define attributeSpread						12	/* Replace the highlight spread factor */
#define attributeTransparency				11	/* Replace the transparency color */
#define attributeRefraction					10	/* Replace the refraction index */
#define attributeReflection					9	/* Replace the reflection color */
#define attributeLuminance					8	/* Replace the luminance color */
#define attributeEdgeTransparencyDropoff	7	/* Replace the edge transparency dropoff */
#define attributeEdgeTransparency			6	/* Replace the edge transparency color */
#define attributeDiffuse					5	/* Replace the diffuse color */
#define attributeTerminator					4	/* Replace the terminator factor */
#define attributeHighlight					3	/* Replace the highlight factor */

/* TMAP Texture Mapping Flags */

#define textureXModBit					15	/* Set to enable multiple copies of the texture in the x direction */
#define textureYModBit					14	/* Set to enable multiple copies of the texture in the y direction */
#define textureNegativeZBit				13	/* Set to allow texture mapping for coordinates with negative z values */
#define textureMipBit					12	/* Obsolete */
#define textureSmoothBit				11	/* Set to enable interpolation smoothing */
#define textureXFlipFlopBit				10	/* Set to flip-flop multiple copies of the texture in the x direction */
#define textureYFlipFlopBit				9	/* Set to flip-flop multiple copies of the texture in the y direction */
#define textureCFilterBit				8	/* Set to enable the custom shading attributes filter */
#define textureInvertRGBBit				7	/* Set to invert the map's RGB channels */
#define textureSharpBit					6	/* Obsolete */
#define textureSumBit					5	/* Set to filter the texture with rectangular direct sampling */
#define textureXClampBit				4	/* Set to clamp to the color of the nearest texture edge in the x direction */
#define textureYClampBit				3	/* Set to clamp to the color of the nearest texture edge in the y direction */
#define textureShaderModeBit			2	/* Set to use textureMode as shader option flags and channel modes (see below) */
#define textureInvertAlphaBit			1	/* Set to invert the map's alpha channel */
#define textureDisableBit				0	/* Set to disable the texture map */

/* TMAP Texture Alpha Types */

#define alphaNone						1	/* Do not use the alpha channel */
#define alphaMix						2	/* Mix the surface color with the ARGB texture color */
#define alphaReplace					3	/* Replace the surface color with the ARGB texture color */
#define alphaBump						4	/* Modify the surface normal by the change in the texture alpha */
#define alphaSurface					5	/* Factor the surface color by the texture alpha */
#define alphaAmbient					6	/* Factor the ambient color by the texture alpha */
#define alphaSpecular					7	/* Factor the specular color by the texture alpha */
#define alphaReflection				8	/* Factor the reflection color by the texture alpha */
#define alphaTransparency				9	/* Factor the transparency alpha by the texture alpha */
#define alphaLuminance					10	/* Factor the luminance color by the texture alpha */
#define alphaEdgeTransparency			11	/* Factor the edge transparency alpha by the texture alpha */
#define alphaDiffuse					12	/* Factor the diffuse color by the texture alpha */
#define alphaReflBlur					13	/* Factor the reflection blur angle by the texture alpha */
#define alphaLast	alphaReflBlur			/* The last texture alpha mode */

/* TMAP Texture RGB Types */

#define rgbNone						1	/* Do not use the rgb channel */
#define rgbSurface					2	/* Change the surface color */
#define rgbAmbient					3	/* Change the ambient color */
#define rgbSpecular					4	/* Change the specular color */
#define rgbReflection					5	/* Change the reflection color */
#define rgbTransparency				6	/* Change the transparency color */
#define rgbLuminance					7	/* Change the luminance color */
#define rgbEdgeTransparency			8	/* Change the edge transparency color */
#define rgbDiffuse					9	/* Change the diffuse color */
#define rgbReflBlur					10	/* Change the reflection blur angle */
#define rgbLast	rgbReflBlur				/* The last texture rgb mode */

/* TMAP Texture Mapping Types */

#define mapNone						0	/* Mapping disabled */
#define mapFlat						1	/* Map a flat texture (same as above) */
#define mapCylindrical					2	/* Map the texture around to a cylinder */
#define mapSpherical					3	/* Map the texture around to a sphere */
#define mapCubic						4	/* Choose the best side of a cube based upon the surface normal */
#define mapPlanar						5	/* Map to the current image pixel's coordinate */
#define mapPyramid					6	/* Map the texture with a pyramid projection */
#define mapSkyDome					7	/* Map the texture around to environment sphere */
#define mapAngular					8	/* Angular (LightProbe) map  */
#define mapVerticalCross				9	/* Cubic (vertical cross) map */

/* Additional experimental texture mapping modes -- these may change without notice */

#define mapCylindricalCaps				10	/* Map the texture around to a cylinder with flat caps */
#define mapSphericalCaps				11	/* Map the texture around to a sphere with flat caps */

/* The shader option flags and channel modes */

#define shaderOptionBlendFlag			15	/* Use the Photoshop blending modes in bits 8-11 */
#define shaderOptionMaskFlag			14	/* Use the alpha channel as a mask (textures only) */
#define shaderOptionInvertAlphaFlag		14	/* Invert the alpha channel (shaders only) */
#define shaderOptionReplaceRGBFlag		13	/* Replace the RGB channels with the alpha channel (textures only) */
#define shaderOptionInvertRGBFlag		13	/* Invert the RGB channels (shaders only) */
#define shaderOptionGrayscaleFlag		12	/* Convert the RGB channels into a gray scale: r * 0.299 + g * 0.587 + b * 0.114 */

/* Bits 8-11 contain the Photoshop blending modes */

#define shaderBlendMultiply			0
#define shaderBlendScreen				1
#define shaderBlendOverlay				2
#define shaderBlendSoftLight			3
#define shaderBlendHardLight			4
#define shaderBlendColorDodge			5
#define shaderBlendColorBurn			6
#define shaderBlendDarken				7
#define shaderBlendLighten				8
#define shaderBlendDifference			9
#define shaderBlendExclusion			10
#define shaderBlendHue					11
#define shaderBlendSaturation			12
#define shaderBlendColor				13
#define shaderBlendLuminosity			14
#define shaderBlendAdd					15

/* Bits 5-7 of the option flags contain the cut out, back side exclusion and disable uv flags */

#define shaderOptionCutoutFlag			7	/* Use the alpha as a cut out by maxing it with the opacity */
#define shaderOptionFrontSideFlag		6	/* Do not cover the back sides of facets */
#define shaderOptionDisableUVFlag		5	/* Use model instead of uv coordinates for texture&shader projection  */

/* The bits 0-4 of the option flags contain the channel mode */

#define shaderChannelNone				0	/* Do not use the shader color */
#define shaderChannelBump				1	/* Bump value */
#define shaderChannelDisplacement		2	/* (Obsolete, use shaderChannelBump instead) */
#define shaderChannelClip				3	/* Clip value */
#define shaderChannelOpacity			4	/* Opacity value */
#define shaderChannelGloss				5	/* Gloss value */
#define shaderChannelShade				6	/* Shade value */
#define shaderChannelAlpha				7	/* Alpha value */
#define shaderChannelBloom				8	/* Bloom value */
#define shaderChannelGlow				9	/* Glow RGB color */
#define shaderChannelReference			10	/* Reference RGB color */
#define shaderChannelDiffuse			11	/* Diffuse RGB color */
#define shaderChannelAmbient			12	/* Ambient RGB color */
#define shaderChannelSpecular			13	/* Specular RGB color */
#define shaderChannelReflection			14	/* Reflection RGB color */
#define shaderChannelTransparency		15	/* Transparency RGB color */
#define shaderChannelLuminance			16	/* Luminance RGB color */
#define shaderChannelTransmission		17	/* Transmission RGB color */
#define shaderChannelEdgeDiffuse		18	/* Edge diffuse value */
#define shaderChannelEdgeAmbient		19	/* Edge ambient value */
#define shaderChannelEdgeSpecular		20	/* Edge specular value */
#define shaderChannelEdgeReflection		21	/* Edge reflection value */
#define shaderChannelEdgeTransparency	22	/* Edge transparency value */
#define shaderChannelEdgeLuminance		23	/* Edge luminance value */
#define shaderChannelEdgeTransmission	24	/* Edge transmission value */
#define shaderChannelReflectionBlur		25	/* Reflection blur value */
#define shaderChannelRefractionBlur		26	/* Refraction blur value */
#define shaderChannelBumpRGB			27	/* RGB Bump value */
#define shaderChannelReflectionMap		28	/* Reflection map value */

#define factCreator	QUADBYTE('E', 'I', 'A', 'M')		/* Default FACT model file creator (ElectricImage Pro) */
#define factType	QUADBYTE('F', 'A', 'C', 'T')		/* Default FACT model file type */

/* FACT types */

typedef UInt32 IFFType;					/* The IFF four character type */
typedef unsigned char MediumInt[3];		/* 3 Byte short */

/* EXTENDED is a 10 byte IEEE floating point number in the FACT file. */
/* In memory it is an 8 byte double */

typedef double EXTENDED;

#ifdef __LITTLE_ENDIAN__

typedef union {
	IFFType theType;
	struct {
		unsigned char c4, c3, c2, c1;
	} c;
} TypeRecord, *TypeRecordPtr;

typedef union {
	short i;
	struct {
		unsigned char b2, b1;
	} b;
} IntRecord, *IntRecordPtr;

typedef union {
	SInt32 l;
	struct {
		short i2, i1;
	} i;
	struct {
		unsigned char b4, b3, b2, b1;
	} b;
} LongRecord, *LongRecordPtr;

typedef union {
	float r;
	SInt32 l;
	struct {
		short i2, i1;
	} i;
	struct {
		unsigned char b4, b3, b2, b1;
	} b;
} RealRecord, *RealRecordPtr;

typedef union {
	double d;
	struct {
		SInt32 l2, l1;
	} l;
	struct {
		short i4, i3, i2, i1;
	} i;
	struct {
		unsigned char b8, b7, b6, b5, b4, b3, b2, b1;
	} b;
} DoubleRecord, *DoubleRecordPtr;

#else

typedef union {
	IFFType theType;
	struct {
		unsigned char c1, c2, c3, c4;
	} c;
} TypeRecord, *TypeRecordPtr;

typedef union {
	short i;
	struct {
		unsigned char b1, b2;
	} b;
} IntRecord, *IntRecordPtr;

typedef union {
	SInt32 l;
	struct {
		short i1, i2;
	} i;
	struct {
		unsigned char b1, b2, b3, b4;
	} b;
} LongRecord, *LongRecordPtr;

typedef union {
	float r;
	SInt32 l;
	struct {
		short i1, i2;
	} i;
	struct {
		unsigned char b1, b2, b3, b4;
	} b;
} RealRecord, *RealRecordPtr;

typedef union {
	double d;
	struct {
		SInt32 l1, l2;
	} l;
	struct {
		short i1, i2, i3, i4;
	} i;
	struct {
		unsigned char b1, b2, b3, b4, b5, b6, b7, b8;
	} b;
} DoubleRecord, *DoubleRecordPtr;
#endif

typedef struct {
	short top;
	short left;
	short bottom;
	short right;
} SRect, *SRectPtr;

typedef struct {
	SInt32 top;
	SInt32 left;
	SInt32 bottom;
	SInt32 right;
	UInt32 width;
	UInt32 height;
} L_Rect, *LRectPtr;

typedef struct {
	float min, max;							/* Minimum and maximum values */
} MinMax, *MinMaxPtr;

typedef struct {
	float x, y, z;							/* x, y and z coordinates */
} Coordinate, *CoordPtr;

#ifndef EI_UNIVERSE	// plug-in mode

typedef struct {
	double x, y, z;							/* x, y and z coordinates */
} DCoordinate, *DCoordPtr;

#endif	// EI_UNIVERSE

typedef Coordinate Normal, *NormalPtr;

#ifndef EI_UNIVERSE	// plug-in mode

typedef struct {
	EXTENDED x, y, z;						/* x, y and z coordinates */
} XCoordinate, *XCoordPtr;

#endif	// EI_UNIVERSE

typedef struct {
	Coordinate minCoord;					/* Minimum x, y and z */
	Coordinate maxCoord;					/* Maximum x, y and z */
} Extent, *ExtentPtr;

typedef double Vector3[3];

typedef union {
	double matrixData[9];
	double matrix[3][3];
	Vector3 mv[3];
	struct {
		double m00, m01, m02;
		double m10, m11, m12;
		double m20, m21, m22;
	} m;
	struct {
		Vector3 mv0;
		Vector3 mv1;
		Vector3 mv2;
	} v;
} Matrix3, *Matrix3Ptr;

typedef EXTENDED XVector3[3];

typedef union {
	EXTENDED matrixData[9];
	EXTENDED matrix[3][3];
	XVector3 mv[3];
	struct {
		EXTENDED m00, m01, m02;
		EXTENDED m10, m11, m12;
		EXTENDED m20, m21, m22;
	} m;
	struct {
		XVector3 mv0;
		XVector3 mv1;
		XVector3 mv2;
	} v;
} XMatrix3, *XMatrix3Ptr;

typedef float RVector3[3];

typedef union {
	float matrixData[9];
	float matrix[3][3];
	RVector3 mv[3];
	struct {
		float m00, m01, m02;
		float m10, m11, m12;
		float m20, m21, m22;
	} m;
	struct {
		RVector3 mv0;
		RVector3 mv1;
		RVector3 mv2;
	} v;
} RMatrix3, *RMatrix3Ptr;

typedef double Vector4[4];

typedef union {
	double matrixData[16];
	double matrix[4][4];
	Vector4 mv[4];
	struct {
		double m00, m01, m02, m03;
		double m10, m11, m12, m13;
		double m20, m21, m22, m23;
		double m30, m31, m32, m33;
	} m;
	struct {
		Vector4 mv0;
		Vector4 mv1;
		Vector4 mv2;
		Vector4 mv3;
	} v;
} Matrix4, *Matrix4Ptr;

typedef EXTENDED XVector4[4];

typedef union {
	EXTENDED matrixData[16];
	EXTENDED matrix[4][4];
	XVector4 mv[4];
	struct {
		EXTENDED m00, m01, m02, m03;
		EXTENDED m10, m11, m12, m13;
		EXTENDED m20, m21, m22, m23;
		EXTENDED m30, m31, m32, m33;
	} m;
	struct {
		XVector4 mv0;
		XVector4 mv1;
		XVector4 mv2;
		XVector4 mv3;
	} v;
} XMatrix4, *XMatrix4Ptr;

typedef float RVector4[4];

typedef union {
	float matrixData[16];
	float matrix[4][4];
	RVector4 mv[4];
	struct {
		float m00, m01, m02, m03;
		float m10, m11, m12, m13;
		float m20, m21, m22, m23;
		float m30, m31, m32, m33;
	} m;
	struct {
		RVector4 mv0;
		RVector4 mv1;
		RVector4 mv2;
		RVector4 mv3;
	} v;
} RMatrix4, *RMatrix4Ptr;

typedef struct {
	double x, y, z, w;
} Quaternion, *QuaternionPtr;

/* Colors are stored as a packed 32 bit word */

#ifndef EI_UNIVERSE	// plug-in mode

typedef union {
	UInt32 colorValue;
	struct {
		unsigned char a, r, g, b;			/* alpha, red, green, and blue */
	} argb;
} ARGB, *ARGBPtr;

typedef ARGB ARGBPalette[256];
typedef ARGB *ARGBPalettePtr;

#endif	// EI_UNIVERSE

/* Alternative color storage format */

typedef union {
	UInt32 colorValue;
	struct {
		unsigned char r, g, b, a;			/* red, green, blue, and alpha */
	} argb;
} RGBA, *RGBAPtr;

/* 16 bit precision RGB color storage format without alpha */

typedef struct {
	short red, green, blue;
} RGB_Color, *RGB_ColorPtr;

/* 16 bit precision RGB color storage format with alpha */

typedef struct {
	short alpha, red, green, blue;
} ARGB_Color, *ARGB_ColorPtr;

/* 16 bit precision HSV color storage format without alpha */

typedef struct {
	short hue, saturation, value;
} HSV_Color, *HSV_ColorPtr;

/* 16 bit precision HSV color storage format with alpha */

typedef struct {
	short alpha, hue, saturation, value;
} AHSV_Color, *AHSV_ColorPtr;

/* Floating point RGB color */

#ifndef EI_UNIVERSE	// plug-in mode

typedef struct {
	float r, g, b;
} RGB_Real, *RGB_RealPtr;

/* Floating point ARGB color */

typedef struct {
	float a, r, g, b;
} ARGB_Real, *ARGB_RealPtr;

#endif	// EI_UNIVERSE

/* Floating point RGBA color */

typedef struct {
	float r, g, b, a;
} RGBA_Real, *RGBA_RealPtr;

/* The PointArray can be used to store vertices to points, lines, triangles and quadrangles.  The */
/* vertex indices are stored as unsigned integers.  Unused vertices contain zeros.  Vertices */
/* following a zero vertex should be skipped.  The entire polygon should be skipped if the first vertex */
/* is zero. */

/* Note that these structures are included as examples only.  In most systems, MediumInt */
/* will be four bytes long and not three so the vertices are usualy read from the file one at a time */
/* as a high order 8 bits followed by the low order 16 bits. */

typedef unsigned char BPointArray[4];		/* 4 Bytes */
typedef unsigned short IPointArray[4];		/* 8 Bytes */
typedef MediumInt MPointArray[4];		/* 12 Bytes */
typedef UInt32 LPointArray[4];			/* 16 Bytes */

/* The Element structure can define any kind of three dimensional component, from simple convex polygons */
/* with fewer than four edges to complex polygons and parametric surfaces.  An Element structure may also */
/* be used for special effects such as modification of shading attributes within a single group. */

/* The data within an Element is defined by the elementType field.  If the elementType=0, the Element */
/* contains a simple polygon as defined by the QuadPoly structure described below.  If the elementType=1, */
/* the Element contains a complex polygon as defined by the MultiPoly structure described below. */

/* The elementType values between 2 and 255 are not yet defined and should be ignored by skipping  */
/* elementSize bytes.  Because elementSize is a long, the possibilities of expansion within */
/* this structure are almost unlimited.  If the number of elementTypes are ever exhausted, types which */
/* contain sub-types could be defined. */

/* One problem which is typical of most complex data formats is that few applications can accept all */
/* of the types of data.  Because of this, a method is provided within the Element structure to provide for */
/* decomposition of complicated polygons and surfaces into simple polygons without having the problem of */
/* duplicated surfaces.  Each complex element (not a QuadPoly) which can be decomposed into simple */
/* QuadPoly elements will contain an elementSkip field.  If the application reading the file can accept */
/* the complex element, it will read that element and then ignore the simple elements that follow it as */
/* defined by the elementSkip field.  If the application cannot accept the complex element, it will be */
/* ignored and the sub elements into which it was decomposed will be read normally and will be used */
/* in place of the complex element. */

/* It is important to note that it is NOT required that an application which creates a FACT file must */
/* decompose the complex elements it writes into simple elements.  Since Electric Image will define and */
/* publish additions to the element structure, applications may depend upon Transporter to decompose */
/* any defined complex elements into simple elements.  This should minimize the amount of work required */
/* to convert from any 3D model format into the FACT format. */

/* The Element structure was designed to store polygons in an optimal amount of memory.  The Element */
/* structure is flexible enough to handle objects containing more than 65535 coordinates when necessary */
/* and yet it can also be used to store polygons in model of 256 or fewer coordinates without wasting */
/* any memory.  This is performed by using variable sized vertex indexes.  The size of an index is */
/* determined by the number of coordinates in the group which precede the Element block.  1..255 */
/* coordinates are indexed by bytes, 256..65535 are indexed by shorts,  65536..16777215 are */
/* indexed by 24 bit integers and 16777216 or more are indexed by 32 bit long integers. */

/* When reading a list of vertexes from a complex element, the polygon should be closed when the */
/* elementSize is reached or a zero index is read.  Any remaining bytes should be skipped (so that the */
/* number of bytes read is exactly equal to elementSize). */

/* QuadPoly is a sub structure of the Element structure which is used to define a simple polygon.  When the */
/* type of an Element structure is zero, it contains a QuadPoly structure.  The points field of this structure */
/* contain exactly four variable size coordinate indexes.  A QuadPoly can define a point, line, triangle */
/* or convex quadrangle.  Any unused vertices must contain zeros.  Only QuadPoly records will be */
/* rendered.  Complex polygons, defined by the MultiPoly structure below, are used only to display */
/* outlines and to define the normals of their sub polygons when rendered */

typedef struct {
	ARGB elementColor;						/* Surface color & matting alpha */
	union {
		BPointArray bPoints;
		IPointArray iPoints;
		MPointArray mPoints;
		LPointArray lPoints;
	} u;
} QuadPoly, *QuadPolyPtr;

/* MultiPoly is a sub structure of the Element structure.  When the type of an Element structure is 1, it indicates */
/* a complex polygon composed of any number of convex or concave vertices.  The number of vertices */
/* contained in the polygon is calculated by:  numVertices=(theSize-8) / indexSize.  The size of the */
/* structure must be even so there may be a zero pad byte at the end of the list. */

/* When the type of the Element structure is 2, it indicates a polyline structure composed of any number of */
/* polyline segments.  The number of polyline vertices contained in the polygon is calculated by: */
/* numVertices=(theSize-8) / indexSize.  The size of the structure must be even so there may be a zero */
/* pad byte at the end of the list. Each polyline consists of one or more vertex index numbers and is */
/* terminated by a zero or the end of the list.  There can be zero or more polylines in a single Element */
/* structure.  When reading a polyline element, a program should continue to read polylines until the end of */
/* the structure has been reached */

/* Points, lines, triangles and convex quadrangles should be written as QuadPoly records and not */
/* as MultiPoly records because MultiPoly records will only ben rendered in wireframe mode */

typedef struct {
	UInt32 elementSize;				/* The total size in bytes of the MultiPoly structure (always even) */
	ARGB elementColor;				/* Surface color & matting alpha */
	UInt32 elementSkip;				/* The number of QuadPoly records into which the MultiPoly was tesselated */
	union {
		unsigned char bVert;
		unsigned short iVert;
		MediumInt mVert;
		UInt32 lVert;
	} elementVertices[1];
} MultiPoly, *MultiPolyPtr;

typedef struct {
	UInt32 elementSize;				/* The total size in bytes of the MultiPoly structure (always even) */
	ARGB hairColor;							/* Hair color & opacity alpha */
	struct {
		union {
			unsigned char bVert;
			unsigned short iVert;
			MediumInt mVert;
			UInt32 lVert;
		} u;
		float thickness;
	} hairVertices[1];
} HairPoly, *HairPolyPtr;

/* MiscBlock is a sub structure of the Element structure.  At this time, such blocks are reserved and should be */
/* skipped by using theSize field. */

typedef struct {
	UInt32 elementSize;				/* The total size in bytes of the MiscBlock structure (always even) */
	void* elementBlock;						/* The data of the block */
} MiscBlock, *MiscBlockPtr;

typedef struct {
	unsigned char elementFlags;				/* The element's flags */
											/* Bits 0, 1: 0=not set, 1=bent facet, 2=broken facet, 3=cone facet */
	unsigned char elementType;				/* The type of element */
	union {
		QuadPoly elementQuadPoly;			/* A simple polygon with between 1 and 4 convex vertices */
		MultiPoly elementMultiPoly;			/* A complex polygon with any number of concave & convex vertices */
		HairPoly elementHairPoly;			/* A series of vertex/thickness segments representing a hair */
		MiscBlock elementMiscBlock;			/* Other entities which are yet to be defined and are reserved */
	} u;
} Element, *ElementPtr;

/* CORD is a list of vertex positions (stored as 12 byte Coordinate structures) */
/* DCOR is a list of vertex positions (stored as 24 byte DCoordinate structures) */
/* CVRT	is a list of vertex colors (stored as 4 byte ARGB structures) */
/* NVRT	is a list of vertex normals (stored as 12 byte Coordinate structures) */
/* TVRT is a list of vertex texture positions (stored as 12 byte Coordinate structures) */
/* BVRT is a list of vertex bump alignment vectors (stored as three 12 byte Coordinate structures) */
/* ELEM is a list of variable length Element records */

typedef struct {
	UInt32 translateVersion;		/* Translation library version number */
	UInt32 translateDate;			/* Date of the last model translation */
} TVER, *TVERPtr;

#ifdef __MACINTOSH__
#if __LP64__
	#pragma pack(4)
#else
	#pragma options align=mac68k
#endif
#endif

typedef struct {
	UInt32 totalCoordinates;		/* Total number of coordinates in the file */
	UInt32 totalPolygons;			/* Total number of polygons in the file */
	UInt32 totalGroups;				/* Total number of groups in the file */
	Extent extents;					/* File extents */
	UInt32 flags;					/* File flags */
	Coordinate anchor;				/* Anchor point for file */
} FINF, *FINFPtr;

/* Child group cycling allows an object to change shape over time.  Child cycling is specified by setting the 'cycle */
/* child groups' bit in the flags variable of a parent groups's GINF structure.  The number of children to cycle is */
/* specified in the cycleCount variable of the same GINF structure.  This number should always be less than or equal */
/* to the childCount variable.  The child groups which should be cycled must appear immediately following their parent */
/* group.  If the cycleCount of the parent is zero, all of its children will be cycled.  During an animation, one cycle */
/* child will be rendered per frame.  The parent's elements and its other cycled child group's elements will not be */
/* rendered.  The cycle parent group's 'invisible group' flag determines the visibility for all of its cycled children. */
/* The 'cycle child' bit should also be set for all cycled child groups for future compatibility.  The flag should be */
/* ignored if the cycle parent group's cycle count field is zero.  If a cycle parent group with no children will be */
/* rendered if it is visible. */

typedef struct {						/* Group Information */
	UInt32 totalCoordinates;		/* Total number of coordinates in the group */
	UInt32 totalPolygons;			/* Total number of polygons in the group */
	UInt32 groupTemp;				/* Reserved (should be zero) */
	Extent extents;					/* Group extents */
	UInt32 flags;					/* See "GINF Group Information Flags" above */
	unsigned char name[32];			/* Group name */
	UInt32 groupDate;				/* Date of last group modification */
	unsigned short groupID;			/* Group ID number */
	XMatrix4 coordinateAbsMatrix;	/* Coordinate absolute (flat) transformation matrix */
	XMatrix4 normalAbsMatrix;		/* Normal absolute (flat) transformation matrix */
	XMatrix4 coordinateRelMatrix;	/* Coordinate relative (hierarchical) transformation matrix */
	XMatrix4 normalRelMatrix;		/* Normal relative (hierarchical) transformation matrix */
	UInt32 childCount;				/* The number of child groups which follow not including their child groups */
	UInt32 childTotal;				/* The number of child groups which follow including their child groups */
	UInt32 cycleCount;				/* The number of child groups to cycle if cycle child groups bit is set (0 to cycle all) */
} GINF, *GINFPtr;

/* Linkage structure:  This structure works in addition to the hierarchical data structure but is not required */
/* Given the complexity of a hierarchical link structure, here are the finer points on the order of */
/* transformations applied with this structure: */
/* 			1. Offset to link point (-x,-y,-z) using localOffset - This is the rotation point around which the group rotates. */
/* 			2. Change to the local coordinate system using localRotation - Apply inverse rotation (-z,-x,-y). */
/* 			3. Do the geometry transformation using key framed motion in animation program. */
/* 			5. Change back to the global coordinate system using localRotation - Apply rotation (y,x,z). */
/* 			4. Undo offset to link point (x,y,z) using localOffset - Put group back to its original point. */
/* 			6. Change to link coordinate system using parentRotation (-z,-x,-y) - This is the initial rotation of the link. */
/* 			7. Offset (x,y,z) to the position on the parent using parentOffset - This is the point where the link is */
/* 			attached to the parent. */

typedef struct {
	unsigned short lType;					/* See "GLNK Hierarchical Linking Types" above */
	XCoordinate positionMinLimit;			/* Position minimum limit */
	XCoordinate positionMaxLimit;			/* Position maximum limit */
	XCoordinate rotationMinLimit;			/* Rotation minimum limit */
	XCoordinate rotationMaxLimit;			/* Rotation maximum limit */
	XCoordinate localOffset;				/* Local Coordinate System - offset to the world */
	XCoordinate parentOffset;				/* Parent link location - starting offset */
	XCoordinate parentRotation;				/* Parent link rotation - starting rotation */
	XCoordinate localRotation;				/* Local Coordinate System - rotation to the world */
	XMatrix4 preRotation;					/* Pre-rotation matrix: this will supercede parentRotation if defined. */
											/* Use parentRotation only if preRotation is all zeros to define prerotation. */
	UInt32 lFlag;							/* Flag used to define limits of hierarchical motion */
											/* Note: bit 31 = zero if old style hierarchy, bit 31 = 1 if new style hierarchy */
	XCoordinate	preScale;
} GLNK, *GLNKPtr;

/* Starting transformation for groups - this may be supplied to setup groups in space w/o calculating new points */
typedef struct {
	XCoordinate offset;						/* Starting offset */
	XCoordinate center;						/* Starting center of rotation */
	XCoordinate rotation;					/* Starting rotation/reference */
	XCoordinate scale;						/* Starting scale */
} GSTR, *GSTRPtr;

/* Group's motion blur motion structure */
typedef struct {
	UInt32 blurFlags;						/* See "GBLR Group Motion Blur Flags" above */
	float blurIntensity;						/* Blur light intensity factor */
	UInt32 blurLayer;						/* Motion blur layer number (reserved-set to zero) */
	XMatrix4 blurMatrix;					/* Coordinate absolute matrix from the previous frame */
} GBLR, *GBLRPtr;

/* Group outline structure */
typedef struct {
	UInt32 outlineFlags;					/* See "GOUT Group Outline Flags" above */
	ARGB outlinePolygonColor;				/* The color of the polygon outlines */
	float outlinePolygonThickness;			/* The thickness of the polygon outlines */
	ARGB outlineEdgeColor;					/* The color of the hard edge outlines */
	float outlineEdgeThickness;				/* The thickness of the hard edge outlines */
	ARGB outlineSilhouetteColor;			/* The color of the silhouette outlines */
	float outlineSilhouetteThickness;		/* The thickness of the silhouette outlines */
} GOUT, *GOUTPtr;

/* Material Attributes (0..100% factored into 0..255) */
typedef struct {
	unsigned short shadeFlags;				/* See "GATR Group Shading Flags" above */
	ARGB reference;							/* Surface color (overrides element color when override color flag is set) */
	ARGB ambient;							/* Ambient reflection color */
	ARGB specular;							/* Specular reflection color, alpha=gloss factor (255=no gloss; 0=100% gloss) */
	float spread;			   				/* Specular highlight spread exponent */
	ARGB transparency;						/* Transparency components, alpha=opacity (255=solid; 0=invisible) */
	float refraction;						/* Refraction index */
	ARGB reflection;						/* Reflection color factor for material, alpha=not used */
	ARGB luminance;							/* Self illumination components, alpha=shade factor */
	float lineWeight;						/* The thickness of all lines and/or points in the group (0.0=use default) */
	float edgeTransparencyDropoff;			/* The degree to which the transparency changes from the center to the edge (0.0=none) */
	ARGB edgeTransparency;					/* The transparency color at the edge of an object (not used in 2.8 and beyond) */
	ARGB diffuse;							/* Diffuse reflection color */
	float terminator;						/* Terminator dropoff factor */
	float highlight;						/* Highlight dropoff factor */
	
	/* The following has been added in version 2.8 */
	
	float edgeDiffuseValue;					/* The diffuse value at the edge of an object */
	float edgeSpecularValue;				/* The specular value at the edge of an object */
	float edgeReflectionValue;				/* The reflection value at the edge of an object */
	float edgeAmbientValue;					/* The ambient value at the edge of an object */
	float edgeLuminanceValue;				/* The luminance value at the edge of an object */
	unsigned short shadeFlags2;				/* See "GATR Group Shading Flags 2" above */
	ARGB transmission;						/* The diffuse shading transmission color, alpha=not used */
	float edgeTransmissionValue;			/* The diffuse shading transmission value at the edge of an object */
	ARGB glow;								/* The glow color, alpha=shaded glow factor */
	
	float edgeDiffuseDropoff;				/* The degree to which the diffuse changes from the center to the edge (0.0=none) */
	float edgeSpecularDropoff;				/* The degree to which the specular changes from the center to the edge (0.0=none) */
	float edgeReflectionDropoff;			/* The degree to which the reflection changes from the center to the edge (0.0=none) */
	float edgeAmbientDropoff;				/* The degree to which the ambient changes from the center to the edge (0.0=none) */
	float edgeLuminanceDropoff;				/* The degree to which the luminance changes from the center to the edge (0.0=none) */
	float edgeTransmissionDropoff;			/* The degree to which the transmission changes from the center to the edge (0.0=none) */
	float edgeTransparencyValue;			/* The transparency value at the edge of an object */
	
	float reflectionBlur;					/* The reflection blur factor; 1.0 = no blur */
	float refractionBlur;					/* The refraction blur factor; 1.0 = no blur */
	
// added October 2003
	SInt32  reflectionBlurType;
	SInt32  reflectionBlurSamples;
	float	reflectionBlurAngle;

	SInt32  refractionBlurType;
	SInt32  refractionBlurSamples;
	float	refractionBlurAngle;

// added February 2004	
	float refrCausticsAmount;
	float refrCausticsFactor;
	float refrCausticsDropoff;
	float refrCausticsReceive;
	
	float reflCausticsAmount;
	ARGB  reflCausticsColor;
	float reflCausticsFactor;
	float reflCausticsDropoff;
	float reflCausticsReceive;
	
// added May 2005
	float translucentAmount;
	float translucentThickness;
	ARGB  translucentColor;
	float translucentScatter;
	float translucentFocus;
	
	float transparentThickness;
} GATR, *GATRPtr;							/* This Block will be extended later to contain more shading information. */

/* Group's texture mapping structure (also used for reflection maps) */
typedef struct {
	unsigned short textureFlags;				/* See "TMAP Texture Mapping Flags" above */
	unsigned short textureType;				/* Alpha mode * 256 + RGB mode or shader option flags and channel modes (see above) */
	unsigned char textureTypeName[32];		/* Reserved - Set to nil string */
	unsigned short mappingType;				/* See "TMAP Texture Mapping Types" above */
	unsigned char mappingTypeName[32];		/* Reserved - Set to nil string */
	XMatrix4 tMatrix;						/* The 4 by 4 transformation matrix for the original group coordinates */
	unsigned char textureName[32];			/* The file name of the associated texture image file */
	SRect textureRect;						/* The rectangle within the image file that will be used for texture mapping */
	XCoordinate offset;						/* The offset of the mapped point within the texture */
	XCoordinate scale;						/* The scale of the mapped point within the texture */
	UInt32 textureReference;				/* The directory reference ID of the texture map */
	unsigned char textureVolName[32];			/* The volume name of the texture map */
	EXTENDED bumpScale;						/* The scale factor applied to the bump normal */
	XMatrix4 bumpMatrix;					/* The rotational matrix for the bump */
	SInt32 textureFrame;					/* The frame of the associtated animation file to use */
	float textureTime;						/* The time as a real number (for custom textures) */
	float textureDensity;					/* The density factor for the texture: >1 blurs; <1 sharpens */
	unsigned short textureCutoff;				/* The maximum density allowed in the map (0=disable) */
	XCoordinate backOffset;					/* Reserved -- must be equal to offset */
	XCoordinate backScale;					/* Reserved -- must be equal to scale */
	XCoordinate leftOffset;					/* Reserved -- must be equal to offset */
	XCoordinate leftScale;					/* Reserved -- must be equal to scale */
	XCoordinate rightOffset;				/* Reserved -- must be equal to offset */
	XCoordinate rightScale;					/* Reserved -- must be equal to scale */
	XCoordinate topOffset;					/* Reserved -- must be equal to offset */
	XCoordinate topScale;					/* Reserved -- must be equal to scale */
	XCoordinate bottomOffset;				/* Reserved -- must be equal to offset */
	XCoordinate bottomScale;				/* Reserved -- must be equal to scale */
} TMAP, *TMAPPtr;

/* The optional TATR structure contains the shading attributes for the texture map.  The TATR structure is only used */
/* if the texture attributes bit is set in textureFlags, the texture alpha mode is alphaMix, and the TATR structure */
/* actually exists.  The shadeFlags field is used as a set of enable flags, one for each value in the structure.  For */
/* future compatibility, all unused flags will be zero. */

typedef GATR TATR, *TATRPtr;

/* Additional Texture map structure - EI will not accept TMAP without this structure. */
/* It will be extended in the future.  This structure is used by ElectricImage to setup and modify the parameters */
/* of a texture map.  Camera only uses the TMAP, RMAP and BMAP records for rendering. */

typedef struct {
	UInt32 texFlags;						/* Additional Texture map flag */
	SRect fullRect;							/* Full texture rectangle before cropping */
	XCoordinate position;					/* Texture origin: usually group's midpoint */
	XCoordinate offset;						/* Offset from texture origin: group is rotated around it */
	XCoordinate rotation;					/* Rotation of group */
	short textureVolumeIndex;				/* Volume index of texture */
	SInt32 textureDirID;					/* Directory ID of texture */
	union {									/* Access these values according to the mapping type TMAP */
		struct {
			EXTENDED imageScaleX;			/* Size factor for height: default is 1.0 (texture height * size factor) */
			EXTENDED imageScaleY;			/* Size factor for width: default is 1.0 (texture width * size factor) */
			EXTENDED imageScaleZ;			/* Size factor for depth: default is 1.0 (infinite positive direction) */
		} Flat;
		struct {
			EXTENDED cylinderRadius;		/* Base Radius: default radius of enclosing sphere for group */
			EXTENDED cylinderWrapAngle;		/* Degree of texture wrap : default 360.0° (full map around perimeter) */
			EXTENDED cylinderHeightScale;	/* Height factor: default 1.0 (texture height * height factor) */
			EXTENDED cylinderHeight;		/* Real Height */
		} Cylindrical;
		struct {
			EXTENDED sphereRadius;			/* Radius: default radius of enclosing sphere of group */
			EXTENDED sphereWrapAngle;		/* Degree of longitude texture wrap: default 360° (full map around sphere) */
			EXTENDED sphereBandAngle;		/* Degree of latitude texture wrap: default 180° (full map pole to pole) */
			EXTENDED sphereBandAngleStart;	/* Degree of latitude start: default 180° (start from "north" pole) */
		} Spherical;
		struct {
			EXTENDED angularBandAngle;		/* Degree of latitude texture wrap: default 180° (full map pole to pole) */
		} Angular;
	} u;
} TMEI, *TMEIPtr;

/*

TMEI Documentation

The TMEI block was added to tell the interface of Electric Image Animation system
how to draw the texture projection primitive.  It adds additional fields for the
texture image file location.


Flags

The field "texFlags" must default to 0x80000000.


Image Rectangles

Electric Image uses the image size (in pixel) to determine the scale factors for
the texture projections.  The field "fullRect" contains the enclosing of the image.
For example, if the image is 640 by 480 pixels, the rectangle in "fullRect" is

fullRect.left = 0,
fullRect.top = 0,
fullRect.right = 640,
fullRect.bottom = 480.

The rectangle field "textureRect" in TMAP contains the cropping rectangle of
the image.  This rectangle specifies the actual rectangular region of the image
to be used and is a subset of the full rectangle "fullRect" in TMEI.  The
default size of "textureRect" is "fullRect" (use the whole image).


Specifying the texture image file location

The TMAP structure defines the name, volume name, and directory ID of the texture
image file.  For historical purposes, the TMEI duplicates the directory ID and
adds the volume index to help find the texture image file.  The field "textureDirID" in
TMEI contains the same directory ID as the field "textureReference" in TMAP.
Additionally, the field "textureVolumeIndex" contains the index of the volume in
which to find the texture image file.  The current version of EI uses the TMEI to
setup its internal file reference and, therefore, must be set.


Projection geometry

The interface draws projection primitives to help the user to align the textures
to the underlying goemetry.  There are three projection types which need the
additional geometry information to draw the primitives: Flat, Cylindrical, and
spherical.

Flat

This projection works like a projector throwing an image onto the geometry.
The fields "imageScaleX" and "imageScaleY" provide a way of scaling the image
to fit the goemetry.  Each pixel of the image corresponds to one unit in floating
point space.  In other words, 1 pixel = 1.0.  The scale factor allows us to modify
that relationship to make the image appear larger or smaller on the geometry.
For default scale, you may choose a side of the extent box enclosing the geometry
and compute the scale by taking the extent box size and dividing it by the image
size.  Use the cropping rectangle for image size of you use a cropping rectangle.

For example, if you have a cube with 10.0 units each side and a texture image of
size 640 by 480, then the scaling factor of the texture map, if the map is projected
on the front of the cube, is

			imageScaleX = 10.0 / 640 = 0.0156;
			imageScaleY = 10.0 / 480 = 0.0208;

The "imageScaleZ" parameter specifies which direction the image is being projected.
In the interface, there is an arrow perpenticular to the texture plane primitive.
It is used to indicate which direction the projection is going to be done if
the user has unselected "Negative Z" in the interface.  The number in "imageScaleZ"
is not as important as its sign.  However, since the arrow is drawn using the
"imageScaleZ", it is best to copy either "imageScaleX" or "imageScaleY" into "imageScaleZ".

If the geometry is flat (2 dimensional), care should be taken, so that the map optimally
projects onto the face of the flat geometry and not on its side.

Cylinder

This projection works by enclosing the geometry with a cylinder.
The field "cylinderRadius" defines the radius of the cylinder.  It is computed
so that it encloses the extent box of the geometry.  This depends on which
axis the cylinder is specified.
The field "cylinderHeight" defines the height of the cylinder.  It is computed
depending on the axis to which the cylinder is aligned.
It is up the developer which axis to use to compute the "cylinderRadius" and
"cylinderHeight".  In the interface, the align tool allows the user to indicate
which axis the cylinder best fits.
The field "cylinderHeightScale" defaults to 1.0 to indicate that the complete
"cylinderHeight" is used to define the scale of the Y dimension of the texture
image.
The field "cylinderWrapAngle" defaults to 360.0 with indicates that the image is
wrapped completely around the geometry (360 degrees) in the X dimension.

Sphere

This projection works by enclosing the geometry with a sphere.  The fields
"sphereWrapAngle", "sphereBandAngle", and "sphereBandAngleStart" are used to
specify how to map the texture image onto goemetry.  The field "sphereRadius"
is used to draw the sphere primitive in the interface.
The field "sphereRadius" contains the radius of the enclosing sphere primitive
and is computed from the midpoint to the furthest corner of the enclosing
extent box of the goemetry.
The field "sphereWrapAngle" defaults to 360.0 with indicates that the image is
wrapped completely around the geometry (360 degrees) in the X dimension.
The field "sphereBandAngle" defaults to 180.0 which wraps the Y dimension of the image
from pole to pole (180 degrees).  A lower value compresses image around the equator.
The field "sphereBandAngleStart" defaults to 180.0 which starts the Y dimension wrapping
at the pole.

General Transformation

All projections use the following transformations.  The field "position" indicates
the origin of the texture within the geometry.  This is the starting point
of the transformation and it is in the local coordinate space of the
geometry.  Usually, it is the mid point of the extent box of the
goemetry, but may be specified somewhere else for a better fit.
The field "offset" specfies the offset of the texture from its origin ("position").
This allows the user to move the texture somewhere other than its default
position.  The default value is (0.0, 0.0, 0.0).
The field "rotation" allows the texture to rotated around its rotation point.
The rotation point of a flat texture is the center of the flat primitive.
The rotation point of a cylindrical texture is the center of the cylinder primitive.
The rotation point of a spherical texture is the center of the sphere primitive.
The default value is (0.0, 0.0, 0.0).

In any case, use the Electric Image texture interface to check what kind of
numbers are being generated when you apply the align tool.
*/

/* Texture maps, secondary texture maps and reflection maps all use the same structures.  In EIAS version 2.8, the */
/* secondary texture map is not used.  Instead, there can an unlimited number of primary texture maps and reflection */
/* maps assigned to a group. */

/* The secondary texture map structure is identical to the texture map structure */

typedef TMAP BMAP, *BMAPPtr;

/* The secondary texture map attribute structure is identical to the group attribute structure */

typedef GATR BATR, *BATRPtr;

/* The secondary texture map additional structure is identical to the texture map additional structure */

typedef TMEI BMEI, *BMEIPtr;

/* The reflection map structure is identical to the texture map structure */

typedef TMAP RMAP, *RMAPPtr;

/* The reflection map additional structure is identical to the texture map additional structure */

typedef TMEI RMEI, *RMEIPtr;

/* The following records define the image animation header and image header: */

#define imageCreator	QUADBYTE('E', 'I', 'A', 'M')
#define imageType		QUADBYTE('E', 'I', 'D', 'I')

typedef struct {
	unsigned short imageVersion;				/* Image file version (5) */
	UInt32 imageFrames;						/* Number of frames in the file (1..?) */
} ImageHeader, *ImageHeaderPtr;

typedef struct {
	float frameTime;							/* Time of frame (0.0) */
	SRect frameRect;						/* Frame Rectangle */
	unsigned char frameBitDepth;				/* Bits Per Pixel (not including alpha) */
	unsigned char frameType;					/* Pixel Type (0=Direct; 1=Indexed, 2=Z-Buffer) */
	SRect framePackRect;					/* Packing rectangle */
	unsigned char framePacking;				/* Packing Mode (0=Not Packed; 1=RL Encoding) */
	unsigned char frameAlpha;					/* Alpha Bits per pixel */
	UInt32 frameSize;						/* Size in bytes of the body of the image */
	unsigned short framePalettes;				/* Number of entries in the color table (1..256) */
	unsigned short frameBackground;				/* The index of the background color (0) */
} ImageFrame, *ImageFramePtr;

#ifdef __MACINTOSH__
#pragma options align=reset
#endif

/* Standard data initialization procedures */

void SetARGB(ARGBPtr theARGB, unsigned char aVal, unsigned char rVal, unsigned char gVal, unsigned char bVal);

#ifndef EI_UNIVERSE
ARGB MakeARGB(unsigned char aVal, unsigned char rVal, unsigned char gVal, unsigned char bVal);
#endif

void PackRGB(ARGB *theARGB, float r, float g, float b);
void UnpackRGB(ARGB *theARGB, float *r, float *g, float *b);
void SetSRect(SRectPtr theSRect, short theLeft, short theTop, short theRight, short theBottom);
void OffsetSRect(SRectPtr theSRect, SInt32 hOffset, SInt32 vOffset);
void SetLRect(LRectPtr theLRect, SInt32 theLeft, SInt32 theTop, SInt32 theRight, SInt32 theBottom);
void SRectToLRect(SRect *theRect, LRectPtr theLRect);
void LRectToSRect(LRectPtr theLRect, SRect *theRect);
void OffsetLRect(LRectPtr theLRect, SInt32 hOffset, SInt32 vOffset);
void InsetLRect(LRectPtr theLRect, SInt32 x, SInt32 y);
void OutsetLRect(LRectPtr theLRect, SInt32 x, SInt32 y);
int EmptyLRect(LRectPtr theLRect);
void UnionLRect(LRectPtr srcA, LRectPtr srcB, LRectPtr dst);
int SectLRect(LRectPtr srcA, LRectPtr srcB, LRectPtr dst);
void InitMatrix3(Matrix3Ptr theMatrix);
void InitXMatrix3(XMatrix3Ptr theMatrix);
void InitRMatrix3(RMatrix3Ptr theMatrix);
void InitMatrix4(Matrix4Ptr theMatrix);
void InitXMatrix4(XMatrix4Ptr theMatrix);
void InitRMatrix4(RMatrix4Ptr theMatrix);
void XMatrix3ToMatrix3(XMatrix3Ptr srcMatrix, Matrix3Ptr dstMatrix);
void Matrix3ToXMatrix3(Matrix3Ptr srcMatrix, XMatrix3Ptr dstMatrix);
void XMatrix4ToMatrix4(XMatrix4Ptr srcMatrix, Matrix4Ptr dstMatrix);
void Matrix4ToXMatrix4(Matrix4Ptr srcMatrix, XMatrix4Ptr dstMatrix);
void InitFINF(FINFPtr theFINF);
void InitGINF(GINFPtr theGINF);
void InitGLNK(GLNKPtr theGLNK);
void InitGSTR(GSTRPtr theGSTR);
void InitGBLR(GBLRPtr theGBLR);
void InitGOUT(GOUTPtr theGOUT);
void InitGATR(GATRPtr theGATR);
void InitTMAP(TMAPPtr theTMAP);
void InitTATR(TATRPtr theTATR);
void InitTMEI(TMEIPtr theTMEI, UInt32 theMapType);
void InitBMAP(BMAPPtr theBMAP);
void InitBATR(BATRPtr theBATR);
void InitBMEI(BMEIPtr theBMEI, UInt32 theMapType);
void InitRMAP(RMAPPtr theRMAP);
void InitRMEI(RMEIPtr theRMEI, UInt32 theMapType);

#ifdef __cplusplus
}
#endif

#endif
