/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#ifndef __PluginClient_h__
#define __PluginClient_h__

#ifndef __FACTStuff_h__
#include "FACTStuff.h"
#endif

#ifndef __PString_h__
#include "PString.h"
#endif

#ifndef __PluginHeaders_h__
#include "PluginHeaders.h"
#endif

#ifndef __System_h__
#include "System.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* This file contains the interfaces for all ElectricImage plugin types.  Message records are passed between */
/* the host application and the plugin code resource.  The host initiates all plugin operations.  During a call */
/* to a plugin, the plugin may call back the host to request functions that only the host can provide such as */
/* allocating memory. */

/* A call to the plugin looks like: */

/* theResult := thePlugin(theCommand, SizeOf(theCommandRec), &theCommandRec); */

/* theCommand is a four character IFFType which the plugin uses to determine what operation to perform. */
/* these commands are defined as constants below. */

/* theCommandRec is a record containing all of the parameters which are necessary for the plugin to perform */
/* the requested operation.  This record always contains a sub record called theInfo.  theInfo contains sizes of */
/* and pointer references to the plugin's control and data records.  It also contains a host reference pointer */
/* which is passed back to the host during a call back. */

/* thePlugin is a pointer to the plugin code resource which must be locked in memory while the plugin is being */
/* called.  At this time, the plugin is responsible for loading its other code segments if it is a multi-segment */
/* code resource.  These segments should be made purgable when the plugin returns to the host application. */
/* This is very preliminary, however, and is expected to change in the future.  Thus far we have not */
/* implemented a multi-segment code resource plugin. */

/* The plugin routine returns a long result code.  Zero is the usual result code and indicates that there */
/* were no errors.  The other result codes are defined below as constants. */

/* A call back to the host looks like: */

/* theResult := theHostProc(theCommand, SizeOf(theCommandRec), &theCommandRec); */

/* theCommand is a four character IFFType which the host uses to determine what operation to perform. */
/* these commands are defined as constants below. */

/* theCommandRec is a record containing all of the parameters which are necessary for the host to perform */
/* the requested operation.  This record always contains a pointer called theReference which is was */
/* passed to the plugin in theInfo.  The host uses this pointer to determine which plugin called it and which */
/* data the plugin is processing.  The plugin has no knowledge of the type of information which this pointer */
/* references so that it is free to change in each host application. */

/* theHostProc is a pointer to the host's call back routine which is passed to the plugin in theInfo. */

/* The host routine returns a long result code.  Zero is the usual result code and indicates that there */
/* were no errors.  The other result codes are defined below as constants. */

/* A host must not make another call to the same plugin during a call back. */

/* The plugin has access to two blocks of memory for storage of permanent or temporary variables.  The */
/* control record contains all of the control variables which the plugin itself sets up with its own user */
/* interface.  This record is maintained by the host applications and survive across plugin activations.  The */
/* data record is used for temporary storage by the plugin and only exists while the plugin is active.  The */
/* host applications have no knowledge of the contents of these records but are responsible for their */
/* maintenance and memory allocation.  These records can become quite large since a plugin may need to */
/* keep track of a large number of elements such as particle records in a particle system plugin.  The */
/* records are referenced by pointer but the host may choose to store them in memory as either handles */
/* or pointers.  When stored as handles, these records can move in memory between plugin calls so it */
/* is required that there must be no pointers to other variables stored in these records because the */
/* addresses may become invalid between plugin calls.  The correct method is to store offsets from the */
/* beginning of the records.  The host must always lock down the control and data records before calling */
/* a plugin. */

/* Whenever a plugin is called, it should validate the contents of its control and data records and allocate */
/* or re-allocate them as necessary.  Usually a single routine can be responsible for the validation of */
/* both the control and data records.  The example plugins are implemented in this way and may be */
/* used as a reference for your own plugins.  The host application expects that all memory allocation */
/* should be done during the Initialize, Interface and Setup calls.  Memory allocation during other calls */
/* may fail.  Memory allocation during a Generate call is very likely to fail because Camera has no spare */
/* memory for allocation during that phase.  If the plugin receives an error result code from a call back */
/* to the host which is critical to its operation, it should exit and pass the same result code back to the */
/* host. */

/* There are several commands which are common to all classes of plugins and host applications and there */
/* are commands which are unique to a particular class of plugin and host.  New commands may be added */
/* to this specification and individual command records may also be expanded with new parameter */
/* variables.  The plugin and host must both verify that the size of the record they receive matches the */
/* size which they expect.  The current implementation accepts larger records but not smaller ones.  This */
/* should allow older plugins to be used with newer host applications.  We may also allow host call backs */
/* to accept smaller records in future versions for the same reason.  Calls to the plugin or the host must */
/* not assume that the requested command is implemented.  The hostCheck and pluginCheck commands may */
/* be used to discover whether a particular command is implemented, however, even the check command */
/* is not required (but is recommended). */

/* When a plugin is called, its resource file may or may not be open.  If the plugin wishes to access one of */
/* its own resources it must call the host with the hostOpenResource command.  The plugin should only call */
/* hostCloseResource if it needs to access another resource file. */

/* A typical model plugin might be called from Camera with the following commands */
/* pluginInitialize				Initialize the plugin's control and data records */
/* pluginModelSetup				Setup the plugin with the current time information and prepare to generate */
/* pluginModelGenerate			Generate model vertex & facet data */
/* pluginFinish					Shut down the plugin.  Some plugins may need to close open files, etc. */

/* A typical light flare plugin might be called from Camera with the following commands */
/* pluginInitialize				Initialize the plugin's control and data records */
/* pluginModelSetup				Setup the plugin with the time & buffer information and prepare to generate */
/* pluginFlareGenerate			Generate a lens flare into the specified buffer (may be called in multiple passes) */
/* pluginFinish					Shut down the plugin.  Some plugins may need to close open files, etc. */

/* In ElectricImage, the model and lens flare plugins are called with additional commands such as those for */
/* setting up animation frame channels and for displaying the plugin's user interface.  The other commands */
/* listed here are also called by Camera and/or ElectricImage but are optional for the plugin to support. */

#define pluginFlareType				QUADBYTE('E', 'I', 'p', 'F')				/* Flare generator plugin */
#define pluginFlareID				16000				/* Flare generator plugin default resource ID */
#define controlFlareType			0x45494346			/* Flare generator control record type */
#define controlFlareID				QUADBYTE('E', 'I', 'C', 'F')				/* Flare generator control record default resource ID */

#define pluginModelType				QUADBYTE('E', 'I', 'p', 'M')				/* Model generator plugin */
#define pluginModelID				16000				/* Model generator plugin default resource ID */
#define controlModelType			QUADBYTE('E', 'I', 'C', 'M')				/* Model generator control record type */
#define controlModelID				pluginModelID		/* Model generator control record default resource ID */


/* Country Codes */

#define countryUSA					QUADBYTE('U', 'S', 'A', '_')
#define countryFrance				QUADBYTE('F', 'R', 'N', 'C')
#define countryBritain				QUADBYTE('B', 'R', 'I', 'T')
#define countryGermany				QUADBYTE('D', 'E', 'U', 'T')
#define countryItaly				QUADBYTE('I', 'T', 'A', 'L')
#define countryNetherlands			QUADBYTE('N', 'E', 'T', 'H')
#define countryLuxembourg			QUADBYTE('B', 'L', 'U', 'X')
#define countrySweden				QUADBYTE('S', 'W', 'E', 'D')
#define countrySpain				QUADBYTE('S', 'P', 'A', 'N')
#define countryDenmark				QUADBYTE('D', 'M', 'R', 'K')
#define countryPortugal				QUADBYTE('P', 'O', 'R', 'T')
#define countryJapan				QUADBYTE('N', 'P', 'P', 'N')
#define countryOther				QUADBYTE('O', 'T', 'H', 'R')

/* Note:  More to be added! */


/* Channel data types							Description									Type			Size */

#define channelLabel				QUADBYTE('L', 'A', 'B', 'L')	/* A container label (contains other channels)	none			0 */
#define channelEnd				QUADBYTE('L', 'E', 'N', 'D')	/* The end of a container label					none			0 */
#define channelReal				QUADBYTE('R', 'E', 'A', 'L')	/* float (floating point) data					double			8 */
#define channelCRGB				QUADBYTE('C', 'R', 'G', 'B')	/* RGB 16 bit color data						RGBColor		6 */
#define channelARGB				QUADBYTE('A', 'R', 'G', 'B')	/* ARGB 16 bit color data						ARGBColor		8 */
#define channelCHSV				QUADBYTE('C', 'H', 'S', 'V')	/* HSV 16 bit color data						HSVColor		6 */
#define channelAHSV				QUADBYTE('A', 'H', 'S', 'V')	/* AHSV 16 bit color data						AHSVColor		8 */
#define channelInteger				QUADBYTE('S', 'I', 'N', 'T')	/* Signed integer count data					long			4 */
#define channelUnsigned			QUADBYTE('U', 'I', 'N', 'T')	/* Unsigned integer count data					long			4 */
#define channelBoolean				QUADBYTE('B', 'O', 'O', 'L')	/* bool flag data								bool			2 */
#define channelUnsignedReal		QUADBYTE('U', 'R', 'E', 'L')	/* Unsigned float (floating point) data			double			8 */
#define channelCoordinate			QUADBYTE('C', 'O', 'R', 'D')	/* Coordinate data (floating point)				DCoordinate	   24 */
#define channelPascalString		QUADBYTE('P', 'S', 'T', 'R')	/* String data									PStr255		  256 */


/* Channel data component types				Description	*/

#define channelPartAll				QUADBYTE('C', 'P', 'A', 'L')	/* This indicates all components of a value */
#define channelPartCoordinate_X		QUADBYTE('C', 'P', 'P', 'X')	/* This indicates the X component of a coordinate */
#define channelPartCoordinate_Y		QUADBYTE('C', 'P', 'P', 'Y')	/* This indicates the Y component of a coordinate */
#define channelPartCoordinate_Z		QUADBYTE('C', 'P', 'P', 'Z')	/* This indicates the Z component of a coordinate */
#define channelPartColor_A			QUADBYTE('C', 'P', 'C', 'A')	/* This indicates the alpha component of a color */
#define channelPartColor_R			QUADBYTE('C', 'P', 'C', 'R')	/* This indicates the red component of a color */
#define channelPartColor_G			QUADBYTE('C', 'P', 'C', 'G')	/* This indicates the green component of a color */
#define channelPartColor_B			QUADBYTE('C', 'P', 'C', 'B')	/* This indicates the blue component of a color */


/* Object data types						Description					Geometry		*/

#define objectTypeUnknown			QUADBYTE('N', 'O', 'N', 'E')	/* Unsupported type							*/
#define objectTypeCamera			QUADBYTE('C', 'A', 'M', 'A')	/* Camera type				No				*/
#define objectTypeLightRadial		QUADBYTE('R', 'L', 'T', 'E')	/* Light type (radial)			No				*/
#define objectTypeLightSpot		QUADBYTE('S', 'L', 'T', 'E')	/* Light type (spot)			No				*/
#define objectTypeLightParallel		QUADBYTE('P', 'L', 'T', 'E')	/* Light type (parallel)		No				*/
#define objectTypeLightTube		QUADBYTE('T', 'L', 'T', 'E')	/* Light type (tube)			No				*/
#define objectTypeLightCamera		QUADBYTE('C', 'L', 'T', 'E')	/* Light type (camera)			No				*/
#define objectTypeLightAmbient		QUADBYTE('A', 'L', 'T', 'E')	/* Light type (ambient)		No				*/
#define objectTypeLightArea		QUADBYTE('A', 'R', 'E', 'A')	/* Light type (area)			No				*/
#define objectTypeLight			QUADBYTE('L', 'I', 'T', 'E')	/* Light type					No				*/
#define objectTypeIlluminatorDome	QUADBYTE('I', 'L', 'M', 'D')	/* Illuminator type	(Dome)	No				*/
#define objectTypeIlluminatorBox	QUADBYTE('I', 'L', 'M', 'B')	/* Illuminator type	(Box)	No				*/
#define objectTypeGroup			QUADBYTE('G', 'R', 'U', 'P')	/* Group type					Yes				*/
#define objectTypePlugin			QUADBYTE('P', 'L', 'U', 'G')	/* Plugin type 				Yes/No			*/
#define objectTypeEffectorBox		QUADBYTE('E', 'F', 'F', 'R')	/* Effector type (box)			No				*/
#define objectTypeEffectorJoint		QUADBYTE('J', 'O', 'I', 'N')	/* Effector type (joint)		No				*/
#define objectTypeEffectorCross		QUADBYTE('C', 'R', 'S', 'S')	/* Effector type (cross)		No				*/
#define objectTypeEffectorBone		QUADBYTE('B', 'O', 'N', 'E')	/* Effector type (bone)		No				*/
#define objectTypeSmoke			QUADBYTE('S', 'M', 'O', 'K')	/* Smoker type 				No				*/
#define objectTypeMaterial			QUADBYTE('M', 'A', 'T', 'R')	/* Material type				No				*/
#define objectTypeFile				QUADBYTE('F', 'I', 'L', 'E')	/* Model File type 			No				*/
#define objectTypeWorld			QUADBYTE('W', 'O', 'R', 'L')	/* World type 				No				*/
#define objectTypeIKHandle			QUADBYTE('I', 'K', 'H', 'N')	/* IK handle type				No				*/

/* Note:  More to be added! */


/* Challenge types */

#define challengeSerial				QUADBYTE('S', 'E', 'R', 'L')	/* Returns the rightmost 7 characters of the user serial string */


/* When a plugin returns from its call, the host should automaticly end any open channel container labels.  Every */
/* channel, even LEND channels, should have a reference provided by the host.  This reference is used by the */
/* plugin to access and delete channels.  It can also be used to create a new channel immediately folowing another */
/* channel in the project list. */

/* Example (This is only an example and does not indicate what a lens flare element will necessarily contain) */

/* 	TYPE					Name			 */

/* 	LABL					Central Flare	 */
/* 		LABL				Rotation		 */
/* 			REAL			X				 */
/* 			REAL			Y				 */
/* 			REAL			Z				 */
/* 		LEND								 */
/* 		CRGB				Color			 */
/* 		REAL				Intensity		 */
/* 		LABL				Scale			 */
/* 			REAL			X				 */
/* 			REAL			Y				 */
/* 		LEND								 */
/* 	LEND									 */


/* Animation Channel key frame modes */

/* The following operations are applied to the set functions and govern */
/* how values are assigned in the current frame */

#define PLUGIN_CHANNEL_MODE_ADD_SYSTEM				1	/* Add the value to the key channel and let the system process it */
#define PLUGIN_CHANNEL_MODE_ADD_KEY					2	/* Add the value to the key channel and create keyframe if necessary */
#define PLUGIN_CHANNEL_MODE_ADD_FRAME				3	/* Add the value to the key channel and create custom frame if necessary */

/* The following operations can be applied to both get and set functions */
/* Setting and Getting the values requires the correct type */
/* Setting the values can also be accomplished using operators (@) in the pascal string format */

#define PLUGIN_CHANNEL_MODE_PROCESS_VALUE			10	/* Get or Set the value of a specified key frame */
#define PLUGIN_CHANNEL_MODE_PROCESS_TIME			11	/* Get or Set the time of a specified key frame */

/* This values are used to specify which key frames are to be changed */
/* Positive non-zero numbers is the actual key frame index (starting at one) */

#define PLUGIN_CHANNEL_SPECIFY_ALLKEY				0	/* This specifies all key frames */
#define PLUGIN_CHANNEL_SPECIFY_ALLFRAME				-1	/* This specifies all key frames */
#define PLUGIN_CHANNEL_SPECIFY_SELECTED				-2	/* This specifies all selected key frames */


/* Animation Channel Information Constants */

/* Type of value filters */

#define PLUGIN_CHANNEL_FILTER_NONE					0	/* No filter specified, all values are allowed */
#define PLUGIN_CHANNEL_FILTER_BYTE_SIGNED			1	/* Signed byte: Value is between -127 and 128 */
#define PLUGIN_CHANNEL_FILTER_BYTE_UNSIGNED			2	/* Unsigned byte: Value is between 0 and 255 */
#define PLUGIN_CHANNEL_FILTER_INTEGER_SIGNED			3	/* Signed integer: Value is signed 32 bit integer */
#define PLUGIN_CHANNEL_FILTER_INTEGER_UNSIGNED		4	/* Unsigned integer: Value is unsigned 32 bit integer */
#define PLUGIN_CHANNEL_FILTER_FLOAT_SIGNED			5	/* Signed float: Value is between -INF and +INF */
#define PLUGIN_CHANNEL_FILTER_FLOAT_UNSIGNED			6	/* Unsigned float: Value is between 0 and +INF */
#define PLUGIN_CHANNEL_FILTER_BOOLEAN				7	/* bool: Value is either on (0) or off (1) */


/* Object Create Constants */

#define PLUGIN_OBJECT_HIERARCHY_MODE_DEFAULT			0
#define PLUGIN_OBJECT_HIERARCHY_MODE_CLASSIC			1

#define PLUGIN_OBJECT_ROTATION_NONE					0
#define PLUGIN_OBJECT_ROTATION_XYZ					1
#define PLUGIN_OBJECT_ROTATION_XZY					2
#define PLUGIN_OBJECT_ROTATION_YXZ					3
#define PLUGIN_OBJECT_ROTATION_YZX					4
#define PLUGIN_OBJECT_ROTATION_ZXY					5
#define PLUGIN_OBJECT_ROTATION_ZYX					6
#define PLUGIN_OBJECT_ROTATION_XYX					7
#define PLUGIN_OBJECT_ROTATION_XZX					8
#define PLUGIN_OBJECT_ROTATION_YXY					9
#define PLUGIN_OBJECT_ROTATION_YZY					10
#define PLUGIN_OBJECT_ROTATION_ZXZ					11
#define PLUGIN_OBJECT_ROTATION_ZYZ					12

#define PLUGIN_OBJECT_HIERARCHY_INHERIT_TRANSLATION	0
#define PLUGIN_OBJECT_HIERARCHY_INHERIT_ROTATION	1
#define PLUGIN_OBJECT_HIERARCHY_INHERIT_SCALE		2
#define PLUGIN_OBJECT_HIERARCHY_INHERIT_VISIBILITY	3
#define PLUGIN_OBJECT_HIERARCHY_INHERIT_DEFORMATION	4

#define PLUGIN_OBJECT_CONSTRAINT_LIMIT_POSITION		1
#define PLUGIN_OBJECT_CONSTRAINT_LIMIT_ROTATION		2
#define PLUGIN_OBJECT_CONSTRAINT_LIMIT_STIFFNESS		3
#define PLUGIN_OBJECT_CONSTRAINT_LIMIT_VISCOSITY		4
#define PLUGIN_OBJECT_CONSTRAINT_LIMIT_GRAVITY		5

/* Object Info Flags	  */

#define OBJECTINFO_FLAG_VISIBILITY					0
#define OBJECTINFO_FLAG_ANIMATABLE					1
#define OBJECTINFO_FLAG_SELECTED					2

/* Standard host commands */

#define hostCheck				QUADBYTE('C', 'H', 'E', 'K')		/* Check to see if the host supports a particular command (return 0 or commandErr) */
#define hostCreator			QUADBYTE('C', 'R', 'E', 'A')		/* Get the host application's creator type */
#define hostVersion			QUADBYTE('V', 'E', 'R', 'S')		/* Get the host application's version info */
#define hostChallenge			QUADBYTE('C', 'H', 'A', 'L')		/* Send a software protection challenge to the host and receive a response */
#define hostFree				QUADBYTE('F', 'R', 'E', 'E')		/* Get the maximum free memory available to the plugin */
#define hostAllocate			QUADBYTE('A', 'L', 'O', 'C')		/* Set the memory allocation of the plugin and control data to the requested size */
#define hostMemMalloc			QUADBYTE('M', 'M', 'A', 'L')		/* Allocate or reallocate a memory pointer */
#define hostMemFree			QUADBYTE('M', 'F', 'R', 'E')		/* Free a memory pointer */
#define hostPermMalloc			QUADBYTE('P', 'M', 'A', 'L')		/* Allocate or reallocate a permnanant memory pointer */
#define hostPermFree			QUADBYTE('P', 'F', 'R', 'E')		/* Free a permnanant memory pointer */
#define hostStatus			QUADBYTE('S', 'T', 'A', 'T')		/* Update the status box information */
#define hostBackground			QUADBYTE('B', 'A', 'C', 'K')		/* Enable host background processing */

#define hostResetChannel		QUADBYTE('R', 'C', 'H', 'N')		/* Reset the plugin's animation channels */
#define hostInsertChannel		QUADBYTE('P', 'C', 'H', 'N')		/* Insert an animation channel to the project list */
#define hostAddChannel			QUADBYTE('A', 'C', 'H', 'N')		/* Add an animation channel to the project list */
#define hostDeleteChannel		QUADBYTE('D', 'C', 'H', 'N')		/* Delete an animation channel from the project list */
#define hostSetChannelName		QUADBYTE('C', 'C', 'H', 'N')		/* Set the animation channel name */
#define hostGetChannel			QUADBYTE('G', 'C', 'H', 'N')		/* Get a frame value for a plugin's animation channel */
#define hostSetChannel			QUADBYTE('S', 'C', 'H', 'N')		/* Set a frame value for a plugin's animation channel */

#define hostAddChannelKey		QUADBYTE('A', 'C', 'H', 'K')		/* Add key(s) to an animation channel */
#define hostDeleteChannelKey	QUADBYTE('D', 'C', 'H', 'K')		/* Delete key(s) from an animation channel */
#define hostGetChannelKey		QUADBYTE('G', 'C', 'H', 'K')		/* Get a key value for a plugin's animation channel */
#define hostSetChannelKey		QUADBYTE('S', 'C', 'H', 'K')		/* Set a key value for a plugin's animation channel */

#define hostGetChannelFrame	QUADBYTE('G', 'C', 'H', 'F')		/* Get frame data of an animation channel */
#define hostAddChannelFrame	QUADBYTE('A', 'C', 'H', 'F')		/* Add frame to an animation channel */
#define hostDeleteChannelFrame	QUADBYTE('D', 'C', 'H', 'F')		/* Delete frame from an animation channel */

#define hostGetChannelInfo		QUADBYTE('G', 'C', 'H', 'I')		/* Get the animation channel information */
#define hostGetChannelName		QUADBYTE('N', 'C', 'H', 'I')		/* Get the animation channel name */
#define hostGetTimingInfo		QUADBYTE('T', 'I', 'N', 'F')		/* Get the animation information */
#define hostSynchronize		QUADBYTE('S', 'Y', 'N', 'C')		/* Synchronize the animation information */
#define hostObjectSynchronize	QUADBYTE('O', 'S', 'Y', 'N')		/* Synchronize the object's transformation */
#define hostProjectInfoRequest	QUADBYTE('P', 'I', 'R', 'Q')		/* Request for pluginProjectInfo call */

#define hostGetNameInfo		QUADBYTE('G', 'N', 'A', 'M')		/* Get the object name */
#define hostSetNameInfo		QUADBYTE('S', 'N', 'A', 'M')		/* Set the object name */

#define hostAddObject			QUADBYTE('A', 'O', 'B', 'J')		/* Create the object */
#define hostObjectInfo			QUADBYTE('U', 'O', 'B', 'J')		/* Update the object Info */
#define hostObjectAnimatedInfo	QUADBYTE('K', 'O', 'B', 'J')		/* Update the animated object Info */
#define hostObjectTransform	QUADBYTE('T', 'O', 'B', 'J')		/* Update the object transform */


/* Standard plugin commands */

#define pluginInitialize		QUADBYTE('I', 'N', 'I', 'T')		/* Initialize the plugin - allocate memory */
#define pluginFinish			QUADBYTE('F', 'I', 'N', 'I')		/* Finish the plugin - all allocated memory is released */
#define pluginInformation		QUADBYTE('I', 'N', 'F', 'O')		/* Get information about the plugin */
#define pluginCheck			QUADBYTE('C', 'H', 'E', 'K')		/* Check to see if the plugin supports a particular command (return 0 or commandErr) */
#define pluginDescribe			QUADBYTE('D', 'E', 'S', 'C')		/* Describe a plugin result code */
#define pluginInterface		QUADBYTE('I', 'N', 'T', 'F')		/* Display the user interface for the plugin and setup parameters */
#define pluginInitializeChannel	QUADBYTE('I', 'C', 'H', 'N')		/* Initialize the plugin's animation channels */
#define pluginFinishChannel	QUADBYTE('F', 'C', 'H', 'N')		/* Finish the plugin's animation channels (plugin should nolonger reference them) */
#define pluginUpdateChannel	QUADBYTE('U', 'C', 'H', 'N')		/* Update a range of frames for each of the plugin's animation channels */
#define pluginLoadChannel		QUADBYTE('L', 'C', 'H', 'N')		/* Load plugin's animation channel information for the current frame */


/* Lens flare specific plugin commands */

#define pluginFlareSetup		QUADBYTE('S', 'E', 'T', 'F')		/* Setup for flare generation */
#define pluginFlareGenerate	QUADBYTE('G', 'E', 'N', 'F')		/* Draw all of the lens flares in the list (pointed to by firstFlare) into an image buffer */


/* Model plugin specific host commands */

#define hostModelReset			QUADBYTE('R', 'S', 'E', 'T')		/* Reset the model data after control parameters have changed */
#define hostModelConfigure		QUADBYTE('C', 'F', 'I', 'G')		/* Configure the model data by sending the total number of vertices */
#define hostModelGetCanvas		QUADBYTE('G', 'E', 'T', 'C')		/* Get information about the rendering canvas */
#define hostModelGetGroup		QUADBYTE('G', 'E', 'T', 'G')		/* Get information about a group group of the group */
#define hostModelSetGroup		QUADBYTE('C', 'P', 'Y', 'G')		/* Copy group information from a referenced group into the current group  */
#define hostModelAddGroup		QUADBYTE('A', 'D', 'D', 'G')		/* Add a new group (it becomes the current group for adding vertices & facets)  */
#define hostModelCopyGroup		QUADBYTE('D', 'U', 'P', 'G')		/* Copy a group (the copy becomes the current group for adding vertices & facets) */
#define hostModelDeleteGroup	QUADBYTE('D', 'E', 'L', 'G')		/* Marks a child group of a plugin parent group for latest deletion  */
#define hostModelGetGroupData	QUADBYTE('G', 'E', 'T', 'D')		/* Get a group's plugin data pointer and ID */
#define hostModelSetGroupData	QUADBYTE('S', 'E', 'T', 'D')		/* Set a group's plugin data pointer and ID */
#define hostModelGetWeight		QUADBYTE('G', 'E', 'T', 'W')		/* Get a weight record in a child group of the plugin parent group  */
#define hostModelGetVertex		QUADBYTE('G', 'E', 'T', 'V')		/* Get a vertex record in a child group of the plugin parent group  */
#define hostModelAddVertex		QUADBYTE('A', 'D', 'D', 'V')		/* Add a new vertex record to the current group  */
#define hostModelGetVertexArray	QUADBYTE('G', 'A', 'R', 'V')		/* Get a vertex sub list in a child group of the plugin parent group  */
#define hostModelAddVertexArray	QUADBYTE('A', 'A', 'R', 'V')		/* Add a new vertex sub list to the current group  */
#define hostModelGetFacet		QUADBYTE('G', 'E', 'T', 'F')		/* Get a facet of the group */
#define hostModelAddFacet		QUADBYTE('A', 'D', 'D', 'F')		/* Add a new facet record to the group */
#define hostModelGetFacetArray	QUADBYTE('G', 'A', 'R', 'F')		/* Get a facet sub list in a child group of the plugin parent group  */
#define hostModelAddFacetArray	QUADBYTE('A', 'A', 'R', 'F')		/* Add a new facet sub list to the current group  */
#define hostModelGINF			QUADBYTE('G', 'I', 'N', 'F')		/* Set the values of the group's GINF record */
#define hostModelGATR			QUADBYTE('G', 'A', 'T', 'R')		/* Set the values of the group's GATR record */
#define hostModelGBLR			QUADBYTE('G', 'B', 'L', 'R')		/* Set the values of the group's GBLR record */
#define hostModelTMAP			QUADBYTE('T', 'M', 'A', 'P')		/* Set the values of the group's TMAP record */
#define hostModelTATR			QUADBYTE('T', 'A', 'T', 'R')		/* Set the values of the group's TATR record */
#define hostModelBMAP			QUADBYTE('B', 'M', 'A', 'P')		/* Set the values of the group's BMAP record (secondary TMAP) */
#define hostModelBATR			QUADBYTE('B', 'A', 'T', 'R')		/* Set the values of the group's BATR record (secondary TATR) */
#define hostModelRMAP			QUADBYTE('R', 'M', 'A', 'P')		/* Set the values of the group's RMAP record */
#define hostModelGLNK			QUADBYTE('G', 'L', 'N', 'K')		/* Set the values of the group's GLNK record */	// BMB 022305
#define hostModelGroupLimits	QUADBYTE('G', 'L', 'M', 'T')        /* Get the values of the group's limit record */


/* Model specific plugin commands  */

#define pluginModelSetup		QUADBYTE('S', 'E', 'T', 'M')		/* Setup for model generation */
#define pluginModelGenerate	QUADBYTE('G', 'E', 'N', 'M')		/* Generate the group's model data */
#define pluginProjectInfo		QUADBYTE('P', 'I', 'N', 'F')		/* Receive the project's information */


/* The following model specific plugin commands are obsolete. */
/* They were sent after the pluginModelSetup command and before the */
/* pluginModelGenerate command.  They should now be ignored */

#define pluginModelGINF			QUADBYTE('G', 'I', 'N', 'F')		/* Set the values of the group's GINF record */
#define pluginModelGATR			QUADBYTE('G', 'A', 'T', 'R')		/* Set the values of the group's GATR record */
#define pluginModelGBLR			QUADBYTE('G', 'B', 'L', 'R')		/* Set the values of the group's GBLR record */
#define pluginModelTMAP			QUADBYTE('T', 'M', 'A', 'P')		/* Set the values of the group's TMAP record */
#define pluginModelTATR			QUADBYTE('T', 'A', 'T', 'R')		/* Set the values of the group's TATR record */
#define pluginModelBMAP			QUADBYTE('B', 'M', 'A', 'P')		/* Set the values of the group's BMAP record (secondary TMAP) */
#define pluginModelBATR			QUADBYTE('B', 'A', 'T', 'R')		/* Set the values of the group's BATR record (secondary TATR) */
#define pluginModelRMAP			QUADBYTE('R', 'M', 'A', 'P')		/* Set the values of the group's RMAP record */


/* Standard result codes */

#define pluginNoErr				0			/* Zero result means no error */
#define pluginMemoryErr			1			/* Cannot allocate enough memory */
#define pluginCommandErr		2			/* Unknown command received */
#define pluginParameterErr		3			/* Error discovered in passed parameter record */
#define pluginCancelErr			4			/* The user canceled or aborted the operation */
#define pluginSequenceErr		5			/* The command was received in the wrong sequence */
#define pluginResourceErr		6			/* A required resource could not be found */
#define pluginProtectionErr		7			/* The copy protection could not be verified*/
#define pluginFeatureErr		8			/* A requested feature or mode is not supported */


/* Plugin copy permission flags from the PluginInformationRec record */

#define pluginPermissionCopyFlag 			31	/* When this bit is set, the host will not copy the plugin to another system */
#define pluginPermissionModeless			30	/* When this bit is set, the host will treat the plugin interface as pseudo modeless */
#define pluginProjectInfoFlag				29	/* When this bit is set, the plug-in receives pluginProjectInfo host commands */
#define pluginProjectSelectionFlag			28	/* When this bit is set, the plug-in receives pluginProjectInfo with projectInfoMode_Selection */
#define pluginSensitivityFlag				0	/* Set if the plug-in supports the sensitivity flags */
#define pluginTimeSensitivityFlag			1	/* Set if the plug-in is sensitive to changes in time */
#define pluginParentSensitivityFlag			2	/* Set if the plug-in is sensitive to changes to its own group */
#define pluginChildSensitivityFlag			3	/* Set if the plug-in is sensitive to changes to its child groups */
#define pluginSetupSensitivityFlag			4	/* Set if the plug-in is sensitive to changes in the reset command (allows the modelsetup call to be repeated) */


/* Group vertex mapping flags definitions */

#define pluginVertexColorFlag	31			/* Set when a group has colors stored at each vertex */
#define pluginVertexNormalFlag	30			/* Set when a group has normals stored at each vertex */
#define pluginVertexTextureFlag	29			/* Set when a group has texture positions stored at each vertex */
#define pluginVertexBlurFlag	28			/* Set when a group has motion blur positions stored at each vertex */


/* Group limits flags definitions */

#define linkRotationLimitFlag    0      	/* Set when a group's rotation limits enabled */
#define linkPositionLimitFlag    1          /* Set when a group's position limits enabled */
#define linkStiffnessFlag        2          /* Set when a group's stiffness enabled */
#define linkViscosityFlag 		 3          /* Set when a group's viscosity enabled */

#ifdef __MACINTOSH__
#if __LP64__
	#pragma pack(4)
#else
	#pragma options align=mac68k
#endif
#endif

/* Prototype for plugin routines */

typedef SInt32 (*EIPluginRoutine)(IFFType theCommand, SInt32 theCommandRecSize, void* theCommandRec);

/* Plugin information record */

typedef struct {
	char		*theControlData;						/* <==  Sends the control parameter data */
	UInt32		theControlDataSize;						/* <==  Sends the control parameter data size */
	char		*thePluginData;							/* <==  Sends the plugin private data */
	UInt32		thePluginDataSize;						/* <==  Sends the plugin privete data size */
	EIPluginRoutine theHostProc;						/* <==  Sends the host callback procedure */
	char *theHostReference;								/* <==  Sends the private data used by host callbacks */
} PluginInfoRec, *PluginInfoPtr;

/* Challenge data block */

typedef union {
	unsigned char b[8];
	struct {
		unsigned char b0, b1, b2, b3, b4, b5, b6, b7;
	} b0;
	struct {
		short i0, i1, i2, i3;
	} i;
	unsigned char s[8];	/* Pascal 7 character string (length + seven characters) */
	struct {
		SInt32 l0, l1;
	} l;
} ChallengeDataRec, *ChallengeDataPtr;

/* Standard host types */

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	IFFType theCommand;									/* <==  Sends the command to check for host acceptance */
} HostCheckRec, *HostCheckPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 versionRelease;								/*  ==> Returns the release version of the host application */
	SInt32 versionRevision;								/*  ==> Returns the revision version of the host application */
	SInt32 versionFix;									/*  ==> Returns the fix version of the host application */
	SInt32 versionInternal;								/*  ==> Returns the internal version of the host application */
	PStr31 versionString;								/*  ==> Returns the short string version of the host application */
	IFFType versionCountry;								/*  ==> Returns the country code of the host application */
} HostVersionRec, *HostVersionPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	IFFType theCreator;									/*  ==> Returns the creator of the host application */
} HostCreatorRec, *HostCreatorPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	IFFType theChallenge;								/* <==  Sends the type of challenge */
	ChallengeDataRec theChallengeData;					/* <==> Sends and returns the challenge data buffer */
} HostChallengeRec, *HostChallengePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	ulong theFreeSize;									/*  ==> Returns the free memory size available to the plugin */
} HostFreeRec, *HostFreePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theControlData;								/*  ==> Returns the new control parameter data location */
	SInt32 theControlDataSize;							/* <==  Sends the requested new control parameter data size */
	char *thePluginData;									/*  ==> Returns the new plugin private data location */
	SInt32 thePluginDataSize;							/* <==  Sends the requested new size of the plugin private data */
} HostAllocateRec, *HostAllocatePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	ulong theSize;										/* <==  Sends the requested memory size */
	void *thePtr;										/* <==> Sends and returns the pointer to the allocated memory */
} HostMemMallocRec, *HostMemMallocPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	void *thePtr;										/* <==  Sends the pointer to the allocated memory */
} HostMemFreeRec, *HostMemFreePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	ulong theSize;										/* <==  Sends the requested memory size */
	void *thePtr;										/* <==> Sends and returns the pointer to the allocated memory */
} HostPermMallocRec, *HostPermMallocPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	void *thePtr;										/* <==  Sends the pointer to the allocated memory */
} HostPermFreeRec, *HostPermFreePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	float theStatusPos;									/* <==  Sends the percentage of the job which has been completed (0.0 ... 1.0) */
	PStr255 theStatusUnit;								/* <==  Sends the unit name for the status box */
	PStr255 theStatusInfo;								/* <==  Sends the information string for the status box */
	SInt32 theStatusMin;								/* <==  Sends the minimum value of the status box range */
	SInt32 theStatusMax;								/* <==  Sends the maximum value of the status box range */
	SInt32 theStatusValue;								/* <==  Sends the current value of the status box */
	SInt32 theStatusTick;								/* <==> Send zero to change the status box, returns the next status update tick */
} HostStatusRec, *HostStatusPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theEnable;									/* <==	Sends the enable for background processing: 0 - disable, 1 - enable */
														/* NOTE: All calls with theEnable set to 0 must be balanced with calls where */
														/* theEnable is set to 1. */
} HostBackgroundRec, *HostBackgroundPtr;


/* Standard host channel types */

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
} HostResetChannelRec, *HostResetChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the new channel  */
	IFFType theChannelType;								/* <==  Sends the type of the new channel */
	PStr255 theChannelName;								/* <==  Sends the name of the new channel */
	char *theChannelDefault;								/* <==  Sends a pointer to the default value(s) for the new channel */
	char *theChannelMinimum;								/* <==  Sends a pointer to the minimum possible value(s) for the new channel (nil=none) */
	char *theChannelMaximum;								/* <==  Sends a pointer to the maximum possible value(s) for the new channel (nil=none) */
	char *theChannelBefore;								/* <==  Sends a reference to the channel before the new channel (nil=add to end of list) */
	char *theChannelReference;							/*  ==> Returns a reference to the new channel */
} HostAddChannelRec, *HostAddChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the new channel  */
	IFFType theChannelType;								/* <==  Sends the type of the new channel */
	PStr255 theChannelName;								/* <==  Sends the name of the new channel */
	char *theChannelDefault;								/* <==  Sends a pointer to the default value(s) for the new channel */
	char *theChannelMinimum;								/* <==  Sends a pointer to the minimum possible value(s) for the new channel (nil=none) */
	char *theChannelMaximum;								/* <==  Sends a pointer to the maximum possible value(s) for the new channel (nil=none) */
	char *theChannelBefore;								/* <==  Sends a reference to the sibling channel before the new channel (nil = add to end of siblings) */
	char *theChannelParent;								/* <==  Sends a reference to the parent channel */
	char *theChannelReference;							/*  ==> Returns a reference to the new channel */
} HostInsertChannelRec, *HostInsertChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelReference;							/* <==  Sends a reference to the channel to be deleted */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel (for validation) */
} HostDeleteChannelRec, *HostDeleteChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel (for validation) */
	PStr255 theChannelName;								/* <==  Sends the new name of the channel */
} HostSetChannelNameRec, *HostSetChannelNamePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Returns the unique ID number of the channel (for validation) */
	IFFType theChannelType;								/*  ==> Returns the type of the requested channel */
	SInt32 theChannelFrame;								/* <==  Sends the requested channel frame number */
	char *theChannelData;								/*  ==> Returns a pointer to the requested channel data */
} HostGetChannelRec, *HostGetChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */
	IFFType theChannelType;								/* <==  Sends the type of the requested channel (for validation) */
	SInt32 theChannelFrame;								/* <==  Sends the requested frame number */
	char *theChannelData;								/* <==  Sends a pointer to the data to be stored in the requested channel frame */
} HostSetChannelRec, *HostSetChannelPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */

	IFFType theDataType;									/* <==  Sends the type of the specified data */
	char *theData;										/* <==  Sends a pointer to the data to be stored in the created channel key */
	double theKeyTime;									/* <==  Sends the time at which the key is to be created */
} HostAddChannelKeyRec, *HostAddChannelKeyPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */

	SInt32 theKeyIndex;									/* <==  Sends the type or index of key to be deleted (see key specifier constants) */					
} HostDelChannelKeyRec, *HostDelChannelKeyPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Returns the unique ID number of the channel (for validation) */

	IFFType theDataType;								/*  ==> Returns the type of the requested channel */
	char *theData;										/*  ==> Returns a pointer to the requested channel data */
	SInt32 theFrameIndex;								/* <==  Sends the requested time frame number */
} HostGetChannelFrameRec, *HostGetChannelFramePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */

	IFFType theDataType;								/* <==  Sends the type of the specified data */
	char *theData;										/* <==  Sends a pointer to the data to be stored in the created channel key */
	SInt32 theFrameIndex;								/* <==  Sends the time at which the key is to be created */
} HostAddChannelFrameRec, *HostAddChannelFramePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */

	SInt32 theFrameIndex;								/* <==  Sends the type or index of key to be deleted (see key specifier constants) */					
} HostDelChannelFrameRec, *HostDelChannelFramePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Returns the unique ID number of the channel (for validation) */

	IFFType theDataType;								/*  ==> Returns the type of the requested channel */
	char *theData;										/*  ==> Returns a pointer to the requested channel data */
	SInt32 theKeyPart;									/* <==  Sends the part of the key to be performed on */
	SInt32 theKeyIndex;									/* <==  Sends the index of the key (only positive values allowed) */
} HostGetChannelKeyRec, *HostGetChannelKeyPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel */

	IFFType theDataType;								/* <==  Sends the type of the requested data */
	char *theData;										/* <==  Sends a pointer to the data to be stored in the requested channel key */
	SInt32 theKeyPart;									/* <==  Sends the part of the key to be performed on */
	SInt32 theKeySubPart;								/* <==  Sends the sub-part of a key value (x, y, z, a, r, g, b, etc.) */
	SInt32 theKeyIndex;									/* <==  Sends the index of the key (see key specifier constants) */
} HostSetChannelKeyRec, *HostSetChannelKeyPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel (for validation) */

	SInt32 theChannelType;								/*  ==> Returns the type of the channel value */
	SInt32 theChannelFilter;							/*  ==> Returns the filter of the channel value */
	SInt32 theChannelVelocity;							/*  ==> Returns the availability of the channel value velocity */
	SInt32 theChannelKeyTotal;							/*  ==> Returns the total number of keys (there is always at least one) */
	SInt32 theChannelFlag;								/*  ==> Returns a flag set of the channel (not used currently) */
	char *theChannelMinimum;								/*  ==> Returns a pointer to the minimum possible value(s) for the new channel (nil=none) */
	char *theChannelMaximum;								/*  ==> Returns a pointer to the maximum possible value(s) for the new channel (nil=none) */
} HostGetChannelInfoRec, *HostGetChannelInfoPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theChannelOwner;								/* <==  Sends the owner pointer to the host, nil if the owner is the plugin */
	char *theChannelReference;							/* <==  Sends a reference to the requested channel */
	SInt32 theChannelID;								/* <==  Sends the unique ID number of the channel (for validation) */

	PStr255 theChannelName;								/*  ==> Returns the name of the new channel */
	SInt32 theChannelNameParent;						/*  ==> Returns the parent channel ID of the specified channel (0 if root level) */
	SInt32 theChannelNameFlag;							/*  ==> Returns a flag set of the channel name (bit 0 - true if name is variable) */
} HostGetChannelNameRec, *HostGetChannelNamePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	double theTimingStart;								/*  ==> Returns the system start time (default is 0.0) */
	double TheTimingCurrent;								/*  ==> Returns the current frame time */
	double theTimingFPS;									/*  ==> Returns the system Frames-Per-Second (default is 30) */
	SInt32 theTimingInterlace;							/*  ==> Returns the availability of interlace (default is 0) */
	SInt32 theTimingTotal;								/*  ==> Returns the total number of frames */
	SInt32 theTimingSkip;								/*  ==> Returns the number of frames to skip (always at least zero) */
	SInt32 theRenderFrameStart;							/*  ==> Returns the start frame of rendering (default is 0) */
	SInt32 theRenderFrameStop;							/*  ==> Returns the stop frame of the rendering (default is theTimingTotal) */
} HostGetTimingInfoRec, *HostGetTimingInfoPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theTimingType;								/* <==	Sends the type of synchronization (0 - current time, 1 - by frame, 2 - by time) */
	SInt32 theTimingFrame;								/* <==  Sends the frame to which the groups are synchronized */
	double theTimingTime;								/* <==	Sends the time to which the groups are synchronized */
} HostSynchronizeRec, *HostSynchronizePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theFrameIndex;								/* <==  Sends the current frame index number */
	double theFrameTime;								/* <==  Sends the time of the current frame */
	double theFrameTimeDelta;							/* <==  Sends the time per frame (1/fps) */
	char  *theObject;									/* <==  Sends the object to synchronize */
} HostObjectSynchronizeRec, *HostObjectSynchronizePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
} HostProjectInfoRequestRec, *HostProjectInfoRequestPtr;

/* Standard host object types */

typedef struct {
	char	*theReference;								/* <==  Sends the private data used by host callbacks  */
	char	*theNameOwner;								/*  ==> Returns the object pointer of the requested Object  */
	PStr255	theName;								/* <==> Get/Set the name of the object */
} HostObjectNameRec, *HostObjectNamePtr;

typedef struct {
	char *theReference;								/* <==  Sends the private data used by host callbacks */

	char *theObject;									/*  ==> Returns the object pointer of the newly created object */
	char *theObjectParent;							/* <==  Sends the parent of the object about to be created (must be specified - default is the plugin group) */
	char *theObjectSibling;							/* <==  Sends the sibling of the object about to be created (must be specified - if not siblings use the parent) */

	SInt32	theObjectType;							/*  ==> Sends the object type (camera, light, etc.) to be created (no groups please) */
	SInt32	theObjectFlag;							/* <==  Sends the inheritance flags (see FACTStuff) */
	SInt32	theObjectHierarchy;						/* <==  Sends the hierarchy method (see Link window) */
	PStr255	theObjectName;							/* <==  Sends the name of the object. */
	DCoordinate theObjectWorldSize;					/* <==  Sends the display size of the object in world space */

	SInt32	theObjectRotationOrder;					/* <==  Sends the rotation order for the rotation transformation */

	DCoordinate theObjectLocalOffset;				/* <==  Sends the local coordinate system origin (see linkage window) */
	DCoordinate theObjectLocalRotation;				/* <==  Sends the local coordinate system orientation (see linkage window) */
	DCoordinate theObjectParentScale;				/* <==  Sends the starting object scale */
	DCoordinate theObjectParentOffset;				/* <==  Sends the starting object offset */
	DCoordinate theObjectParentRotation;			/* <==  Sends the starting object rotation */

	SInt32	theConstraintType;						/* <==  Sends the constraint type (see FACTStuff) */
	SInt32	theConstraintFlag;						/* <==  Sends the constraint flags (see FACTStuff) */
	
	DCoordinate	theConstraintLimitMinRotation;			/* <==  Sends the constraint minimum rotation (see linkage window) */
	DCoordinate theConstraintLimitMaxRotation;			/* <==  Sends the constraint maximum rotation (see linkage window) */
	DCoordinate theConstraintLimitCenterRotation;		/* <==  Sends the constraint center rotation (see linkage window) */
	DCoordinate theConstraintLimitStiffness;			/* <==  Sends the constraint rotation stiffness (see linkage window) */
	DCoordinate theConstraintLimitViscosity;			/* <==  Sends the constraint rotation viscosity (see linkage window) */

	DCoordinate theConstraintLimitMinPosition;			/* <==  Sends the constraint minimum position (see linkage window) */
	DCoordinate theConstraintLimitMaxPosition;			/* <==  Sends the constraint maximum position (see linkage window) */

	DCoordinate theConstraintGravity;					/* <==  Sends the constraint local gravity (see linkage window) */
	
} HostObjectCreateRec, *HostObjectCreatePtr;

typedef struct {
	char	*theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32	theReferenceIndex;							/* <==  Sends the index of the object in list (1 based - 0 is total) */

	char	*theObject;
	SInt32	theObjectType;								/*  ==> Returns the object type (camera, light, etc.) */
	PStr255	theObjectName;								/*  ==> Returns the object name */
	SInt32	theObjectIndex;								/*  ==> Returns the object's index or total */
	char	*	theObjectParent;							/*  ==> Returns the object's parent */
	SInt32	theObjectFlag;								/*  ==> Returns the object's flags (see "Object Info Flags")  */
} HostObjectInfoRec, *HostObjectInfoPtr;

typedef struct {
	char	*theReference;								/* <==  Sends the private data used by host callbacks  */
	char	*theTransformOwner;							/* <==  Sends the object ID of the requested Object  */

	Matrix4	theTransform;								/*  ==> Returns the local to global transformation of the object */
} HostObjectTransformRec, *HostObjectTransformPtr;



/* Standard plugin types */

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
} PluginInitializeRec, *PluginInitializePtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
} PluginFinishRec, *PluginFinishPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	IFFType thePluginType;								/*  ==> Returns the plugin's unique type */
	SInt32 thePluginVersion;							/*  ==> Returns the plugin's version number */
	PStr255 thePluginName;								/*  ==> Returns the plugin's original name */
	PStr255 thePluginCopyright;							/*  ==> Returns the plugin's copyright notice */
	PStr255 thePluginAuthor;							/*  ==> Returns the plugin's author name | address | phone */
	PStr255 thePluginDescription;						/*  ==> Returns the plugin's description string */
	UInt32 thePluginPermission;							/*  ==> Returns the plugin's permission flag field (defined above) */
} PluginInformationRec, *PluginInformationPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	IFFType theCommand;									/* <==  Sends the command to check */
} PluginCheckRec, *PluginCheckPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theResult;									/* <==  Sends the result code */
	PStr255 theDescription;								/*  ==> Returns the description of the error result code */
} PluginDescribeRec, *PluginDescribePtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
} PluginInterfaceRec, *PluginInterfacePtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theStartFrame;								/* <==  The starting frame number for the animation */
	SInt32 theEndFrame;									/* <==  The ending frame number for the animation */
	double theStartTime;									/* <==  The time of the starting frame for the animation */
	double theFrameTime;									/* <==  The time per frame for the animation */
} PluginInitializeChannelRec, *PluginInitializeChannelPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
} PluginFinishChannelRec, *PluginFinishChannelPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theStartFrame;								/* <==  The starting frame number for the animation */
	SInt32 theEndFrame;									/* <==  The ending frame number for the animation */
	double theStartTime;									/* <==  The time of the starting frame for the animation */
	double theFrameTime;									/* <==  The time per frame for the animation */
} PluginUpdateChannelRec, *PluginUpdateChannelPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theStartFrame;								/* <==  The starting frame number for the animation */
	SInt32 theEndFrame;									/* <==  The ending frame number for the animation */
	double theStartTime;									/* <==  The time of the starting frame for the animation */
	double theFrameTime;									/* <==  The time per frame for the animation */
	SInt32 theFrame;									/* <==  The current frame number */
} PluginLoadChannelRec, *PluginLoadChannelPtr;

/* Lens flare specific types */

typedef struct {
	unsigned short red, green, blue;						/* 0..16384 */
} FlareColor, *FlareColorPtr;

typedef struct {											/* The flare buffer contains pixels of type FlareColor (0..16384) */
	char *bufferBase;									/* The buffer's base address */
	SInt32 bufferRowBytes;								/* The number of bytes in a single scanline of the buffer */
	L_Rect bufferRect;									/* The enclosing rectangle of the buffer */
	short bufferDepth;									/* The bits per pixel of the flare buffer (8 or 32) (8 or 48 pre-2.8)  */
	SInt32 bufferIndex;									/* The index color of flare rings for 8 bit wire frame buffers  */
} FlareBuffer, *FlareBufferPtr;

/* Lens flare plugin records */

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theFrameIndex;								/* <==  Sends the current frame index number */
	double theFrameTime;									/* <==  Sends the time of the current frame */
	double theFrameTimeDelta;							/* <==  Sends the time per frame (1/fps) */
	L_Rect theFrameRect;								/* <==  Sends the canvas frame rectangle */
	double theFrameXScale;								/* <==  Sends flareGlobalXScale */
	double theFrameYScale;								/* <==  Sends flareGlobalYScale */
	double theFrameXCenter;								/* <==  Sends flareGlobalXCenter */
	double theFrameYCenter;								/* <==  Sends flareGlobalYCenter */
	double	theBlurShutter;								/* <==  Sends the motion blur shutter angle (usually between 0 and 1) */
	Matrix4 theCameraMatrix;							/* <==  Sends the camera matrix */
	Matrix4 theCameraRotMatrix;							/* <==  Sends the camera rotation matrix */
	Matrix4 theCameraBlurMatrix;						/* <==  Sends the camera motion blur matrix */
	DCoordinate theLightWorldPosition;					/* <==  Sends the light world space poisition */
	DCoordinate theLightBlurWorldPosition;				/* <==  Sends the light motion blur world space position */
	DCoordinate theLightFramePosition;					/* <==  Sends the light perspective space position */
	DCoordinate theLightBlurFramePosition;				/* <==  Sends the light motion blur perspective space position */
	DCoordinate theLightDirection;						/* <==  Sends the light direction normal */
	DCoordinate theLightBlurDirection;					/* <==  Sends the light motion blur direction normal */
	ARGB theLightColor;									/* <==  Sends the light color */
	double theLightIntensity;							/* <==  Sends the light intensity */
	double theLightDropoff;								/* <==  Sends the light dropoff distance */
	double theLightWorldRadius;							/* <==  Sends the light radius in world units */
	double theLightFrameRadius;							/* <==  Sends the light radius in frame pixels */
	double theLightOuterAngle;							/* <==  Sends the spotlight outer cone angle */
	double theLightInnerAngle;							/* <==  Sends the spotlight inner cone angle */
	double theLightSoftExponent;						/* <==  Sends the spotlight transition expoinent */

	/* New parameters since EIAS 2.7 */
	
	double theCameraView;								/* <==  Sends the camera perspective (1.0 for perspective, 0.0 for isometric) */
	double theCameraZoom;								/* <==  Sends the camera zoom factor */
	double theCameraXScale;								/* <==  Sends the camera x pixel aspect scale */
	double theCameraYScale;								/* <==  Sends the camera y pixel aspect scale */
	double theCameraXOffset;							/* <==  Sends the camera x image offset */
	double theCameraYOffset;							/* <==  Sends the camera y image offset */
	Matrix4 theLightMatrix;								/* <==  Sends the light matrix */
	Matrix4 theLightRotMatrix;							/* <==  Sends the light rotation matrix  */
	Matrix4 theLightBlurMatrix;							/* <==  Sends the light motion blur matrix */
} PluginFlareSetupRec, *PluginFlareSetupPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	FlareBuffer theFlareBuffer;							/* <==  Sends the flare pixel buffer information */
} PluginFlareGenerateRec, *PluginFlareGeneratePtr;

/* Model generator host records */

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks  */
} HostModelResetRec, *HostModelResetPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks  */
	SInt32 totalCoordinates;							/* <==  Sends the group's total coordinates and normals */
	unsigned char includeColors;							/* <==  Sends the group's colors if any (1 for true, 0 for false) */
	unsigned char includeBlurCoordinates;					/* <==  Sends the group's blur coordinates if any (1 for true, 0 for false) */
	unsigned char includeTextures;						/* <==  Sends the group's texture coordinates if any (1 for true, 0 for false) */

} HostModelConfigureRec, *HostModelConfigurePtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks  */
	UInt32 thePerspectiveFlag;							/* <==  Sends 1 for perspective rendering, 0 for isometric rendering */
	L_Rect theCanvasRect;								/* <==  Sends the canvas frame rectangle */
	float thePixelWidth;									/* <==  Sends the antialiasing pixel width */
	float thePixelScale;									/* <==  Sends the canvas zoom factor */
	float thePixelRatio;									/* <==  Sends the canvas y/x pixel ratio */
	float thePixelXOffset;								/* <==  Sends the canvas x offset */
	float thePixelYOffset;								/* <==  Sends the canvas y offset */
}  HostModelGetCanvasRec, *HostModelGetCanvasPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks  */
	SInt32 theGroupIndex;								/* <==  Sends the index number of the requested child group (0 gets the parent) */
	char *theGroup;										/*  ==> Returns the group pointer of the requested child group  */
	UInt32 theGroupVertexFlags;							/*  ==> Returns vertex mapping flags of the requested group (defined above) */
	SInt32 theGroupVertices;							/*  ==> Returns the number of vertices of the requested child gruop  */
	SInt32 theGroupFacets;                    			/*  ==> Returns the number of facets of the requested child group */
	Matrix4 theGroupMatrix;          					/*  ==> Returns the group's transformation matrix */
	Matrix4 theGroupRotMatrix;   						/*  ==> Returns the group's rotation matrix */
	Matrix4 theGroupBlurMatrix;  						/*  ==> Returns the group's motion blur transformation */
} HostModelGetGroupRec, *HostModelGetGroupPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroupTemplate;								/* <==  Sends the template group pointer (nil for plug-in parent group) */
	UInt32 theGroupVertexFlags;							/* <==  Sends vertex mapping flags of the new group (defined above) */
} HostModelSetGroupRec, *HostModelSetGroupPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroupTemplate;								/* <==  Sends the template group pointer (nil for plug-in parent group) */
	UInt32 theGroupVertexFlags;							/* <==  Sends vertex mapping flags of the new group (defined above) */
	char *theGroup;										/* ==>  Returns the new group pointer */
} HostModelAddGroupRec, *HostModelAddGroupPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks  */
	char *theGroupTemplate;								/* <==  Sends the template group pointer (nil for plug-in parent group) */
	char *theGroup;										/* ==>  Returns the new group pointer */
	Matrix4 theGroupMatrix;          					/* <==  Sends the group's transformation matrix */
	Matrix4 theGroupBlurMatrix;  						/* <==  Sends the group's motion blur transformation */
} HostModelCopyGroupRec, *HostModelCopyGroupPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer */
} HostModelDeleteGroupRec, *HostModelDeleteGroupPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer */
	char *thePluginData;									/* ==>  Returns the group's plugin data pointer */
	SInt32 thePluginID;									/* ==>  Returns the group's plugin ID */
	UInt32 thePluginSize;								/* ==>  Returns the group's plugin data size */
} HostModelGetGroupDataRec, *HostModelGetGroupDataPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer */
	char *thePluginData;									/* <==  Sends the group's plugin data pointer */
	UInt32 thePluginID;									/* <==  Sends the group's plugin ID */
} HostModelSetGroupDataRec, *HostModelSetGroupDataPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks*/
	char *theGroup;										/* <==  Sends the requested vertex's group pointer*/
	SInt32 theVertexIndex;								/* <==  Sends the index number of the requested vertex */
	ARGB theColor;										/*  ==> Returns the color of the requested vertex */
	DCoordinate theNormal;								/*  ==> Returns the normal of the requested vertex */
	DCoordinate thePosition;							/*  ==> Returns the position of the requested vertex*/
	DCoordinate theBlurPosition;						/*  ==> Returns the motion blur position of the requested vertex */
	DCoordinate theTexturePosition;						/*  ==> Returns the texture position of the requested vertex */
} HostModelGetVertexRec, *HostModelGetVertexPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	DCoordinate	thePosition;							/* <==  Sends the vertex position */
	ARGB theColor;										/* <==  Sends the vertex color */
	DCoordinate theNormal;								/* <==  Sends the vertex normal */
	DCoordinate theBlurPosition;						/* <==  Sends the vertex motion blur position */
	DCoordinate theTexturePosition;						/* <==  Sends the vertex texture position */
	SInt32 theIndex;									/*  ==> Returns the index of the coordinate (zero if an error occurs) */
} HostModelAddVertexRec, *HostModelAddVertexPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks*/
	char *theGroup;										/* <==  Sends the requested vertex's group pointer*/
	SInt32 theVertexIndex;								/* <==  Sends the index offset of the requested vertex data */
	SInt32 theVertexTotal;								/* <==> Sends the total of the requested vertex data (contains actual total on return) */
	ARGB *theColor;										/* <==  Sends the color array to put values into */
	DCoordinate *theNormal;								/* <==  Sends the normal array to put values into */
	DCoordinate *thePosition;							/* <==  Sends the position array to put values into */
	DCoordinate *theBlurPosition;						/* <==  Sends the motion blur position array to put values into */
	DCoordinate *theTexturePosition;					/* <==  Sends the texture position array to put values into */
} HostModelGetVertexArrayRec, *HostModelGetVertexArrayPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theVertexTotal;								/* <==  Sends the total of the requested vertex data to copy from */
	DCoordinate	*thePosition;							/* <==  Sends the vertex position array to copy data from */
	ARGB *theColor;										/* <==  Sends the vertex color array to copy data from */
	DCoordinate *theNormal;								/* <==  Sends the vertex normal array to copy data from */
	DCoordinate *theBlurPosition;						/* <==  Sends the vertex motion blur position array to copy data from */
	DCoordinate *theTexturePosition;					/* <==  Sends the vertex texture position array to copy data from */
	SInt32 theIndex;									/*  ==> Returns the offset of the coordinate list (zero if an error occurs) */
} HostModelAddVertexArrayRec, *HostModelAddVertexArrayPtr;

typedef	UInt32	Vertices[4];								/* Vertex array type */

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the requested facet's group pointer  */
	SInt32 theFacetIndex;								/* <==  Sends the index number of the requested facet */
	ARGB theColor;										/*  ==> Returns the facet color  */
	short theNumVertices;								/*  ==> Returns the number of vertices in the requested facet (1..4) */
	Vertices theVertices;								/*  ==> Returns the facet vertices of the requested facet (0 padded) */
} HostModelGetFacetRec, *HostModelGetFacetPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	ARGB theColor;										/* <==  Sends the facet color */
	short theNumVertices;								/* <==  Sends the number of vertices in the new facet (1..4) */
	Vertices theVertices;								/* <==  Sends the facet vertices of the new facet (0 padded) */
} HostModelAddFacetRec, *HostModelAddFacetPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks*/
	char *theGroup;										/* <==  Sends the requested facet's group pointer*/
	SInt32 theFacetIndex;								/* <==  Sends the index offset of the requested facet data */
	SInt32 theFacetTotal;								/* <==> Sends the total of the requested facet data (contains actual total on return) */
	ARGB *theColor;										/* <==  Sends the color array to put values into */
	short *theNumVertices;								/* <==  Sends the array of vertices number to put values into */
	Vertices *theVertices;								/* <==  Sends the array of vertices to put values into */
} HostModelGetFacetArrayRec, *HostModelGetFacetArrayPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	SInt32 theFacetTotal;								/* <==  Sends the total of the requested facet data to copy from */
	ARGB *theColor;										/* <==  Sends the facet color array to copy data from */
	short *theNumVertices;								/* <==  Sends the array of vertices number to copy data from */
	Vertices *theVertices;								/* <==  Sends the array of vertices to copy data from */
	SInt32 theIndex;									/*  ==> Returns the offset of the facets list (zero if an error occurs) */
} HostModelAddFacetArrayRec, *HostModelAddFacetArrayPtr;

typedef struct {
	char		*theReference;								/* <==  Sends the private data used by host callbacks*/
	char		*theGroup;									/* <==  Sends the requested weight's group pointer*/
	SInt32	theWeightMapIndex;							/* <=>  Sen 0 to return the total number of weight maps or the 1-based index of the requested weight data */
	PStr255	theWeightMapName;							/* ==>  Returns the name of the weight map. */
	unsigned char*	theWeightMapData;					/* ==>	Returns the data of the weight map as normalized unsinged byte (0 is 0.0, and 255 is 1.0) */
} HostModelGetWeightRec, *HostModelGetWeightPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	GINFPtr theGINF;									/*  ==> Returns the pointer to the group information record */
} HostModelGINFRec, *HostModelGINFPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	GATRPtr theGATR;									/*  ==> Returns the pointer to the group attributes record */
} HostModelGATRRec, *HostModelGATRPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	GBLRPtr theGBLR;									/*  ==> Returns the pointer to the group motion blur record */
} HostModelGBLRRec, *HostModelGBLRPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	TMAPPtr theTMAP;									/*  ==> Returns the pointer to the group texture map record */
} HostModelTMAPRec, *HostModelTMAPPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	TATRPtr theTATR;									/*  ==> Returns the pointer to the group texture attribute record */
} HostModelTATRRec, *HostModelTATRPtr; 

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	BMAPPtr theBMAP;									/*  ==> Returns the pointer to the group secondary texture map record */
} HostModelBMAPRec, *HostModelBMAPPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	BATRPtr theBATR;									/*  ==> Returns the pointer to the group secondary texture attribute record */
} HostModelBATRRec, *HostModelBATRPtr;

typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	RMAPPtr theRMAP;									/*  ==> Returns the pointer to the group reflection map record */
} HostModelRMAPRec, *HostModelRMAPPtr;

// BMB 022305
typedef struct {
	char *theReference;									/* <==  Sends the private data used by host callbacks */
	char *theGroup;										/* <==  Sends the group pointer of the requested record's group  */
	GLNKPtr theGLNK;									/*  ==> Returns the pointer to the group link record */
} HostModelGLNKRec, *HostModelGLNKPtr;

typedef struct {
	char *theReference;                                 /* <==  Sends the private data used by host callbacks */
	char *theGroup;                                     /* <==  Sends the group pointer of the requested record's group */
       
	XCoordinate thePositionMinLimit;                    /* ==> returns the position minimum limit */
    XCoordinate thePositionMaxLimit;     	            /* ==> returns the position maximum limit */

    XCoordinate theRotationMinLimit;                    /* Rotation minimum limit */
    XCoordinate theRotationMaxLimit;                    /* Rotation maximum limit */

    XCoordinate theRotationCenter;                      /* Rotation center */
    XCoordinate theStiffness;                  		    /* Stiffness */
    XCoordinate theViscosity;   		                /* the Viscosity */

    UInt32 theFlag;										/* the Flags */
} HostModelGroupLimitsRec, * HostModelGroupLimitsPtr;

/* Model generator plugin records */

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theFrameIndex;								/* <==  Sends the current frame index number */
	double theFrameTime;									/* <==  Sends the time of the current frame */
	double theFrameTimeDelta;							/* <==  Sends the time per frame (1/fps) */
	Matrix4 theCameraMatrix;							/* <==  Sends the camera matrix */
	Matrix4 theCameraRotMatrix;							/* <==  Sends the camera rotation matrix */
	Matrix4 theCameraBlurMatrix;						/* <==  Sends the camera motion blur matrix */
	Matrix4 theGroupMatrix;								/* <==  Sends the parent group absolute matrix */
	Matrix4 theGroupRotMatrix;							/* <==  Sends the parent group absolute rotation matrix */
	Matrix4 theGroupBlurMatrix;							/* <==  Sends the parent group absolute motion blur matrix */
} PluginModelSetupRec, *PluginModelSetupPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	GINFPtr theGINF;									/* <==  Sends the pointer to the group information record */
	unsigned char theChangeFlag;						/*  ==> Returns true if the record was modified */
} PluginModelGINFRec, *PluginModelGINFPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	GATRPtr theGATR;									/* <==  Sends the pointer to the group attributes record */
	unsigned char theChangeFlag;						/*  ==> Returns true if the record was modified */
} PluginModelGATRRec, *PluginModelGATRPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	GBLRPtr theGBLR;									/* <==  Sends the pointer to the group motion blur record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelGBLRRec, *PluginModelGBLRPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	TMAPPtr theTMAP;									/* <==  Sends the pointer to the group texture map record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelTMAPRec, *PluginModelTMAPPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	TATRPtr theTATR;									/* <==  Sends the pointer to the group texture attrubte record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelTATRRec, *PluginModelTATRPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	BMAPPtr theBMAP;									/* <==  Sends the pointer to the group secondary texture map record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelBMAPRec, *PluginModelBMAPPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	BATRPtr theBATR;									/* <==  Sends the pointer to the group secondary texture attrubte record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelBATRRec, *PluginModelBATRPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	RMAPPtr theRMAP;									/* <==  Sends the pointer to the group reflection map record */
	unsigned char theChangeFlag;							/*  ==> Returns true if the record was modified */
} PluginModelRMAPRec, *PluginModelRMAPPtr;

typedef struct {
	PluginInfoRec theInfo;								/* <==  Sends the plugin data, control and host data references */
	SInt32 theChildGroups;								/* <==  Sends the number of child groups of the plugin parent group  */
	UInt32 theVertexFlags;								/* <==  Sends vertex mapping flags of the plugin group (defined above) */
} PluginModelGenerateRec, *PluginModelGeneratePtr;

#ifdef __MACINTOSH__
#pragma options align=reset
#endif

/* The plugin's main routine */

EXPDLL SInt32 Main(IFFType theCommand, SInt32 theCommandRecSize, void* theCommandRec);

/* CallPluginRoutine may be used to call a plugin or host routine. (Provided for Pascal) */

SInt32 CallPluginRoutine (IFFType command, SInt32 parameterSize, void* parameters, EIPluginRoutine theRoutine);

/* DescribePluginError will return a description of a plugin's result code as a string */

void DescribePluginError(SInt32 theResultCode, PStrPtr theString);

/* EIPluginMalloc is called by a plug-in to malloc or realloc a pointer */

void EIPluginMalloc(PluginInfoPtr theInfo, SInt32 *theResult, UInt32 theSize, void **thePtr);

/* EIPluginFree is called by a plug-in to free a pointer */

void EIPluginFree(PluginInfoPtr theInfo, SInt32 *theResult, void *thePtr);

/* EIPluginStatusTime will return true when it is time to update the status */

#define EIPluginStatusTime(theStatus) (((theStatus)->theStatusTick == 0) || ((theStatus)->theStatusTick <= TICKCOUNT()))

/* EIPluginStatusCheck will check for user aborts */

void EIPluginStatusCheck(PluginInfoPtr theInfo, HostStatusRec *theStatus, SInt32 *theResult);

/* EIPluginStatusUpdate will update the status */

void EIPluginStatusUpdate(PluginInfoPtr theInfo, HostStatusRec *theStatus, float thePos, SInt32 theValue, SInt32 *theResult);

/* EIPluginStatusSetup will setup the status information */

void EIPluginStatusSetup(PluginInfoPtr theInfo, HostStatusRec *theStatus, PStrPtr theUnit, PStrPtr theHeading, SInt32 min, SInt32 max, float thePos, SInt32 theValue, SInt32 *theResult);

/* EIPluginStatusFinish will finish updating a status box */

void EIPluginStatusFinish(PluginInfoPtr theInfo, HostStatusRec *theStatus, SInt32 *theResult);

#ifdef __cplusplus
}
#endif

#endif
