/*  Copyright 1993-2011 EIAS3D, This Source code may be included in plug-ins, 
shaders, and utilities intended for use with the Electric Image Animation System.
This copyright notice must be included in its original form in all such products. 
This  source code may not be offered for sale or distributed by any entity 
without the express permission of EIAS3D.*/

#include "Matrix.h"
#include "FPU.h"

#define matrix_nearZero		1.0e-200					/* Quaternion near zero */
#define matrix_minScale		1.0e-10						/* Minimum matrix scale factor */
#define matrix_epsilon		0.000001					/* Floating point error */
#define matrix_epsilon2		0.00000016666667			/* epsilon / 6 */

static void PivotSolve4(Matrix4Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector);
static void PivotSolveR4(RMatrix4Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector);
static void PivotSolve3(Matrix3Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector);
static void PivotSolveR3(RMatrix3Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector);

/*
 *		Static Functions
 */

#if 1
static void PivotSolve4(Matrix4Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector)
{
	long i, j;

	for (i = 0; i <= 2; i++)
		for (j = i + 1; j <= 3; j++)
			coeffVector[indexVector[j]] = coeffVector[indexVector[j]] - srcMatrix->matrix[indexVector[j]][i] * coeffVector[indexVector[i]];
	resultVector[3] = coeffVector[indexVector[3]] / srcMatrix->matrix[indexVector[3]][3];
	for (i = 2; i >= 0; i--) {
		double sum = coeffVector[indexVector[i]];
		for (j = i + 1; j <= 3; j++)
			sum -= srcMatrix->matrix[indexVector[i]][j] * resultVector[j];
		resultVector[i] = sum / srcMatrix->matrix[indexVector[i]][i];
	}
}  /* PivotSolve4 */

static void PivotSolveR4(RMatrix4Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector)
{
	long i, j;

	for (i = 0; i <= 2; i++)
		for (j = i + 1; j <= 3; j++)
			coeffVector[indexVector[j]] = coeffVector[indexVector[j]] - srcMatrix->matrix[indexVector[j]][i] * coeffVector[indexVector[i]];
	resultVector[3] = coeffVector[indexVector[3]] / srcMatrix->matrix[indexVector[3]][3];
	for (i = 2; i >= 0; i--) {
		float sum = coeffVector[indexVector[i]];
		for (j = i + 1; j <= 3; j++)
			sum -= srcMatrix->matrix[indexVector[i]][j] * resultVector[j];
		resultVector[i] = sum / srcMatrix->matrix[indexVector[i]][i];
	}
}  /* PivotSolveR4 */

static void PivotSolve3(Matrix3Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector)
{
	long i, j;

	for (i = 0; i <= 1; i++)
		for (j = i + 1; j <= 2; j++)
			coeffVector[indexVector[j]] = coeffVector[indexVector[j]] - srcMatrix->matrix[indexVector[j]][i] * coeffVector[indexVector[i]];
	resultVector[2] = coeffVector[indexVector[2]] / srcMatrix->matrix[indexVector[2]][2];
	for (i = 1; i >= 0; i--) {
		double sum = coeffVector[indexVector[i]];
		for (j = i + 1; j <= 2; j++)
			sum -= srcMatrix->matrix[indexVector[i]][j] * resultVector[j];
		resultVector[i] = sum / srcMatrix->matrix[indexVector[i]][i];
	}
}  /* PivotSolve3 */

static void PivotSolveR3(RMatrix3Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector)
{
	long i, j;

	for (i = 0; i <= 1; i++)
		for (j = i + 1; j <= 2; j++)
			coeffVector[indexVector[j]] = coeffVector[indexVector[j]] - srcMatrix->matrix[indexVector[j]][i] * coeffVector[indexVector[i]];
	resultVector[2] = coeffVector[indexVector[2]] / srcMatrix->matrix[indexVector[2]][2];
	for (i = 1; i >= 0; i--) {
		float sum = coeffVector[indexVector[i]];
		for (j = i + 1; j <= 2; j++)
			sum -= srcMatrix->matrix[indexVector[i]][j] * resultVector[j];
		resultVector[i] = sum / srcMatrix->matrix[indexVector[i]][i];
	}
}  /* PivotSolveR3 */
#else
static void PivotSolve4(Matrix4Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector)
{
	long i0 = indexVector[0];
	long i1 = indexVector[1];
	long i2 = indexVector[2];
	long i3 = indexVector[3];

	coeffVector[i1] -= srcMatrix->matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix->matrix[i2][0] * coeffVector[i0] + srcMatrix->matrix[i2][1] * coeffVector[i1];
	coeffVector[i3] -= srcMatrix->matrix[i3][0] * coeffVector[i0] + srcMatrix->matrix[i3][1] * coeffVector[i1] + srcMatrix->matrix[i3][2] * coeffVector[i2];

	resultVector[3] = coeffVector[i3] / srcMatrix->matrix[i3][3];
	resultVector[2] = (coeffVector[i2] - srcMatrix->matrix[i2][3] * resultVector[3]) / srcMatrix->matrix[i2][2];
	resultVector[1] = (coeffVector[i1] - srcMatrix->matrix[i1][2] * resultVector[2] - srcMatrix->matrix[i1][3] * resultVector[3]) / srcMatrix->matrix[i1][1];
	resultVector[0] = (coeffVector[i0] - srcMatrix->matrix[i0][1] * resultVector[1] - srcMatrix->matrix[i0][2] * resultVector[2] - srcMatrix->matrix[i0][3] * resultVector[3]) / srcMatrix->matrix[i0][0];
	
}  /* PivotSolve4 */

static void PivotSolveR4(RMatrix4Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector)
{
	long i0 = indexVector[0];
	long i1 = indexVector[1];
	long i2 = indexVector[2];
	long i3 = indexVector[3];

	coeffVector[i1] -= srcMatrix->matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix->matrix[i2][0] * coeffVector[i0] + srcMatrix->matrix[i2][1] * coeffVector[i1];
	coeffVector[i3] -= srcMatrix->matrix[i3][0] * coeffVector[i0] + srcMatrix->matrix[i3][1] * coeffVector[i1] + srcMatrix->matrix[i3][2] * coeffVector[i2];

	resultVector[3] = coeffVector[i3] / srcMatrix->matrix[i3][3];
	resultVector[2] = (coeffVector[i2] - srcMatrix->matrix[i2][3] * resultVector[3]) / srcMatrix->matrix[i2][2];
	resultVector[1] = (coeffVector[i1] - srcMatrix->matrix[i1][2] * resultVector[2] - srcMatrix->matrix[i1][3] * resultVector[3]) / srcMatrix->matrix[i1][1];
	resultVector[0] = (coeffVector[i0] - srcMatrix->matrix[i0][1] * resultVector[1] - srcMatrix->matrix[i0][2] * resultVector[2] - srcMatrix->matrix[i0][3] * resultVector[3]) / srcMatrix->matrix[i0][0];
}  /* PivotSolveR4 */

static void PivotSolve3(Matrix3Ptr srcMatrix, double *coeffVector, double *resultVector, long *indexVector)
{
	long i0 = indexVector[0];
	long i1 = indexVector[1];
	long i2 = indexVector[2];
	
	coeffVector[i1] -= srcMatrix->matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix->matrix[i2][0] * coeffVector[i0] + srcMatrix->matrix[i2][1] * coeffVector[i1];

	resultVector[2] = coeffVector[i2] / srcMatrix->matrix[i2][2];
	resultVector[1] = (coeffVector[i1] - srcMatrix->matrix[i1][2] * resultVector[2]) / srcMatrix->matrix[i1][1];
	resultVector[0] = (coeffVector[i0] - srcMatrix->matrix[i0][1] * resultVector[1] - srcMatrix->matrix[i0][2] * resultVector[2]) / srcMatrix->matrix[i0][0];
}  /* PivotSolve3 */

static void PivotSolveR3(RMatrix3Ptr srcMatrix, float *coeffVector, float *resultVector, long *indexVector)
{
	long i0 = indexVector[0];
	long i1 = indexVector[1];
	long i2 = indexVector[2];
	
	coeffVector[i1] -= srcMatrix->matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix->matrix[i2][0] * coeffVector[i0] + srcMatrix->matrix[i2][1] * coeffVector[i1];

	resultVector[2] = coeffVector[i2] / srcMatrix->matrix[i2][2];
	resultVector[1] = (coeffVector[i1] - srcMatrix->matrix[i1][2] * resultVector[2]) / srcMatrix->matrix[i1][1];
	resultVector[0] = (coeffVector[i0] - srcMatrix->matrix[i0][1] * resultVector[1] - srcMatrix->matrix[i0][2] * resultVector[2]) / srcMatrix->matrix[i0][0];
}  /* PivotSolveR3 */
#endif

/* Math Routines */

void Normalize(double *a, double *b, double *c)
{
	double aNum = *a;
	double bNum = *b;
	double cNum = *c;
	double distance = SQR(aNum) + SQR(bNum) + SQR(cNum);
	
	if (distance > 0.0) {
		distance = 1.0 / SQRT(distance);
		*a = aNum * distance;
		*b = bNum * distance;
		*c = cNum * distance;
	} else {
		*a = 0.0;
		*b = 0.0;
		*c = 1.0;
	}
}  /* Normalize */

void NormalizeR(float *a, float *b, float *c)
{
	float aNum = *a;
	float bNum = *b;
	float cNum = *c;
	float distance = SQR(aNum) + SQR(bNum) + SQR(cNum);
	
	if (distance > 0.0f) {
		distance = 1.0f / (float)SQRT(distance);
		*a = aNum * distance;
		*b = bNum * distance;
		*c = cNum * distance;
	} else {
		*a = 0.0f;
		*b = 0.0f;
		*c = 1.0f;
	}
}  /* NormalizeR */

unsigned char CheckNormalize(double *a, double *b, double *c)
{
	double aNum = *a;
	double bNum = *b;
	double cNum = *c;
	double distance = SQR(aNum) + SQR(bNum) + SQR(cNum);

	if (distance > matrix_epsilon) {
		distance = 1.0 / SQRT(distance);
		*a = aNum * distance;
		*b = bNum * distance;
		*c = cNum * distance;
		return 1;
	} else
		return 0;
}  /* CheckNormalize */

unsigned char CheckNormalizeR(float *a, float *b, float *c)
{
	float aNum = *a;
	float bNum = *b;
	float cNum = *c;
	float distance = SQR(aNum) + SQR(bNum) + SQR(cNum);
	
	if (distance > matrix_epsilon) {
		distance = 1.0f / (float)SQRT(distance);
		*a = aNum * distance;
		*b = bNum * distance;
		*c = cNum * distance;
		return 1;
	} else
		return 0;
}  /* CheckNormalizeR */

void Normalize2D(double* a, double* b)
{
	double aNum = *a;
	double bNum = *b;
	double distance = SQR(aNum) + SQR(bNum);
	
	if (distance > 0.0f) {
		distance = 1.0f / SQRT(distance);
		*a *= distance;
		*b *= distance;
	} else {
		*a = 0.0;
		*b = 1.0;
	}
}  /* Normalize2D */

void Normalize2DR(float* a, float* b)
{
	float aNum = *a;
	float bNum = *b;
	float distance = SQR(aNum) + SQR(bNum);
	
	if (distance > 0.0f) {
		distance = 1.0f / (float)SQRT(distance);
		*a = aNum * distance;
		*b = bNum * distance;
	} else {
		*a = 0.0f;
		*b = 1.0f;
	}
}  /* Normalize2DR */

void VectorProduct(double a1, double b1, double c1, double a2, double b2, double c2, double* a3, double* b3, double* c3)
{
	*a3 = b1 * c2 - c1 * b2;
	*b3 = c1 * a2 - a1 * c2;
	*c3 = a1 * b2 - b1 * a2;
}  /* VectorProduct */

void VectorProductR(float a1, float b1, float c1, float a2, float b2, float c2, float* a3, float* b3, float* c3)
{
	*a3 = b1 * c2 - c1 * b2;
	*b3 = c1 * a2 - a1 * c2;
	*c3 = a1 * b2 - b1 * a2;
}  /* VectorProductR */

double DotProduct(double a1, double b1, double c1, double a2, double b2, double c2)
{
	return a1 * a2 + b1 * b2 + c1 * c2;
}  /* DotProduct */

float DotProductR(float a1, float b1, float c1, float a2, float b2, float c2)
{
	return a1 * a2 + b1 * b2 + c1 * c2;
}  /* DotProductR */

void AlignVectors(double axisA, double axisB, double axisC, double srcA, double srcB, double srcC, double *dstA, double *dstB, double *dstC)
{
	double tempA = axisB * srcC - axisC * srcB;
	double tempB = axisC * srcA - axisA * srcC;
	double tempC = axisA * srcB - axisB * srcA;

	*dstA = axisB * tempC - axisC * tempB;
	*dstB = axisC * tempA - axisA * tempC;
	*dstC = axisA * tempB - axisB * tempA;
}  /* AlignVectors */

void AlignVectorsR(float axisA, float axisB, float axisC, float srcA, float srcB, float srcC, float *dstA, float *dstB, float *dstC)
{
	float tempA = axisB * srcC - axisC * srcB;
	float tempB = axisC * srcA - axisA * srcC;
	float tempC = axisA * srcB - axisB * srcA;

	*dstA = axisB * tempC - axisC * tempB;
	*dstB = axisC * tempA - axisA * tempC;
	*dstC = axisA * tempB - axisB * tempA;
}  /* AlignVectorsR */

double TriangleArea(double x1, double y1, double x2, double y2, double x3, double y3)
{
	double area = 0.5 * ((x1 * y2 - y1 * x2) + (x2 * y3 - y2 * x3) + (x3 * y1 - y3 * x1));

	return FABS(area);
}  /* TriangleArea */

double TriangleArea3(double x1, double y1, double z1, double x2, double y2, double z2, double x3, double y3, double z3)
{
	double area;

	x1 -= x3;
	y1 -= y3;
	z1 -= z3;
	x2 -= x3;
	y2 -= y3;
	z2 -= z3;
	area = 0.5 * ((y1 * z2 - z1 * y2) + (z1 * x2 - x1 * z2) + (x1 * y2 - y1 * x2));
	return FABS(area);
}  /* TriangleArea3 */

double QuadrangleArea(double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4)
{
	double factor = x3 * y1 - y3 * x1;
	double area1 = (x1 * y2 - y1 * x2) + (x2 * y3 - y2 * x3) + factor;
	double area2 = (x1 * y4 - y1 * x4) + (x4 * y3 - y4 * x3) + factor;

	return 0.5 * (FABS(area1) + FABS(area2));
}  /* QuadrangeArea */

double QuadrangleArea3(double x1, double y1, double z1, double x2, double y2, double z2, double x3, double y3, double z3, double x4, double y4, double z4)
{
	double area1, area2;

	x1 -= x3;
	y1 -= y3;
	z1 -= z3;
	x2 -= x3;
	y2 -= y3;
	z2 -= z3;
	x4 -= x3;
	y4 -= y3;
	z4 -= z3;
	area1 = (y1 * z2 - z1 * y2) + (z1 * x2 - x1 * z2) + (x1 * y2 - y1 * x2);
	area2 = (y1 * z4 - z1 * y4) + (z1 * x4 - x1 * z4) + (x1 * y4 - y1 * x4);
	return 0.5 * (FABS(area1) + FABS(area2));
}  /* QuadrangleArea3 */

double CalcDotValue(double theAngle)
{
	return COS(theAngle * PIFac);
}  /* CalcDotValue */

double CalcAngle(double dotValue)
{
	return ACOS(dotValue) * (180.0 / PI);
}  /* CalcAngle */

void Rot(double sinTheta, double cosTheta, double* n1, double* n2)
{
	double n3 = *n1 * cosTheta - *n2 * sinTheta;

	*n2 = *n1 * sinTheta + *n2 * cosTheta;
	*n1 = n3;
}  /* Rot */

void RotR(float sinTheta, float cosTheta, float* n1, float* n2)
{
	float n3 = *n1 * cosTheta - *n2 * sinTheta;

	*n2 = *n1 * sinTheta + *n2 * cosTheta;
	*n1 = n3;
}  /* RotR */

long Sgn(double n)
{
	return n >= 0.0 ? 1 : -1;
}  /* Sgn */

unsigned char Equal0(float n)
{
	n += 1.0f;
	return n == 1.0f;
}  /* Equal0 */

unsigned char Equal1(float n)
{
	return n == 1.0f;
}  /* Equal1 */

double LineDistance(double xp1, double yp1, double zp1, double xn1, double yn1, double zn1, double length1, double xp2, double yp2, double zp2, double xn2, double yn2, double zn2, double length2)
{
	double xp3, yp3, zp3;
	
	/* VectorProduct(xn1, yn1, zn1, xn2, yn2, zn2, m02, m12, m22); */
	
	double m1 = yn1 * zn2 - zn1 * yn2;
	double m2 = zn1 * xn2 - xn1 * zn2;
	double m3 = xn1 * yn2 - yn1 * xn2;
	double distance = SQR(m1) + SQR(m2) + SQR(m3);
	
	if (distance > 0.0) {
		xp3 = xp2 - xp1;
		yp3 = yp2 - yp1;
		zp3 = zp2 - zp1;
		
		/* distance=Det(a)/distance */
		
		distance = (xp3 * yn2 * m3 + xn2 * m2 * zp3 + m1 * yp3 * zn2 - m1 * yn2 * zp3 - xn2 * yp3 * m3 - xp3 * m2 * zn2) / distance;
		if (distance > 0.0) {
			if (distance > length1)
				distance = length1;
			xp3 = xp1 + xn1 * distance;
			yp3 = yp1 + yn1 * distance;
			zp3 = zp1 + zn1 * distance;
		} else {
			xp3 = xp1;
			yp3 = yp1;
			zp3 = zp1;
		}
	} else {
		xp3 = xp1;
		yp3 = yp1;
		zp3 = zp1;
	}
	distance = (xp3 * xn2 + yp3 * yn2 + zp3 * zn2) - (xp2 * xn2 + yp2 * yn2 + zp2 * zn2);
	if (distance >= 0.0)
		if (distance > length2) {
			distance = length2;
			xp2 += xn2 * distance;
			yp2 += yn2 * distance;
			zp2 += zn2 * distance;
			distance = (xp2 * xn1 + yp2 * yn1 + zp2 * zn1) - (xp1 * xn1 + yp1 * yn1 + zp1 * zn1);
			if (distance > 0.0) {
				if (distance > length1)
					distance = length1;
				xp1 += xn1 * distance;
				yp1 += yn1 * distance;
				zp1 += zn1 * distance;
			}
		} else {
			xp2 += xn2 * distance;
			yp2 += yn2 * distance;
			zp2 += zn2 * distance;
			xp1 = xp3;
			yp1 = yp3;
			zp1 = zp3;
		}
	else {
		distance = (xp2 * xn1 + yp2 * yn1 + zp2 * zn1) - (xp1 * xn1 + yp1 * yn1 + zp1 * zn1);
		if (distance > 0.0) {
			if (distance > length1)
				distance = length1;
			xp1 += xn1 * distance;
			yp1 += yn1 * distance;
			zp1 += zn1 * distance;
		}
	}
	xp2 -= xp1;
	yp2 -= yp1;
	zp2 -= zp1;
	return SQRT(SQR(xp2) + SQR(yp2) + SQR(zp2));
}  /* LineDistance */

void Line1Closest(double* xp1, double* yp1, double* zp1, double xn1, double yn1, double zn1, double length1, double xp2, double yp2, double zp2, double xn2, double yn2, double zn2)
{
	Matrix3 m;
	double distance;

	VectorProduct(xn1, yn1, zn1, xn2, yn2, zn2, &m.m.m02, &m.m.m12, &m.m.m22);
	distance = SQR(m.m.m02) + SQR(m.m.m12) + SQR(m.m.m22);
	if (distance > 0.0) {
		m.m.m00 = xp2 - *xp1;
		m.m.m10 = yp2 - *yp1;
		m.m.m20 = zp2 - *zp1;
		m.m.m01 = xn2;
		m.m.m11 = yn2;
		m.m.m21 = zn2;
		distance = Det(&m) / distance;
		if (distance > 0.0) {
			if (distance > length1)
				distance = length1;
			*xp1 += xn1 * distance;
			*yp1 += yn1 * distance;
			*zp1 += zn1 * distance;
		}
	}
}  /* Line1Closest */

void Line2Closest(double* xp1, double* yp1, double* zp1, double xn1, double yn1, double zn1, double length1, double* xp2, double* yp2, double* zp2, double xn2, double yn2, double zn2, double length2)
{
	Matrix3 m;
	double distance;
	double xp3, yp3, zp3;

	VectorProduct(xn1, yn1, zn1, xn2, yn2, zn2, &m.m.m02, &m.m.m12, &m.m.m22);
	distance = SQR(m.m.m02) + SQR(m.m.m12) + SQR(m.m.m22);
	if (distance > 0.0) {
		m.m.m00 = *xp2 - *xp1;
		m.m.m10 = *yp2 - *yp1;
		m.m.m20 = *zp2 - *zp1;
		m.m.m01 = xn2;
		m.m.m11 = yn2;
		m.m.m21 = zn2;
		distance = Det(&m) / distance;
		if (distance > 0.0) {
			if (distance > length1)
				distance = length1;
			xp3 = *xp1 + xn1 * distance;
			yp3 = *yp1 + yn1 * distance;
			zp3 = *zp1 + zn1 * distance;
		} else {
			xp3 = *xp1;
			yp3 = *yp1;
			zp3 = *zp1;
		}
	} else {
		xp3 = *xp1;
		yp3 = *yp1;
		zp3 = *zp1;
	}
	distance = (xp3 * xn2 + yp3 * yn2 + zp3 * zn2) - (*xp2 * xn2 + *yp2 * yn2 + *zp2 * zn2);
	if (distance >= 0.0)
		if (distance > length2) {
			distance = length2;
			*xp2 += xn2 * distance;
			*yp2 += yn2 * distance;
			*zp2 += zn2 * distance;
			distance = (*xp2 * xn1 + *yp2 * yn1 + *zp2 * zn1) - (*xp1 * xn1 + *yp1 * yn1 + *zp1 * zn1);
			if (distance > 0.0) {
				if (distance > length1)
					distance = length1;
				*xp1 += xn1 * distance;
				*yp1 += yn1 * distance;
				*zp1 += zn1 * distance;
			}
		} else {
			*xp2 += xn2 * distance;
			*yp2 += yn2 * distance;
			*zp2 += zn2 * distance;
			*xp1 = xp3;
			*yp1 = yp3;
			*zp1 = zp3;
		}
	else {
		distance = (*xp2 * xn1 + *yp2 * yn1 + *zp2 * zn1) - (*xp1 * xn1 + *yp1 * yn1 + *zp1 * zn1);
		if (distance > 0.0) {
			if (distance > length1)
				distance = length1;
			*xp1 += xn1 * distance;
			*yp1 += yn1 * distance;
			*zp1 += zn1 * distance;
		}
	}
}  /* Line2Closest */

void CalcVector(CoordPtr coord1, CoordPtr coord2, CoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
}  /* CalcVector */

unsigned char CalcVector2(CoordPtr coord1, CoordPtr coord2, CoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	return (distance >= matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) || (distance >= matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2))));
}  /* CalcVector2 */

void CalcDVector(DCoordPtr coord1, DCoordPtr coord2, DCoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
}  /* CalcDVector */

unsigned char CalcDVector2(DCoordPtr coord1, DCoordPtr coord2, DCoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	return (distance >= matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) || (distance >= matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2))));
}  /* CalcDVector2 */

unsigned char CalcNormal(CoordPtr coord1, CoordPtr coord2, CoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	if ((distance < matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) && (distance < matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2)))))
		return 0;
	else {
		distance = 1.0 / distance;
		*aNum *= distance;
		*bNum *= distance;
		*cNum *= distance;
		return 1;
	}
}  /* CalcNormal */

unsigned char CalcDNormal(DCoordPtr coord1, DCoordPtr coord2, DCoordPtr coord3, double* aNum, double* bNum, double* cNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	if ((distance < matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) && (distance < matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2)))))
		return 0;
	else {
		distance = 1.0 / distance;
		*aNum *= distance;
		*bNum *= distance;
		*cNum *= distance;
		return 1;
	}
}  /* CalcDNormal */

void CalcPlaneFormula(CoordPtr coord1, CoordPtr coord2, CoordPtr coord3, double* aNum, double* bNum, double* cNum, double* dNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;
	
	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	if ((distance < matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) && (distance < matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2))))) {
		*aNum = 0.0;
		*bNum = 0.0;
		*cNum = 1.0;
	} else {
		distance = 1.0 / distance;
		*aNum *= distance;
		*bNum *= distance;
		*cNum *= distance;
	}
	*dNum = -(*aNum * coord1->x + *bNum * coord1->y + *cNum * coord1->z);
}  /* CalcPlaneFormula */

void CalcDPlaneFormula(DCoordPtr coord1, DCoordPtr coord2, DCoordPtr coord3, double* aNum, double* bNum, double* cNum, double* dNum)
{
	double distance;
	double a1 = coord2->x - coord1->x;
	double b1 = coord2->y - coord1->y;
	double c1 = coord2->z - coord1->z;
	double a2 = coord3->x - coord2->x;
	double b2 = coord3->y - coord2->y;
	double c2 = coord3->z - coord2->z;

	*aNum = b1 * c2 - b2 * c1;
	*bNum = c1 * a2 - c2 * a1;
	*cNum = a1 * b2 - a2 * b1;
	distance = SQRT(SQR(*aNum) + SQR(*bNum) + SQR(*cNum));
	if ((distance < matrix_epsilon2 * (FABS(a1) + FABS(b1) + FABS(c1) + FABS(a2) + FABS(b2) + FABS(c2))) && (distance < matrix_epsilon2 * (SQRT(SQR(a1) + SQR(b1) + SQR(c1)) + SQRT(SQR(a2) + SQR(b2) + SQR(c2))))) {
		*aNum = 0.0;
		*bNum = 0.0;
		*cNum = 1.0;
	} else {
		distance = 1.0 / distance;
		*aNum *= distance;
		*bNum *= distance;
		*cNum *= distance;
	}
	*dNum = -(*aNum * coord1->x + *bNum * coord1->y + *cNum * coord1->z);
}  /* CalcDPlaneFormula */

/* Matrix Routines */

unsigned char TestIdentityMatrix4(Matrix4Ptr theMatrix)
{
	return EQUAL1(theMatrix->m.m00) && EQUAL1(theMatrix->m.m11) && EQUAL1(theMatrix->m.m22) && EQUAL1(theMatrix->m.m33) && EQUAL0(theMatrix->m.m01) && EQUAL0(theMatrix->m.m02) && EQUAL0(theMatrix->m.m03) && EQUAL0(theMatrix->m.m10) && EQUAL0(theMatrix->m.m12) && EQUAL0(theMatrix->m.m13) && EQUAL0(theMatrix->m.m20) && EQUAL0(theMatrix->m.m21) && EQUAL0(theMatrix->m.m23) && EQUAL0(theMatrix->m.m30) && EQUAL0(theMatrix->m.m31) && EQUAL0(theMatrix->m.m32);
}  /* TestIdentityMatrix4 */

unsigned char TestIdentityRMatrix4(RMatrix4Ptr theMatrix)
{
	return EQUAL1(theMatrix->m.m00) && EQUAL1(theMatrix->m.m11) && EQUAL1(theMatrix->m.m22) && EQUAL1(theMatrix->m.m33) && EQUAL0(theMatrix->m.m01) && EQUAL0(theMatrix->m.m02) && EQUAL0(theMatrix->m.m03) && EQUAL0(theMatrix->m.m10) && EQUAL0(theMatrix->m.m12) && EQUAL0(theMatrix->m.m13) && EQUAL0(theMatrix->m.m20) && EQUAL0(theMatrix->m.m21) && EQUAL0(theMatrix->m.m23) && EQUAL0(theMatrix->m.m30) && EQUAL0(theMatrix->m.m31) && EQUAL0(theMatrix->m.m32);
}  /* TestIdentityRMatrix4 */

unsigned char TestIdentityMatrix3(Matrix3Ptr theMatrix)
{
	return EQUAL1(theMatrix->m.m00) && EQUAL1(theMatrix->m.m11) && EQUAL1(theMatrix->m.m22) && EQUAL0(theMatrix->m.m01) && EQUAL0(theMatrix->m.m02) && EQUAL0(theMatrix->m.m10) && EQUAL0(theMatrix->m.m12) && EQUAL0(theMatrix->m.m20) && EQUAL0(theMatrix->m.m21);
}  /* TestIdentityMatrix3 */

unsigned char TestIdentityRMatrix3(RMatrix3Ptr theMatrix)
{
	return EQUAL1(theMatrix->m.m00) && EQUAL1(theMatrix->m.m11) && EQUAL1(theMatrix->m.m22) && EQUAL0(theMatrix->m.m01) && EQUAL0(theMatrix->m.m02) && EQUAL0(theMatrix->m.m10) && EQUAL0(theMatrix->m.m12) && EQUAL0(theMatrix->m.m20) && EQUAL0(theMatrix->m.m21);
}  /* TestIdentityRMatrix3 */

unsigned char TestZeroMatrix4(Matrix4Ptr theMatrix)
{
	double sum = 10.0;

	sum += theMatrix->m.m00;
	sum += theMatrix->m.m01;
	sum += theMatrix->m.m02;
	sum += theMatrix->m.m10;
	sum += theMatrix->m.m11;
	sum += theMatrix->m.m12;
	sum += theMatrix->m.m20;
	sum += theMatrix->m.m21;
	sum += theMatrix->m.m22;
	return sum == 10.0;
}  /* TestZeroMatrix4 */

unsigned char TestZeroRMatrix4(RMatrix4Ptr theMatrix)
{
	float sum = 1.0f;

	sum += theMatrix->m.m00;
	sum += theMatrix->m.m01;
	sum += theMatrix->m.m02;
	sum += theMatrix->m.m10;
	sum += theMatrix->m.m11;
	sum += theMatrix->m.m12;
	sum += theMatrix->m.m20;
	sum += theMatrix->m.m21;
	sum += theMatrix->m.m22;
	return sum == 1.0;
}  /* TestZeroRMatrix4 */

unsigned char TestZeroMatrix3(Matrix3Ptr theMatrix)
{
	double sum = 10.0;

	sum += theMatrix->m.m00;
	sum += theMatrix->m.m01;
	sum += theMatrix->m.m02;
	sum += theMatrix->m.m10;
	sum += theMatrix->m.m11;
	sum += theMatrix->m.m12;
	sum += theMatrix->m.m20;
	sum += theMatrix->m.m21;
	sum += theMatrix->m.m22;
	return sum == 10.0;
}  /* TestZeroMatrix3 */

unsigned char TestZeroRMatrix3(RMatrix3Ptr theMatrix)
{
	float sum = 1.0f;

	sum += theMatrix->m.m00;
	sum += theMatrix->m.m01;
	sum += theMatrix->m.m02;
	sum += theMatrix->m.m10;
	sum += theMatrix->m.m11;
	sum += theMatrix->m.m12;
	sum += theMatrix->m.m20;
	sum += theMatrix->m.m21;
	sum += theMatrix->m.m22;
	return sum == 1.0;
}  /* TestZeroRMatrix3 */

void IdentityMatrix4(Matrix4Ptr theMatrix)
{
	static Matrix4 ident = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0};

	*theMatrix = ident;
}  /* IdentityMatrix4 */

void IdentityRMatrix4(RMatrix4Ptr theMatrix)
{
	static RMatrix4 ident = {1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f};

	*theMatrix = ident;
}  /* IdentityMatrix4 */

void IdentityMatrix3(Matrix3Ptr theMatrix)
{
	static Matrix3 ident = {1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0};

	*theMatrix = ident;
}  /* IdentityMatrix3 */

void IdentityRMatrix3(RMatrix3Ptr theMatrix)
{
	static RMatrix3 ident = {1.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 1.0f};

	*theMatrix = ident;
}  /* IdentityRMatrix3 */

/* SenseMatrix tests the handedness of the matrix.  Returns 1 if the third row of the */
/* matrix points in the same direction as the cross products of the first two rows */

unsigned char SenseMatrix4(Matrix4Ptr theMatrix)
{
	double a = theMatrix->m.m01 * theMatrix->m.m12 - theMatrix->m.m02 * theMatrix->m.m11;
	double b = theMatrix->m.m02 * theMatrix->m.m10 - theMatrix->m.m00 * theMatrix->m.m12;
	double c = theMatrix->m.m00 * theMatrix->m.m11 - theMatrix->m.m01 * theMatrix->m.m10;
	
	return (a * theMatrix->m.m20 + b * theMatrix->m.m21 + c * theMatrix->m.m22) > 0.0;
}  /* SenseMatrix4 */

unsigned char SenseRMatrix4(RMatrix4Ptr theMatrix)
{
	float a = theMatrix->m.m01 * theMatrix->m.m12 - theMatrix->m.m02 * theMatrix->m.m11;
	float b = theMatrix->m.m02 * theMatrix->m.m10 - theMatrix->m.m00 * theMatrix->m.m12;
	float c = theMatrix->m.m00 * theMatrix->m.m11 - theMatrix->m.m01 * theMatrix->m.m10;
	
	return (a * theMatrix->m.m20 + b * theMatrix->m.m21 + c * theMatrix->m.m22) > 0.0;
}  /* SenseRMatrix4 */

unsigned char SenseMatrix3(Matrix3Ptr theMatrix)
{
	double a = theMatrix->m.m01 * theMatrix->m.m12 - theMatrix->m.m02 * theMatrix->m.m11;
	double b = theMatrix->m.m02 * theMatrix->m.m10 - theMatrix->m.m00 * theMatrix->m.m12;
	double c = theMatrix->m.m00 * theMatrix->m.m11 - theMatrix->m.m01 * theMatrix->m.m10;
	
	return (a * theMatrix->m.m20 + b * theMatrix->m.m21 + c * theMatrix->m.m22) > 0.0;
}  /* SenseMatrix3 */

unsigned char SenseRMatrix3(RMatrix3Ptr theMatrix)
{
	float a = theMatrix->m.m01 * theMatrix->m.m12 - theMatrix->m.m02 * theMatrix->m.m11;
	float b = theMatrix->m.m02 * theMatrix->m.m10 - theMatrix->m.m00 * theMatrix->m.m12;
	float c = theMatrix->m.m00 * theMatrix->m.m11 - theMatrix->m.m01 * theMatrix->m.m10;
	
	return (a * theMatrix->m.m20 + b * theMatrix->m.m21 + c * theMatrix->m.m22) > 0.0;
}  /* SenseRMatrix3 */

#if 0
void InvertMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	static Matrix4 identity = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0};
	Vector4 coeffVector, resultVector, scaleVector;
	long indexVector[4];
	long i, j, k;

	/* PivotGaus */
	
	for (i = 0; i <= 3; i++) {
		double sMax, temp;

		indexVector[i] = i;
		sMax = FABS(srcMatrix.matrix[i][0]);
		temp = FABS(srcMatrix.matrix[i][1]);
		if (sMax < temp)
			sMax = temp;
		temp = FABS(srcMatrix.matrix[i][2]);
		if (sMax < temp)
			sMax = temp;
		temp = FABS(srcMatrix.matrix[i][3]);
		if (sMax < temp)
			scaleVector[i] = temp;
		else
			scaleVector[i] = sMax;
	}
	for (k = 0; k <= 2; k++) {
		double rMax = 0.0;
		long indexSwap;

		j = 0;
		for (i = k; i <= 3; i++) {
			double r = FABS(srcMatrix.matrix[indexVector[i]][k]) / scaleVector[indexVector[i]];
			if (r > rMax) {
				j = i;
				rMax = r;
			}
		}
		indexSwap = indexVector[j];
		indexVector[j] = indexVector[k];
		indexVector[k] = indexSwap;
		for (i = k + 1; i <= 3; i++) {
			double xMult = srcMatrix.matrix[indexVector[i]][k] / srcMatrix.matrix[indexSwap][k];
			for (j = k + 1; j <= 3; j++)
				srcMatrix.matrix[indexVector[i]][j] = srcMatrix.matrix[indexVector[i]][j] - xMult * srcMatrix.matrix[indexSwap][j];
			srcMatrix.matrix[indexVector[i]][k] = xMult;
		}
	}

	/* Solve the matrix */
	
	for (i = 0; i <= 3; i++) {
		coeffVector[0] = identity.mv[i][0];
		coeffVector[1] = identity.mv[i][1];
		coeffVector[2] = identity.mv[i][2];
		coeffVector[3] = identity.mv[i][3];
		PivotSolve4(&srcMatrix, coeffVector, resultVector, indexVector);
		theMatrix->matrix[0][i] = resultVector[0];
		theMatrix->matrix[1][i] = resultVector[1];
		theMatrix->matrix[2][i] = resultVector[2];
		theMatrix->matrix[3][i] = resultVector[3];
	}
}  /* InvertMatrix4 */

void InvertRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	static RMatrix4 identity = {1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f};
	RVector4 coeffVector, resultVector, scaleVector;
	long indexVector[4];
	long i, j, k;

	/* PivotGaus */
	
	for (i = 0; i <= 3; i++) {
		float sMax, temp;

		indexVector[i] = i;
		sMax = (float)FABS(srcMatrix.matrix[i][0]);
		temp = (float)FABS(srcMatrix.matrix[i][1]);
		if (sMax < temp)
			sMax = temp;
		temp = (float)FABS(srcMatrix.matrix[i][2]);
		if (sMax < temp)
			sMax = temp;
		temp = (float)FABS(srcMatrix.matrix[i][3]);
		if (sMax < temp)
			scaleVector[i] = temp;
		else
			scaleVector[i] = sMax;
	}
	for (k = 0; k <= 2; k++) {
		float rMax = 0.0f;
		long indexSwap;

		j = 0;
		for (i = k; i <= 3; i++) {
			float r = (float)FABS(srcMatrix.matrix[indexVector[i]][k]) / scaleVector[indexVector[i]];
			if (r > rMax) {
				j = i;
				rMax = r;
			}
		}
		indexSwap = indexVector[j];
		indexVector[j] = indexVector[k];
		indexVector[k] = indexSwap;
		for (i = k + 1; i <= 3; i++) {
			float xMult = srcMatrix.matrix[indexVector[i]][k] / srcMatrix.matrix[indexSwap][k];
			for (j = k + 1; j <= 3; j++)
				srcMatrix.matrix[indexVector[i]][j] = srcMatrix.matrix[indexVector[i]][j] - xMult * srcMatrix.matrix[indexSwap][j];
			srcMatrix.matrix[indexVector[i]][k] = xMult;
		}
	}

	/* Solve the matrix */
	
	for (i = 0; i <= 3; i++) {
		coeffVector[0] = identity.mv[i][0];
		coeffVector[1] = identity.mv[i][1];
		coeffVector[2] = identity.mv[i][2];
		coeffVector[3] = identity.mv[i][3];
		PivotSolveR4(&srcMatrix, coeffVector, resultVector, indexVector);
		theMatrix->matrix[0][i] = resultVector[0];
		theMatrix->matrix[1][i] = resultVector[1];
		theMatrix->matrix[2][i] = resultVector[2];
		theMatrix->matrix[3][i] = resultVector[3];
	}
}  /* InvertRMatrix4 */

void InvertMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	static Matrix3 identity = {1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0};
	Vector3 coeffVector, resultVector, scaleVector;
	long indexVector[3];
	long i, j, k;

	/* PivotGaus */
	
	for (i = 0; i <= 2; i++) {
		double sMax, temp;

		indexVector[i] = i;
		sMax = FABS(srcMatrix.matrix[i][0]);
		temp = FABS(srcMatrix.matrix[i][1]);
		if (sMax < temp)
			sMax = temp;
		temp = FABS(srcMatrix.matrix[i][2]);
		if (sMax < temp)
			scaleVector[i] = temp;
		else
			scaleVector[i] = sMax;
	}
	for (k = 0; k <= 1; k++) {
		double rMax = 0.0;
		long indexSwap;

		j = 0;
		for (i = k; i <= 2; i++) {
			double r = FABS(srcMatrix.matrix[indexVector[i]][k]) / scaleVector[indexVector[i]];
			if (r > rMax) {
				j = i;
				rMax = r;
			}
		}
		indexSwap = indexVector[j];
		indexVector[j] = indexVector[k];
		indexVector[k] = indexSwap;
		for (i = k + 1; i <= 2; i++) {
			double xMult = srcMatrix.matrix[indexVector[i]][k] / srcMatrix.matrix[indexSwap][k];
			for (j = k + 1; j <= 2; j++)
				srcMatrix.matrix[indexVector[i]][j] = srcMatrix.matrix[indexVector[i]][j] - xMult * srcMatrix.matrix[indexSwap][j];
			srcMatrix.matrix[indexVector[i]][k] = xMult;
		}
	}

	/* Solve the matrix */
	
	for (i = 0; i <= 2; i++) {
		coeffVector[0] = identity.mv[i][0];
		coeffVector[1] = identity.mv[i][1];
		coeffVector[2] = identity.mv[i][2];
		PivotSolve3(&srcMatrix, coeffVector, resultVector, indexVector);
		theMatrix->matrix[0][i] = resultVector[0];
		theMatrix->matrix[1][i] = resultVector[1];
		theMatrix->matrix[2][i] = resultVector[2];
	}
}  /* InvertMatrix3 */

void InvertRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	static RMatrix3 identity = {1.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 1.0f};
	RVector3 coeffVector, resultVector, scaleVector;
	long indexVector[3];
	long i, j, k;

	/* PivotGaus */
	
	for (i = 0; i <= 2; i++) {
		float sMax, temp;

		indexVector[i] = i;
		sMax = (float)FABS(srcMatrix.matrix[i][0]);
		temp = (float)FABS(srcMatrix.matrix[i][1]);
		if (sMax < temp)
			sMax = temp;
		temp = (float)FABS(srcMatrix.matrix[i][2]);
		if (sMax < temp)
			scaleVector[i] = temp;
		else
			scaleVector[i] = sMax;
	}
	for (k = 0; k <= 1; k++) {
		float rMax = 0.0f;
		long indexSwap;

		j = 0;
		for (i = k; i <= 2; i++) {
			float r = (float)FABS(srcMatrix.matrix[indexVector[i]][k]) / scaleVector[indexVector[i]];
			if (r > rMax) {
				j = i;
				rMax = r;
			}
		}
		indexSwap = indexVector[j];
		indexVector[j] = indexVector[k];
		indexVector[k] = indexSwap;
		for (i = k + 1; i <= 2; i++) {
			float xMult = srcMatrix.matrix[indexVector[i]][k] / srcMatrix.matrix[indexSwap][k];
			for (j = k + 1; j <= 2; j++)
				srcMatrix.matrix[indexVector[i]][j] = srcMatrix.matrix[indexVector[i]][j] - xMult * srcMatrix.matrix[indexSwap][j];
			srcMatrix.matrix[indexVector[i]][k] = xMult;
		}
	}

	/* Solve the matrix */
	
	for (i = 0; i <= 2; i++) {
		coeffVector[0] = identity.mv[i][0];
		coeffVector[1] = identity.mv[i][1];
		coeffVector[2] = identity.mv[i][2];
		PivotSolveR3(&srcMatrix, coeffVector, resultVector, indexVector);
		theMatrix->matrix[0][i] = resultVector[0];
		theMatrix->matrix[1][i] = resultVector[1];
		theMatrix->matrix[2][i] = resultVector[2];
	}
}  /* InvertRMatrix3 */
#else
void InvertMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	static Vector3 coeffVector;	/* Fixed a PC compiler optimizer bug by making this variable a static */
	Vector3 scaleVector;
	double n1, n2;
	double r0, r1, r2;
	double f0, f1, f2;
	double xMult, xFac;
	long indexSwap;
	long i0, i1, i2;
	
	/* PivotGaus */
	
	r0 = srcMatrix.matrix[0][0];
	n1 = srcMatrix.matrix[0][1];
	n2 = srcMatrix.matrix[0][2];
	r0 = FABS(r0);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r0 *= scaleVector[0] = 1.0 / MAX3(r0, n1, n2);

	r1 = srcMatrix.matrix[1][0];
	n1 = srcMatrix.matrix[1][1];
	n2 = srcMatrix.matrix[1][2];
	r1 = FABS(r1);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r1 *= scaleVector[1] = 1.0 / MAX3(r1, n1, n2);

	r2 = srcMatrix.matrix[2][0];
	n1 = srcMatrix.matrix[2][1];
	n2 = srcMatrix.matrix[2][2];
	r2 = FABS(r2);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r2 *= scaleVector[2] = 1.0 / MAX3(r2, n1, n2);

	if (r0 >= r1)
		if (r0 >= r2) {
			i0 = 0;
			i1 = 1;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	else
		if (r1 >= r2) {
			i0 = 1;
			i1 = 0;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}

	xFac = 1.0 / srcMatrix.matrix[i0][0];
	
	xMult = srcMatrix.matrix[i1][0] *= xFac;
	srcMatrix.matrix[i1][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i1][2] -= xMult * srcMatrix.matrix[i0][2];

	xMult = srcMatrix.matrix[i2][0] *= xFac;
	srcMatrix.matrix[i2][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i0][2];

	r1 = srcMatrix.matrix[i1][1] * scaleVector[i1];
	r2 = srcMatrix.matrix[i2][1] * scaleVector[i2];
	r1 = FABS(r1);
	r2 = FABS(r2);

	if (r2 > r1) {
		indexSwap = i2;
		i2 = i1;
		i1 = indexSwap;
	}

	xMult = srcMatrix.matrix[i2][1] /= srcMatrix.matrix[i1][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i1][2];

	/* Solve the matrix */
	
	f0 = 1.0 / srcMatrix.matrix[i0][0];
	f1 = 1.0 / srcMatrix.matrix[i1][1];
	f2 = 1.0 / srcMatrix.matrix[i2][2];

	coeffVector[0] = 1.0;
	coeffVector[1] = 0.0;
	coeffVector[2] = 0.0;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][0] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][0] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][0] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0;
	coeffVector[1] = 1.0;
	coeffVector[2] = 0.0;
	
	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][1] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][1] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][1] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0;
	coeffVector[1] = 0.0;
	coeffVector[2] = 1.0;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][2] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][2] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][2] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	/* Calculate the inverse translation */
	
	f0 = -srcMatrix.m.m30;
	f1 = -srcMatrix.m.m31;
	f2 = -srcMatrix.m.m32;
	theMatrix->m.m30 = f0 * theMatrix->m.m00 + f1 * theMatrix->m.m10 + f2 * theMatrix->m.m20;
	theMatrix->m.m31 = f0 * theMatrix->m.m01 + f1 * theMatrix->m.m11 + f2 * theMatrix->m.m21;
	theMatrix->m.m32 = f0 * theMatrix->m.m02 + f1 * theMatrix->m.m12 + f2 * theMatrix->m.m22;
	
	/* Set the part of the matrix that never changes */
	
	theMatrix->m.m03 = 0.0;
	theMatrix->m.m13 = 0.0;
	theMatrix->m.m23 = 0.0;
	theMatrix->m.m33 = 1.0;
}  /* InvertMatrix4 */

void InvertRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	static RVector3 coeffVector;	/* Fixed a PC compiler optimizer bug by making this variable a static */
	RVector3 scaleVector;
	float n1, n2;
	float r0, r1, r2;
	float f0, f1, f2;
	float xMult, xFac;
	long indexSwap;
	long i0, i1, i2;

	/* PivotGaus */

	r0 = srcMatrix.matrix[0][0];
	n1 = srcMatrix.matrix[0][1];
	n2 = srcMatrix.matrix[0][2];
	r0 = (float)FABS(r0);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r0 *= scaleVector[0] = 1.0f / MAX3(r0, n1, n2);

	r1 = srcMatrix.matrix[1][0];
	n1 = srcMatrix.matrix[1][1];
	n2 = srcMatrix.matrix[1][2];
	r1 = (float)FABS(r1);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r1 *= scaleVector[1] = 1.0f / MAX3(r1, n1, n2);

	r2 = srcMatrix.matrix[2][0];
	n1 = srcMatrix.matrix[2][1];
	n2 = srcMatrix.matrix[2][2];
	r2 = (float)FABS(r2);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r2 *= scaleVector[2] = 1.0f / MAX3(r2, n1, n2);

	if (r0 >= r1)
		if (r0 >= r2) {
			i0 = 0;
			i1 = 1;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	else
		if (r1 >= r2) {
			i0 = 1;
			i1 = 0;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	
	xFac = 1.0f / srcMatrix.matrix[i0][0];
	
	xMult = srcMatrix.matrix[i1][0] *= xFac;
	srcMatrix.matrix[i1][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i1][2] -= xMult * srcMatrix.matrix[i0][2];

	xMult = srcMatrix.matrix[i2][0] *= xFac;
	srcMatrix.matrix[i2][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i0][2];

	r1 = srcMatrix.matrix[i1][1] * scaleVector[i1];
	r2 = srcMatrix.matrix[i2][1] * scaleVector[i2];
	r1 = (float)FABS(r1);
	r2 = (float)FABS(r2);

	if (r2 > r1) {
		indexSwap = i2;
		i2 = i1;
		i1 = indexSwap;
	}

	xMult = srcMatrix.matrix[i2][1] /= srcMatrix.matrix[i1][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i1][2];

	/* Solve the matrix */
	
	f0 = 1.0f / srcMatrix.matrix[i0][0];
	f1 = 1.0f / srcMatrix.matrix[i1][1];
	f2 = 1.0f / srcMatrix.matrix[i2][2];
	
	coeffVector[0] = 1.0f;
	coeffVector[1] = 0.0f;
	coeffVector[2] = 0.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][0] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][0] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][0] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0f;
	coeffVector[1] = 1.0f;
	coeffVector[2] = 0.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][1] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][1] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][1] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0f;
	coeffVector[1] = 0.0f;
	coeffVector[2] = 1.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][2] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][2] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][2] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	/* Calculate the inverse translation */
	
	f0 = -srcMatrix.m.m30;
	f1 = -srcMatrix.m.m31;
	f2 = -srcMatrix.m.m32;
	theMatrix->m.m30 = f0 * theMatrix->m.m00 + f1 * theMatrix->m.m10 + f2 * theMatrix->m.m20;
	theMatrix->m.m31 = f0 * theMatrix->m.m01 + f1 * theMatrix->m.m11 + f2 * theMatrix->m.m21;
	theMatrix->m.m32 = f0 * theMatrix->m.m02 + f1 * theMatrix->m.m12 + f2 * theMatrix->m.m22;
	
	/* Set the part of the matrix that never changes */
	
	theMatrix->m.m03 = 0.0f;
	theMatrix->m.m13 = 0.0f;
	theMatrix->m.m23 = 0.0f;
	theMatrix->m.m33 = 1.0f;
}  /* InvertRMatrix4 */

void InvertMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	static Vector3 coeffVector;	/* Fixed a PC compiler optimizer bug by making this variable a static */
	Vector3 scaleVector;
	double n1, n2;
	double r0, r1, r2;
	double f0, f1, f2;
	double xMult, xFac;
	long indexSwap;
	long i0, i1, i2;

	/* PivotGaus */

	r0 = srcMatrix.matrix[0][0];
	n1 = srcMatrix.matrix[0][1];
	n2 = srcMatrix.matrix[0][2];
	r0 = FABS(r0);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r0 *= scaleVector[0] = 1.0 / MAX3(r0, n1, n2);

	r1 = srcMatrix.matrix[1][0];
	n1 = srcMatrix.matrix[1][1];
	n2 = srcMatrix.matrix[1][2];
	r1 = FABS(r1);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r1 *= scaleVector[1] = 1.0 / MAX3(r1, n1, n2);

	r2 = srcMatrix.matrix[2][0];
	n1 = srcMatrix.matrix[2][1];
	n2 = srcMatrix.matrix[2][2];
	r2 = FABS(r2);
	n1 = FABS(n1);
	n2 = FABS(n2);
	r2 *= scaleVector[2] = 1.0 / MAX3(r2, n1, n2);

	if (r0 >= r1)
		if (r0 >= r2) {
			i0 = 0;
			i1 = 1;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	else
		if (r1 >= r2) {
			i0 = 1;
			i1 = 0;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	
	xFac = 1.0 / srcMatrix.matrix[i0][0];
	
	xMult = srcMatrix.matrix[i1][0] *= xFac;
	srcMatrix.matrix[i1][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i1][2] -= xMult * srcMatrix.matrix[i0][2];

	xMult = srcMatrix.matrix[i2][0] *= xFac;
	srcMatrix.matrix[i2][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i0][2];

	r1 = srcMatrix.matrix[i1][1] * scaleVector[i1];
	r2 = srcMatrix.matrix[i2][1] * scaleVector[i2];
	r1 = FABS(r1);
	r2 = FABS(r2);

	if (r2 > r1) {
		indexSwap = i2;
		i2 = i1;
		i1 = indexSwap;
	}

	xMult = srcMatrix.matrix[i2][1] /= srcMatrix.matrix[i1][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i1][2];

	/* Solve the matrix */
	
	f0 = 1.0 / srcMatrix.matrix[i0][0];
	f1 = 1.0 / srcMatrix.matrix[i1][1];
	f2 = 1.0 / srcMatrix.matrix[i2][2];
	
	coeffVector[0] = 1.0;
	coeffVector[1] = 0.0;
	coeffVector[2] = 0.0;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][0] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][0] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][0] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0;
	coeffVector[1] = 1.0;
	coeffVector[2] = 0.0;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][1] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][1] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][1] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0;
	coeffVector[1] = 0.0;
	coeffVector[2] = 1.0;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][2] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][2] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][2] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;
}  /* InvertMatrix3 */

void InvertRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	static RVector3 coeffVector;	/* Fixed a PC compiler optimizer bug by making this variable a static */
	RVector3 scaleVector;
	float n1, n2;
	float r0, r1, r2;
	float f0, f1, f2;
	float xMult, xFac;
	long indexSwap;
	long i0, i1, i2;

	/* PivotGaus */

	r0 = srcMatrix.matrix[0][0];
	n1 = srcMatrix.matrix[0][1];
	n2 = srcMatrix.matrix[0][2];
	r0 = (float)FABS(r0);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r0 *= scaleVector[0] = 1.0f / MAX3(r0, n1, n2);

	r1 = srcMatrix.matrix[1][0];
	n1 = srcMatrix.matrix[1][1];
	n2 = srcMatrix.matrix[1][2];
	r1 = (float)FABS(r1);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r1 *= scaleVector[1] = 1.0f / MAX3(r1, n1, n2);

	r2 = srcMatrix.matrix[2][0];
	n1 = srcMatrix.matrix[2][1];
	n2 = srcMatrix.matrix[2][2];
	r2 = (float)FABS(r2);
	n1 = (float)FABS(n1);
	n2 = (float)FABS(n2);
	r2 *= scaleVector[2] = 1.0f / MAX3(r2, n1, n2);

	if (r0 >= r1)
		if (r0 >= r2) {
			i0 = 0;
			i1 = 1;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	else
		if (r1 >= r2) {
			i0 = 1;
			i1 = 0;
			i2 = 2;
		} else {
			i0 = 2;
			i1 = 1;
			i2 = 0;
		}
	
	xFac = 1.0f / srcMatrix.matrix[i0][0];
	
	xMult = srcMatrix.matrix[i1][0] *= xFac;
	srcMatrix.matrix[i1][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i1][2] -= xMult * srcMatrix.matrix[i0][2];

	xMult = srcMatrix.matrix[i2][0] *= xFac;
	srcMatrix.matrix[i2][1] -= xMult * srcMatrix.matrix[i0][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i0][2];

	r1 = srcMatrix.matrix[i1][1] * scaleVector[i1];
	r2 = srcMatrix.matrix[i2][1] * scaleVector[i2];
	r1 = (float)FABS(r1);
	r2 = (float)FABS(r2);

	if (r2 > r1) {
		indexSwap = i2;
		i2 = i1;
		i1 = indexSwap;
	}

	xMult = srcMatrix.matrix[i2][1] /= srcMatrix.matrix[i1][1];
	srcMatrix.matrix[i2][2] -= xMult * srcMatrix.matrix[i1][2];

	/* Solve the matrix */
	
	f0 = 1.0f / srcMatrix.matrix[i0][0];
	f1 = 1.0f / srcMatrix.matrix[i1][1];
	f2 = 1.0f / srcMatrix.matrix[i2][2];
	
	coeffVector[0] = 1.0f;
	coeffVector[1] = 0.0f;
	coeffVector[2] = 0.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][0] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][0] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][0] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0f;
	coeffVector[1] = 1.0f;
	coeffVector[2] = 0.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][1] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][1] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][1] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;

	coeffVector[0] = 0.0f;
	coeffVector[1] = 0.0f;
	coeffVector[2] = 1.0f;

	coeffVector[i1] -= srcMatrix.matrix[i1][0] * coeffVector[i0];
	coeffVector[i2] -= srcMatrix.matrix[i2][0] * coeffVector[i0] + srcMatrix.matrix[i2][1] * coeffVector[i1];

	theMatrix->matrix[2][2] = n2 = coeffVector[i2] * f2;
	theMatrix->matrix[1][2] = n1 = (coeffVector[i1] - srcMatrix.matrix[i1][2] * n2) * f1;
	theMatrix->matrix[0][2] = (coeffVector[i0] - srcMatrix.matrix[i0][1] * n1 - srcMatrix.matrix[i0][2] * n2) * f0;
}  /* InvertRMatrix3 */
#endif

void TransposeMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m01 = srcMatrix.m.m10;
	theMatrix->m.m02 = srcMatrix.m.m20;
	theMatrix->m.m03 = srcMatrix.m.m30;
	theMatrix->m.m10 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m12 = srcMatrix.m.m21;
	theMatrix->m.m13 = srcMatrix.m.m31;
	theMatrix->m.m20 = srcMatrix.m.m02;
	theMatrix->m.m21 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
	theMatrix->m.m23 = srcMatrix.m.m32;
	theMatrix->m.m30 = srcMatrix.m.m03;
	theMatrix->m.m31 = srcMatrix.m.m13;
	theMatrix->m.m32 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* TransposeMatrix4 */

void TransposeRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m01 = srcMatrix.m.m10;
	theMatrix->m.m02 = srcMatrix.m.m20;
	theMatrix->m.m03 = srcMatrix.m.m30;
	theMatrix->m.m10 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m12 = srcMatrix.m.m21;
	theMatrix->m.m13 = srcMatrix.m.m31;
	theMatrix->m.m20 = srcMatrix.m.m02;
	theMatrix->m.m21 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
	theMatrix->m.m23 = srcMatrix.m.m32;
	theMatrix->m.m30 = srcMatrix.m.m03;
	theMatrix->m.m31 = srcMatrix.m.m13;
	theMatrix->m.m32 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* TransposeRMatrix4 */

void TransposeMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m01 = srcMatrix.m.m10;
	theMatrix->m.m02 = srcMatrix.m.m20;
	theMatrix->m.m10 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m12 = srcMatrix.m.m21;
	theMatrix->m.m20 = srcMatrix.m.m02;
	theMatrix->m.m21 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
}  /* TransposeMatrix3 */

void TransposeRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m01 = srcMatrix.m.m10;
	theMatrix->m.m02 = srcMatrix.m.m20;
	theMatrix->m.m10 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m12 = srcMatrix.m.m21;
	theMatrix->m.m20 = srcMatrix.m.m02;
	theMatrix->m.m21 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
}  /* TransposeRMatrix3 */

double Det(Matrix3Ptr a)
{
	return a->m.m00 * a->m.m11 * a->m.m22 + a->m.m01 * a->m.m12 * a->m.m20 + a->m.m02 * a->m.m10 * a->m.m21 - a->m.m02 * a->m.m11 * a->m.m20 - a->m.m01 * a->m.m10 * a->m.m22 - a->m.m00 * a->m.m12 * a->m.m21;
}  /* Det */

float DetR(RMatrix3Ptr a)
{
	return a->m.m00 * a->m.m11 * a->m.m22 + a->m.m01 * a->m.m12 * a->m.m20 + a->m.m02 * a->m.m10 * a->m.m21 - a->m.m02 * a->m.m11 * a->m.m20 - a->m.m01 * a->m.m10 * a->m.m22 - a->m.m00 * a->m.m12 * a->m.m21;
}  /* DetD */

void AppendMatrix4(Matrix4Ptr theMatrix, Matrix4Ptr b)
{
	Matrix4 a = *theMatrix;

	theMatrix->m.m00 = a.m.m00 * b->m.m00 + a.m.m01 * b->m.m10 + a.m.m02 * b->m.m20 + a.m.m03 * b->m.m30;
	theMatrix->m.m01 = a.m.m00 * b->m.m01 + a.m.m01 * b->m.m11 + a.m.m02 * b->m.m21 + a.m.m03 * b->m.m31;
	theMatrix->m.m02 = a.m.m00 * b->m.m02 + a.m.m01 * b->m.m12 + a.m.m02 * b->m.m22 + a.m.m03 * b->m.m32;
	theMatrix->m.m03 = a.m.m00 * b->m.m03 + a.m.m01 * b->m.m13 + a.m.m02 * b->m.m23 + a.m.m03 * b->m.m33;
	
	theMatrix->m.m10 = a.m.m10 * b->m.m00 + a.m.m11 * b->m.m10 + a.m.m12 * b->m.m20 + a.m.m13 * b->m.m30;
	theMatrix->m.m11 = a.m.m10 * b->m.m01 + a.m.m11 * b->m.m11 + a.m.m12 * b->m.m21 + a.m.m13 * b->m.m31;
	theMatrix->m.m12 = a.m.m10 * b->m.m02 + a.m.m11 * b->m.m12 + a.m.m12 * b->m.m22 + a.m.m13 * b->m.m32;
	theMatrix->m.m13 = a.m.m10 * b->m.m03 + a.m.m11 * b->m.m13 + a.m.m12 * b->m.m23 + a.m.m13 * b->m.m33;
	
	theMatrix->m.m20 = a.m.m20 * b->m.m00 + a.m.m21 * b->m.m10 + a.m.m22 * b->m.m20 + a.m.m23 * b->m.m30;
	theMatrix->m.m21 = a.m.m20 * b->m.m01 + a.m.m21 * b->m.m11 + a.m.m22 * b->m.m21 + a.m.m23 * b->m.m31;
	theMatrix->m.m22 = a.m.m20 * b->m.m02 + a.m.m21 * b->m.m12 + a.m.m22 * b->m.m22 + a.m.m23 * b->m.m32;
	theMatrix->m.m23 = a.m.m20 * b->m.m03 + a.m.m21 * b->m.m13 + a.m.m22 * b->m.m23 + a.m.m23 * b->m.m33;
	
	theMatrix->m.m30 = a.m.m30 * b->m.m00 + a.m.m31 * b->m.m10 + a.m.m32 * b->m.m20 + a.m.m33 * b->m.m30;
	theMatrix->m.m31 = a.m.m30 * b->m.m01 + a.m.m31 * b->m.m11 + a.m.m32 * b->m.m21 + a.m.m33 * b->m.m31;
	theMatrix->m.m32 = a.m.m30 * b->m.m02 + a.m.m31 * b->m.m12 + a.m.m32 * b->m.m22 + a.m.m33 * b->m.m32;
	theMatrix->m.m33 = a.m.m30 * b->m.m03 + a.m.m31 * b->m.m13 + a.m.m32 * b->m.m23 + a.m.m33 * b->m.m33;
}  /* AppendMatrix4 */

void AppendRMatrix4(RMatrix4Ptr theMatrix, RMatrix4Ptr b)
{
	RMatrix4 a = *theMatrix;

	theMatrix->m.m00 = a.m.m00 * b->m.m00 + a.m.m01 * b->m.m10 + a.m.m02 * b->m.m20 + a.m.m03 * b->m.m30;
	theMatrix->m.m01 = a.m.m00 * b->m.m01 + a.m.m01 * b->m.m11 + a.m.m02 * b->m.m21 + a.m.m03 * b->m.m31;
	theMatrix->m.m02 = a.m.m00 * b->m.m02 + a.m.m01 * b->m.m12 + a.m.m02 * b->m.m22 + a.m.m03 * b->m.m32;
	theMatrix->m.m03 = a.m.m00 * b->m.m03 + a.m.m01 * b->m.m13 + a.m.m02 * b->m.m23 + a.m.m03 * b->m.m33;
	
	theMatrix->m.m10 = a.m.m10 * b->m.m00 + a.m.m11 * b->m.m10 + a.m.m12 * b->m.m20 + a.m.m13 * b->m.m30;
	theMatrix->m.m11 = a.m.m10 * b->m.m01 + a.m.m11 * b->m.m11 + a.m.m12 * b->m.m21 + a.m.m13 * b->m.m31;
	theMatrix->m.m12 = a.m.m10 * b->m.m02 + a.m.m11 * b->m.m12 + a.m.m12 * b->m.m22 + a.m.m13 * b->m.m32;
	theMatrix->m.m13 = a.m.m10 * b->m.m03 + a.m.m11 * b->m.m13 + a.m.m12 * b->m.m23 + a.m.m13 * b->m.m33;
	
	theMatrix->m.m20 = a.m.m20 * b->m.m00 + a.m.m21 * b->m.m10 + a.m.m22 * b->m.m20 + a.m.m23 * b->m.m30;
	theMatrix->m.m21 = a.m.m20 * b->m.m01 + a.m.m21 * b->m.m11 + a.m.m22 * b->m.m21 + a.m.m23 * b->m.m31;
	theMatrix->m.m22 = a.m.m20 * b->m.m02 + a.m.m21 * b->m.m12 + a.m.m22 * b->m.m22 + a.m.m23 * b->m.m32;
	theMatrix->m.m23 = a.m.m20 * b->m.m03 + a.m.m21 * b->m.m13 + a.m.m22 * b->m.m23 + a.m.m23 * b->m.m33;
	
	theMatrix->m.m30 = a.m.m30 * b->m.m00 + a.m.m31 * b->m.m10 + a.m.m32 * b->m.m20 + a.m.m33 * b->m.m30;
	theMatrix->m.m31 = a.m.m30 * b->m.m01 + a.m.m31 * b->m.m11 + a.m.m32 * b->m.m21 + a.m.m33 * b->m.m31;
	theMatrix->m.m32 = a.m.m30 * b->m.m02 + a.m.m31 * b->m.m12 + a.m.m32 * b->m.m22 + a.m.m33 * b->m.m32;
	theMatrix->m.m33 = a.m.m30 * b->m.m03 + a.m.m31 * b->m.m13 + a.m.m32 * b->m.m23 + a.m.m33 * b->m.m33;
}  /* AppendRMatrix4 */

void AppendMatrix3(Matrix3Ptr theMatrix, Matrix3Ptr b)
{
	Matrix3 a = *theMatrix;

	theMatrix->m.m00 = a.m.m00 * b->m.m00 + a.m.m01 * b->m.m10 + a.m.m02 * b->m.m20;
	theMatrix->m.m01 = a.m.m00 * b->m.m01 + a.m.m01 * b->m.m11 + a.m.m02 * b->m.m21;
	theMatrix->m.m02 = a.m.m00 * b->m.m02 + a.m.m01 * b->m.m12 + a.m.m02 * b->m.m22;
	
	theMatrix->m.m10 = a.m.m10 * b->m.m00 + a.m.m11 * b->m.m10 + a.m.m12 * b->m.m20;
	theMatrix->m.m11 = a.m.m10 * b->m.m01 + a.m.m11 * b->m.m11 + a.m.m12 * b->m.m21;
	theMatrix->m.m12 = a.m.m10 * b->m.m02 + a.m.m11 * b->m.m12 + a.m.m12 * b->m.m22;
	
	theMatrix->m.m20 = a.m.m20 * b->m.m00 + a.m.m21 * b->m.m10 + a.m.m22 * b->m.m20;
	theMatrix->m.m21 = a.m.m20 * b->m.m01 + a.m.m21 * b->m.m11 + a.m.m22 * b->m.m21;
	theMatrix->m.m22 = a.m.m20 * b->m.m02 + a.m.m21 * b->m.m12 + a.m.m22 * b->m.m22;
}  /* AppendMatrix3 */

void AppendRMatrix3(RMatrix3Ptr theMatrix, RMatrix3Ptr b)
{
	RMatrix3 a = *theMatrix;

	theMatrix->m.m00 = a.m.m00 * b->m.m00 + a.m.m01 * b->m.m10 + a.m.m02 * b->m.m20;
	theMatrix->m.m01 = a.m.m00 * b->m.m01 + a.m.m01 * b->m.m11 + a.m.m02 * b->m.m21;
	theMatrix->m.m02 = a.m.m00 * b->m.m02 + a.m.m01 * b->m.m12 + a.m.m02 * b->m.m22;
	
	theMatrix->m.m10 = a.m.m10 * b->m.m00 + a.m.m11 * b->m.m10 + a.m.m12 * b->m.m20;
	theMatrix->m.m11 = a.m.m10 * b->m.m01 + a.m.m11 * b->m.m11 + a.m.m12 * b->m.m21;
	theMatrix->m.m12 = a.m.m10 * b->m.m02 + a.m.m11 * b->m.m12 + a.m.m12 * b->m.m22;
	
	theMatrix->m.m20 = a.m.m20 * b->m.m00 + a.m.m21 * b->m.m10 + a.m.m22 * b->m.m20;
	theMatrix->m.m21 = a.m.m20 * b->m.m01 + a.m.m21 * b->m.m11 + a.m.m22 * b->m.m21;
	theMatrix->m.m22 = a.m.m20 * b->m.m02 + a.m.m21 * b->m.m12 + a.m.m22 * b->m.m22;
}  /* AppendRMatrix3 */

void InsertMatrix4(Matrix4Ptr theMatrix, Matrix4Ptr a)
{
	Matrix4 b = *theMatrix;

	theMatrix->m.m00 = a->m.m00 * b.m.m00 + a->m.m01 * b.m.m10 + a->m.m02 * b.m.m20 + a->m.m03 * b.m.m30;
	theMatrix->m.m10 = a->m.m10 * b.m.m00 + a->m.m11 * b.m.m10 + a->m.m12 * b.m.m20 + a->m.m13 * b.m.m30;
	theMatrix->m.m20 = a->m.m20 * b.m.m00 + a->m.m21 * b.m.m10 + a->m.m22 * b.m.m20 + a->m.m23 * b.m.m30;
	theMatrix->m.m30 = a->m.m30 * b.m.m00 + a->m.m31 * b.m.m10 + a->m.m32 * b.m.m20 + a->m.m33 * b.m.m30;
	
	theMatrix->m.m01 = a->m.m00 * b.m.m01 + a->m.m01 * b.m.m11 + a->m.m02 * b.m.m21 + a->m.m03 * b.m.m31;
	theMatrix->m.m11 = a->m.m10 * b.m.m01 + a->m.m11 * b.m.m11 + a->m.m12 * b.m.m21 + a->m.m13 * b.m.m31;
	theMatrix->m.m21 = a->m.m20 * b.m.m01 + a->m.m21 * b.m.m11 + a->m.m22 * b.m.m21 + a->m.m23 * b.m.m31;
	theMatrix->m.m31 = a->m.m30 * b.m.m01 + a->m.m31 * b.m.m11 + a->m.m32 * b.m.m21 + a->m.m33 * b.m.m31;
	
	theMatrix->m.m02 = a->m.m00 * b.m.m02 + a->m.m01 * b.m.m12 + a->m.m02 * b.m.m22 + a->m.m03 * b.m.m32;
	theMatrix->m.m12 = a->m.m10 * b.m.m02 + a->m.m11 * b.m.m12 + a->m.m12 * b.m.m22 + a->m.m13 * b.m.m32;
	theMatrix->m.m22 = a->m.m20 * b.m.m02 + a->m.m21 * b.m.m12 + a->m.m22 * b.m.m22 + a->m.m23 * b.m.m32;
	theMatrix->m.m32 = a->m.m30 * b.m.m02 + a->m.m31 * b.m.m12 + a->m.m32 * b.m.m22 + a->m.m33 * b.m.m32;
	
	theMatrix->m.m03 = a->m.m00 * b.m.m03 + a->m.m01 * b.m.m13 + a->m.m02 * b.m.m23 + a->m.m03 * b.m.m33;
	theMatrix->m.m13 = a->m.m10 * b.m.m03 + a->m.m11 * b.m.m13 + a->m.m12 * b.m.m23 + a->m.m13 * b.m.m33;
	theMatrix->m.m23 = a->m.m20 * b.m.m03 + a->m.m21 * b.m.m13 + a->m.m22 * b.m.m23 + a->m.m23 * b.m.m33;
	theMatrix->m.m33 = a->m.m30 * b.m.m03 + a->m.m31 * b.m.m13 + a->m.m32 * b.m.m23 + a->m.m33 * b.m.m33;
}  /* InsertMatrix4 */

void InsertRMatrix4(RMatrix4Ptr theMatrix, RMatrix4Ptr a)
{
	RMatrix4 b = *theMatrix;

	theMatrix->m.m00 = a->m.m00 * b.m.m00 + a->m.m01 * b.m.m10 + a->m.m02 * b.m.m20 + a->m.m03 * b.m.m30;
	theMatrix->m.m10 = a->m.m10 * b.m.m00 + a->m.m11 * b.m.m10 + a->m.m12 * b.m.m20 + a->m.m13 * b.m.m30;
	theMatrix->m.m20 = a->m.m20 * b.m.m00 + a->m.m21 * b.m.m10 + a->m.m22 * b.m.m20 + a->m.m23 * b.m.m30;
	theMatrix->m.m30 = a->m.m30 * b.m.m00 + a->m.m31 * b.m.m10 + a->m.m32 * b.m.m20 + a->m.m33 * b.m.m30;
	
	theMatrix->m.m01 = a->m.m00 * b.m.m01 + a->m.m01 * b.m.m11 + a->m.m02 * b.m.m21 + a->m.m03 * b.m.m31;
	theMatrix->m.m11 = a->m.m10 * b.m.m01 + a->m.m11 * b.m.m11 + a->m.m12 * b.m.m21 + a->m.m13 * b.m.m31;
	theMatrix->m.m21 = a->m.m20 * b.m.m01 + a->m.m21 * b.m.m11 + a->m.m22 * b.m.m21 + a->m.m23 * b.m.m31;
	theMatrix->m.m31 = a->m.m30 * b.m.m01 + a->m.m31 * b.m.m11 + a->m.m32 * b.m.m21 + a->m.m33 * b.m.m31;
	
	theMatrix->m.m02 = a->m.m00 * b.m.m02 + a->m.m01 * b.m.m12 + a->m.m02 * b.m.m22 + a->m.m03 * b.m.m32;
	theMatrix->m.m12 = a->m.m10 * b.m.m02 + a->m.m11 * b.m.m12 + a->m.m12 * b.m.m22 + a->m.m13 * b.m.m32;
	theMatrix->m.m22 = a->m.m20 * b.m.m02 + a->m.m21 * b.m.m12 + a->m.m22 * b.m.m22 + a->m.m23 * b.m.m32;
	theMatrix->m.m32 = a->m.m30 * b.m.m02 + a->m.m31 * b.m.m12 + a->m.m32 * b.m.m22 + a->m.m33 * b.m.m32;
	
	theMatrix->m.m03 = a->m.m00 * b.m.m03 + a->m.m01 * b.m.m13 + a->m.m02 * b.m.m23 + a->m.m03 * b.m.m33;
	theMatrix->m.m13 = a->m.m10 * b.m.m03 + a->m.m11 * b.m.m13 + a->m.m12 * b.m.m23 + a->m.m13 * b.m.m33;
	theMatrix->m.m23 = a->m.m20 * b.m.m03 + a->m.m21 * b.m.m13 + a->m.m22 * b.m.m23 + a->m.m23 * b.m.m33;
	theMatrix->m.m33 = a->m.m30 * b.m.m03 + a->m.m31 * b.m.m13 + a->m.m32 * b.m.m23 + a->m.m33 * b.m.m33;
}  /* InsertRMatrix4 */

void InsertMatrix3(Matrix3Ptr theMatrix, Matrix3Ptr a)
{
	Matrix3 b = *theMatrix;

	theMatrix->m.m00 = a->m.m00 * b.m.m00 + a->m.m01 * b.m.m10 + a->m.m02 * b.m.m20;
	theMatrix->m.m10 = a->m.m10 * b.m.m00 + a->m.m11 * b.m.m10 + a->m.m12 * b.m.m20;
	theMatrix->m.m20 = a->m.m20 * b.m.m00 + a->m.m21 * b.m.m10 + a->m.m22 * b.m.m20;
	
	theMatrix->m.m01 = a->m.m00 * b.m.m01 + a->m.m01 * b.m.m11 + a->m.m02 * b.m.m21;
	theMatrix->m.m11 = a->m.m10 * b.m.m01 + a->m.m11 * b.m.m11 + a->m.m12 * b.m.m21;
	theMatrix->m.m21 = a->m.m20 * b.m.m01 + a->m.m21 * b.m.m11 + a->m.m22 * b.m.m21;
	
	theMatrix->m.m02 = a->m.m00 * b.m.m02 + a->m.m01 * b.m.m12 + a->m.m02 * b.m.m22;
	theMatrix->m.m12 = a->m.m10 * b.m.m02 + a->m.m11 * b.m.m12 + a->m.m12 * b.m.m22;
	theMatrix->m.m22 = a->m.m20 * b.m.m02 + a->m.m21 * b.m.m12 + a->m.m22 * b.m.m22;
}  /* InsertMatrix3 */

void InsertRMatrix3(RMatrix3Ptr theMatrix, RMatrix3Ptr a)
{
	RMatrix3 b = *theMatrix;

	theMatrix->m.m00 = a->m.m00 * b.m.m00 + a->m.m01 * b.m.m10 + a->m.m02 * b.m.m20;
	theMatrix->m.m10 = a->m.m10 * b.m.m00 + a->m.m11 * b.m.m10 + a->m.m12 * b.m.m20;
	theMatrix->m.m20 = a->m.m20 * b.m.m00 + a->m.m21 * b.m.m10 + a->m.m22 * b.m.m20;
	
	theMatrix->m.m01 = a->m.m00 * b.m.m01 + a->m.m01 * b.m.m11 + a->m.m02 * b.m.m21;
	theMatrix->m.m11 = a->m.m10 * b.m.m01 + a->m.m11 * b.m.m11 + a->m.m12 * b.m.m21;
	theMatrix->m.m21 = a->m.m20 * b.m.m01 + a->m.m21 * b.m.m11 + a->m.m22 * b.m.m21;
	
	theMatrix->m.m02 = a->m.m00 * b.m.m02 + a->m.m01 * b.m.m12 + a->m.m02 * b.m.m22;
	theMatrix->m.m12 = a->m.m10 * b.m.m02 + a->m.m11 * b.m.m12 + a->m.m12 * b.m.m22;
	theMatrix->m.m22 = a->m.m20 * b.m.m02 + a->m.m21 * b.m.m12 + a->m.m22 * b.m.m22;
}  /* InsertRMatrix3 */

void NormalMatrix4(Matrix4Ptr theMatrix)
{
	InvertMatrix4(theMatrix, *theMatrix);
	TransposeMatrix4(theMatrix, *theMatrix);
}  /* NormalMatrix4 */

void NormalRMatrix4(RMatrix4Ptr theMatrix)
{
	InvertRMatrix4(theMatrix, *theMatrix);
	TransposeRMatrix4(theMatrix, *theMatrix);
}  /* NormalRMatrix4 */

void NormalMatrix3(Matrix3Ptr theMatrix)
{
	InvertMatrix3(theMatrix, *theMatrix);
	TransposeMatrix3(theMatrix, *theMatrix);
}  /* NormalMatrix3 */

void NormalRMatrix3(RMatrix3Ptr theMatrix)
{
	InvertRMatrix3(theMatrix, *theMatrix);
	TransposeRMatrix3(theMatrix, *theMatrix);
}  /* NormalRMatrix3 */

void NormalizeMatrix3(Matrix3Ptr theMatrix)
{
	double m00 = theMatrix->m.m00;
	double m01 = theMatrix->m.m01;
	double m02 = theMatrix->m.m02;
	double m10 = theMatrix->m.m10;
	double m11 = theMatrix->m.m11;
	double m12 = theMatrix->m.m12;
	double m20 = theMatrix->m.m20;
	double m21 = theMatrix->m.m21;
	double m22 = theMatrix->m.m22;
	double distance;
	
	distance = SQR(m00) + SQR(m01) + SQR(m02);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m00 = 1.0;
			m01 = 0.0;
			m02 = 0.0;
		} else {
			distance = 1.0 / SQRT(distance);
			m00 *= distance;
			m01 *= distance;
			m02 *= distance;
		}
	
	distance = SQR(m10) + SQR(m11) + SQR(m12);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m10 = 1.0;
			m11 = 0.0;
			m12 = 0.0;
		} else {
			distance = 1.0 / SQRT(distance);
			m10 *= distance;
			m11 *= distance;
			m12 *= distance;
		}
	
	distance = SQR(m20) + SQR(m21) + SQR(m22);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m20 = 1.0;
			m21 = 0.0;
			m22 = 0.0;
		} else {
			distance = 1.0 / SQRT(distance);
			m20 *= distance;
			m21 *= distance;
			m22 *= distance;
		}

	theMatrix->m.m00 = m00;
	theMatrix->m.m01 = m01;
	theMatrix->m.m02 = m02;
	theMatrix->m.m10 = m10;
	theMatrix->m.m11 = m11;
	theMatrix->m.m12 = m12;
	theMatrix->m.m20 = m20;
	theMatrix->m.m21 = m21;
	theMatrix->m.m22 = m22;
}  /* NormalizeMatrix3 */

void NormalizeRMatrix3(RMatrix3Ptr theMatrix)
{
	float m00 = theMatrix->m.m00;
	float m01 = theMatrix->m.m01;
	float m02 = theMatrix->m.m02;
	float m10 = theMatrix->m.m10;
	float m11 = theMatrix->m.m11;
	float m12 = theMatrix->m.m12;
	float m20 = theMatrix->m.m20;
	float m21 = theMatrix->m.m21;
	float m22 = theMatrix->m.m22;
	float distance;
	
	distance = SQR(m00) + SQR(m01) + SQR(m02);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m00 = 1.0f;
			m01 = 0.0f;
			m02 = 0.0f;
		} else {
			distance = 1.0f / (float)SQRT(distance);
			m00 *= distance;
			m01 *= distance;
			m02 *= distance;
		}
	
	distance = SQR(m10) + SQR(m11) + SQR(m12);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m10 = 1.0f;
			m11 = 0.0f;
			m12 = 0.0f;
		} else {
			distance = 1.0f / (float)SQRT(distance);
			m10 *= distance;
			m11 *= distance;
			m12 *= distance;
		}
	
	distance = SQR(m20) + SQR(m21) + SQR(m22);
	if (!EQUAL1(distance)) 
		if (EQUAL0(distance)) {
			m20 = 1.0f;
			m21 = 0.0f;
			m22 = 0.0f;
		} else {
			distance = 1.0f / (float)SQRT(distance);
			m20 *= distance;
			m21 *= distance;
			m22 *= distance;
		}

	theMatrix->m.m00 = m00;
	theMatrix->m.m01 = m01;
	theMatrix->m.m02 = m02;
	theMatrix->m.m10 = m10;
	theMatrix->m.m11 = m11;
	theMatrix->m.m12 = m12;
	theMatrix->m.m20 = m20;
	theMatrix->m.m21 = m21;
	theMatrix->m.m22 = m22;
}  /* NormalizeRMatrix3 */

void SwapXYMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m01;
	theMatrix->m.m10 = srcMatrix.m.m11;
	theMatrix->m.m20 = srcMatrix.m.m21;
	theMatrix->m.m30 = srcMatrix.m.m31;

	theMatrix->m.m01 = srcMatrix.m.m00;
	theMatrix->m.m11 = srcMatrix.m.m10;
	theMatrix->m.m21 = srcMatrix.m.m20;
	theMatrix->m.m31 = srcMatrix.m.m30;

	theMatrix->m.m02 = srcMatrix.m.m02;
	theMatrix->m.m12 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
	theMatrix->m.m32 = srcMatrix.m.m32;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapXYMatrix4 */

void SwapXYRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m01;
	theMatrix->m.m10 = srcMatrix.m.m11;
	theMatrix->m.m20 = srcMatrix.m.m21;
	theMatrix->m.m30 = srcMatrix.m.m31;

	theMatrix->m.m01 = srcMatrix.m.m00;
	theMatrix->m.m11 = srcMatrix.m.m10;
	theMatrix->m.m21 = srcMatrix.m.m20;
	theMatrix->m.m31 = srcMatrix.m.m30;

	theMatrix->m.m02 = srcMatrix.m.m02;
	theMatrix->m.m12 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
	theMatrix->m.m32 = srcMatrix.m.m32;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapXYRMatrix4 */

void SwapXYMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m01;
	theMatrix->m.m10 = srcMatrix.m.m11;
	theMatrix->m.m20 = srcMatrix.m.m21;

	theMatrix->m.m01 = srcMatrix.m.m00;
	theMatrix->m.m11 = srcMatrix.m.m10;
	theMatrix->m.m21 = srcMatrix.m.m20;

	theMatrix->m.m02 = srcMatrix.m.m02;
	theMatrix->m.m12 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
}  /* SwapXYMatrix3 */

void SwapXYRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m01;
	theMatrix->m.m10 = srcMatrix.m.m11;
	theMatrix->m.m20 = srcMatrix.m.m21;

	theMatrix->m.m01 = srcMatrix.m.m00;
	theMatrix->m.m11 = srcMatrix.m.m10;
	theMatrix->m.m21 = srcMatrix.m.m20;

	theMatrix->m.m02 = srcMatrix.m.m02;
	theMatrix->m.m12 = srcMatrix.m.m12;
	theMatrix->m.m22 = srcMatrix.m.m22;
}  /* SwapXYRMatrix3 */

void SwapXZMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m02;
	theMatrix->m.m10 = srcMatrix.m.m12;
	theMatrix->m.m20 = srcMatrix.m.m22;
	theMatrix->m.m30 = srcMatrix.m.m32;

	theMatrix->m.m01 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m21 = srcMatrix.m.m21;
	theMatrix->m.m31 = srcMatrix.m.m31;

	theMatrix->m.m02 = srcMatrix.m.m00;
	theMatrix->m.m12 = srcMatrix.m.m10;
	theMatrix->m.m22 = srcMatrix.m.m20;
	theMatrix->m.m32 = srcMatrix.m.m30;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapXZMatrix4 */

void SwapXZRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m02;
	theMatrix->m.m10 = srcMatrix.m.m12;
	theMatrix->m.m20 = srcMatrix.m.m22;
	theMatrix->m.m30 = srcMatrix.m.m32;

	theMatrix->m.m01 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m21 = srcMatrix.m.m21;
	theMatrix->m.m31 = srcMatrix.m.m31;

	theMatrix->m.m02 = srcMatrix.m.m00;
	theMatrix->m.m12 = srcMatrix.m.m10;
	theMatrix->m.m22 = srcMatrix.m.m20;
	theMatrix->m.m32 = srcMatrix.m.m30;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapXZRMatrix4 */

void SwapXZMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m02;
	theMatrix->m.m10 = srcMatrix.m.m12;
	theMatrix->m.m20 = srcMatrix.m.m22;

	theMatrix->m.m01 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m21 = srcMatrix.m.m21;

	theMatrix->m.m02 = srcMatrix.m.m00;
	theMatrix->m.m12 = srcMatrix.m.m10;
	theMatrix->m.m22 = srcMatrix.m.m20;
}  /* SwapXZMatrix3 */

void SwapXZRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m02;
	theMatrix->m.m10 = srcMatrix.m.m12;
	theMatrix->m.m20 = srcMatrix.m.m22;

	theMatrix->m.m01 = srcMatrix.m.m01;
	theMatrix->m.m11 = srcMatrix.m.m11;
	theMatrix->m.m21 = srcMatrix.m.m21;

	theMatrix->m.m02 = srcMatrix.m.m00;
	theMatrix->m.m12 = srcMatrix.m.m10;
	theMatrix->m.m22 = srcMatrix.m.m20;
}  /* SwapXZRMatrix3 */

void SwapYZMatrix4(Matrix4Ptr theMatrix, Matrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m10 = srcMatrix.m.m10;
	theMatrix->m.m20 = srcMatrix.m.m20;
	theMatrix->m.m30 = srcMatrix.m.m30;

	theMatrix->m.m01 = srcMatrix.m.m02;
	theMatrix->m.m11 = srcMatrix.m.m12;
	theMatrix->m.m21 = srcMatrix.m.m22;
	theMatrix->m.m31 = srcMatrix.m.m32;

	theMatrix->m.m02 = srcMatrix.m.m01;
	theMatrix->m.m12 = srcMatrix.m.m11;
	theMatrix->m.m22 = srcMatrix.m.m21;
	theMatrix->m.m32 = srcMatrix.m.m31;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapYZMatrix4 */

void SwapYZRMatrix4(RMatrix4Ptr theMatrix, RMatrix4 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m10 = srcMatrix.m.m10;
	theMatrix->m.m20 = srcMatrix.m.m20;
	theMatrix->m.m30 = srcMatrix.m.m30;

	theMatrix->m.m01 = srcMatrix.m.m02;
	theMatrix->m.m11 = srcMatrix.m.m12;
	theMatrix->m.m21 = srcMatrix.m.m22;
	theMatrix->m.m31 = srcMatrix.m.m32;

	theMatrix->m.m02 = srcMatrix.m.m01;
	theMatrix->m.m12 = srcMatrix.m.m11;
	theMatrix->m.m22 = srcMatrix.m.m21;
	theMatrix->m.m32 = srcMatrix.m.m31;

	theMatrix->m.m03 = srcMatrix.m.m03;
	theMatrix->m.m13 = srcMatrix.m.m13;
	theMatrix->m.m23 = srcMatrix.m.m23;
	theMatrix->m.m33 = srcMatrix.m.m33;
}  /* SwapYZRMatrix4 */

void SwapYZMatrix3(Matrix3Ptr theMatrix, Matrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m10 = srcMatrix.m.m10;
	theMatrix->m.m20 = srcMatrix.m.m20;

	theMatrix->m.m01 = srcMatrix.m.m02;
	theMatrix->m.m11 = srcMatrix.m.m12;
	theMatrix->m.m21 = srcMatrix.m.m22;

	theMatrix->m.m02 = srcMatrix.m.m01;
	theMatrix->m.m12 = srcMatrix.m.m11;
	theMatrix->m.m22 = srcMatrix.m.m21;
}  /* SwapYZMatrix3 */

void SwapYZRMatrix3(RMatrix3Ptr theMatrix, RMatrix3 srcMatrix)
{
	theMatrix->m.m00 = srcMatrix.m.m00;
	theMatrix->m.m10 = srcMatrix.m.m10;
	theMatrix->m.m20 = srcMatrix.m.m20;

	theMatrix->m.m01 = srcMatrix.m.m02;
	theMatrix->m.m11 = srcMatrix.m.m12;
	theMatrix->m.m21 = srcMatrix.m.m22;

	theMatrix->m.m02 = srcMatrix.m.m01;
	theMatrix->m.m12 = srcMatrix.m.m11;
	theMatrix->m.m22 = srcMatrix.m.m21;
}  /* SwapYZRMatrix3 */

void TranslateMatrix4(Matrix4Ptr theMatrix, double tx, double ty, double tz)
{
	theMatrix->m.m00 += theMatrix->m.m03 * tx;
	theMatrix->m.m01 += theMatrix->m.m03 * ty;
	theMatrix->m.m02 += theMatrix->m.m03 * tz;
	theMatrix->m.m10 += theMatrix->m.m13 * tx;
	theMatrix->m.m11 += theMatrix->m.m13 * ty;
	theMatrix->m.m12 += theMatrix->m.m13 * tz;
	theMatrix->m.m20 += theMatrix->m.m23 * tx;
	theMatrix->m.m21 += theMatrix->m.m23 * ty;
	theMatrix->m.m22 += theMatrix->m.m23 * tz;
	theMatrix->m.m30 += theMatrix->m.m33 * tx;
	theMatrix->m.m31 += theMatrix->m.m33 * ty;
	theMatrix->m.m32 += theMatrix->m.m33 * tz;
}  /* TranslateMatrix4 */

void TranslateRMatrix4(RMatrix4Ptr theMatrix, float tx, float ty, float tz)
{
	theMatrix->m.m00 += theMatrix->m.m03 * tx;
	theMatrix->m.m01 += theMatrix->m.m03 * ty;
	theMatrix->m.m02 += theMatrix->m.m03 * tz;
	theMatrix->m.m10 += theMatrix->m.m13 * tx;
	theMatrix->m.m11 += theMatrix->m.m13 * ty;
	theMatrix->m.m12 += theMatrix->m.m13 * tz;
	theMatrix->m.m20 += theMatrix->m.m23 * tx;
	theMatrix->m.m21 += theMatrix->m.m23 * ty;
	theMatrix->m.m22 += theMatrix->m.m23 * tz;
	theMatrix->m.m30 += theMatrix->m.m33 * tx;
	theMatrix->m.m31 += theMatrix->m.m33 * ty;
	theMatrix->m.m32 += theMatrix->m.m33 * tz;
}  /* TranslateRMatrix4 */

void XRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m01;
	double temp1 = theMatrix->m.m02;
	
	theMatrix->m.m01 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m02 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m11;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m11 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m12 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m21;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m21 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m22 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m31;
	temp1 = theMatrix->m.m32;
	theMatrix->m.m31 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m32 = temp0 * sinTheta + temp1 * cosTheta;
}  /* XRotateMatrix4 */

void XRotateRMatrix4(RMatrix4Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m01;
	float temp1 = theMatrix->m.m02;
	
	theMatrix->m.m01 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m02 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m11;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m11 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m12 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m21;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m21 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m22 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m31;
	temp1 = theMatrix->m.m32;
	theMatrix->m.m31 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m32 = temp0 * sinTheta + temp1 * cosTheta;
}  /* XRotateRMatrix4 */

void XRotateMatrix3(Matrix3Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m01;
	double temp1 = theMatrix->m.m02;
	
	theMatrix->m.m01 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m02 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m11;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m11 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m12 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m21;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m21 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m22 = temp0 * sinTheta + temp1 * cosTheta;
}  /* XRotateMatrix3 */

void XRotateRMatrix3(RMatrix3Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m01;
	float temp1 = theMatrix->m.m02;
	
	theMatrix->m.m01 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m02 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m11;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m11 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m12 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m21;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m21 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m22 = temp0 * sinTheta + temp1 * cosTheta;
}  /* XRotateRMatrix3 */

void YRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m00;
	double temp1 = theMatrix->m.m02;

	theMatrix->m.m00 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m02 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m10 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m12 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m20 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m22 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m30;
	temp1 = theMatrix->m.m32;
	theMatrix->m.m30 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m32 = temp1 * cosTheta - temp0 * sinTheta;
}  /* YRotateMatrix4 */

void YRotateRMatrix4(RMatrix4Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m00;
	float temp1 = theMatrix->m.m02;

	theMatrix->m.m00 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m02 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m10 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m12 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m20 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m22 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m30;
	temp1 = theMatrix->m.m32;
	theMatrix->m.m30 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m32 = temp1 * cosTheta - temp0 * sinTheta;
}  /* YRotateRMatrix4 */

void YRotateMatrix3(Matrix3Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m00;
	double temp1 = theMatrix->m.m02;

	theMatrix->m.m00 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m02 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m10 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m12 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m20 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m22 = temp1 * cosTheta - temp0 * sinTheta;
}  /* YRotateMatrix3 */

void YRotateRMatrix3(RMatrix3Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m00;
	float temp1 = theMatrix->m.m02;
	theMatrix->m.m00 = temp0 * cosTheta + temp1 * sinTheta;

	theMatrix->m.m02 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m12;
	theMatrix->m.m10 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m12 = temp1 * cosTheta - temp0 * sinTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m22;
	theMatrix->m.m20 = temp0 * cosTheta + temp1 * sinTheta;
	theMatrix->m.m22 = temp1 * cosTheta - temp0 * sinTheta;
}  /* YRotateRMatrix3 */

void ZRotateMatrix4(Matrix4Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m00;
	double temp1 = theMatrix->m.m01;

	theMatrix->m.m00 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m01 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m11;
	theMatrix->m.m10 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m11 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m21;
	theMatrix->m.m20 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m21 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m30;
	temp1 = theMatrix->m.m31;
	theMatrix->m.m30 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m31 = temp0 * sinTheta + temp1 * cosTheta;
}  /* ZRotateMatrix4 */

void ZRotateRMatrix4(RMatrix4Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m00;
	float temp1 = theMatrix->m.m01;

	theMatrix->m.m00 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m01 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m11;
	theMatrix->m.m10 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m11 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m21;
	theMatrix->m.m20 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m21 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m30;
	temp1 = theMatrix->m.m31;
	theMatrix->m.m30 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m31 = temp0 * sinTheta + temp1 * cosTheta;
}  /* ZRotateRMatrix4 */

void ZRotateMatrix3(Matrix3Ptr theMatrix, double sinTheta, double cosTheta)
{
	double temp0 = theMatrix->m.m00;
	double temp1 = theMatrix->m.m01;

	theMatrix->m.m00 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m01 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m11;
	theMatrix->m.m10 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m11 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m21;
	theMatrix->m.m20 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m21 = temp0 * sinTheta + temp1 * cosTheta;
}  /* ZRotateMatrix3 */

void ZRotateRMatrix3(RMatrix3Ptr theMatrix, float sinTheta, float cosTheta)
{
	float temp0 = theMatrix->m.m00;
	float temp1 = theMatrix->m.m01;

	theMatrix->m.m00 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m01 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m10;
	temp1 = theMatrix->m.m11;
	theMatrix->m.m10 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m11 = temp0 * sinTheta + temp1 * cosTheta;
	temp0 = theMatrix->m.m20;
	temp1 = theMatrix->m.m21;
	theMatrix->m.m20 = temp0 * cosTheta - temp1 * sinTheta;
	theMatrix->m.m21 = temp0 * sinTheta + temp1 * cosTheta;
}  /* ZRotateRMatrix3 */

void RotateMatrix4(Matrix4Ptr theMatrix, double x, double y, double z, double theta)
{
	Matrix4 newMatrix;
	double t = SQR(x) + SQR(y) + SQR(z);

	if (t > 0.0) {
		double s = SIN(theta);
		double c = COS(theta);
		double tx, ty, txy, txz, tyz, sx, sy, sz;

		t = 1.0 / SQRT(t);
		x *= t;
		y *= t;
		z *= t;

		t = 1.0 - c;
		tx = t * x;
		ty = t * y;
		
		txy = tx * y;
		txz = tx * z;
		tyz = ty * z;
		
		sx = s * x;
		sy = s * y;
		sz = s * z;
		
		newMatrix.m.m00 = tx * x + c;
		newMatrix.m.m01 = txy + sz;
		newMatrix.m.m02 = txz - sy;
		newMatrix.m.m03 = 0.0;
		newMatrix.m.m10 = txy - sz;
		newMatrix.m.m11 = ty * y + c;
		newMatrix.m.m12 = tyz + sx;
		newMatrix.m.m13 = 0.0;
		newMatrix.m.m20 = txz + sy;
		newMatrix.m.m21 = tyz - sx;
		newMatrix.m.m22 = t * z * z + c;
		newMatrix.m.m23 = 0.0;
		newMatrix.m.m30 = 0.0;
		newMatrix.m.m31 = 0.0;
		newMatrix.m.m32 = 0.0;
		newMatrix.m.m33 = 1.0;
		
		AppendMatrix4(theMatrix, &newMatrix);
	}
}  /* RotateMatrix4 */

void RotateRMatrix4(RMatrix4Ptr theMatrix, float x, float y, float z, float theta)
{
	RMatrix4 newMatrix;
	float t = SQR(x) + SQR(y) + SQR(z);

	if (t > 0.0) {
		float s = (float)SIN(theta);
		float c = (float)COS(theta);
		float tx, ty, txy, txz, tyz, sx, sy, sz;

		t = 1.0f / (float)SQRT(t);
		x *= t;
		y *= t;
		z *= t;

		t = 1.0f - c;
		tx = t * x;
		ty = t * y;
		
		txy = tx * y;
		txz = tx * z;
		tyz = ty * z;
		
		sx = s * x;
		sy = s * y;
		sz = s * z;
		
		newMatrix.m.m00 = tx * x + c;
		newMatrix.m.m01 = txy + sz;
		newMatrix.m.m02 = txz - sy;
		newMatrix.m.m03 = 0.0f;
		newMatrix.m.m10 = txy - sz;
		newMatrix.m.m11 = ty * y + c;
		newMatrix.m.m12 = tyz + sx;
		newMatrix.m.m13 = 0.0f;
		newMatrix.m.m20 = txz + sy;
		newMatrix.m.m21 = tyz - sx;
		newMatrix.m.m22 = t * z * z + c;
		newMatrix.m.m23 = 0.0f;
		newMatrix.m.m30 = 0.0f;
		newMatrix.m.m31 = 0.0f;
		newMatrix.m.m32 = 0.0f;
		newMatrix.m.m33 = 1.0f;
		
		AppendRMatrix4(theMatrix, &newMatrix);
	}
}  /* RotateRMatrix4 */

void RotateMatrix3(Matrix3Ptr theMatrix, double x, double y, double z, double theta)
{
	Matrix3 newMatrix;
	double t = SQR(x) + SQR(y) + SQR(z);

	if (t > 0.0) {
		double s = SIN(theta);
		double c = COS(theta);
		double tx, ty, txy, txz, tyz, sx, sy, sz;

		t = 1.0 / SQRT(t);
		x *= t;
		y *= t;
		z *= t;

		t = 1.0 - c;
		tx = t * x;
		ty = t * y;
		
		txy = tx * y;
		txz = tx * z;
		tyz = ty * z;
		
		sx = s * x;
		sy = s * y;
		sz = s * z;
		
		newMatrix.m.m00 = tx * x + c;
		newMatrix.m.m01 = txy + sz;
		newMatrix.m.m02 = txz - sy;
		newMatrix.m.m10 = txy - sz;
		newMatrix.m.m11 = ty * y + c;
		newMatrix.m.m12 = tyz + sx;
		newMatrix.m.m20 = txz + sy;
		newMatrix.m.m21 = tyz - sx;
		newMatrix.m.m22 = t * z * z + c;
		
		AppendMatrix3(theMatrix, &newMatrix);
	}
}  /* RotateMatrix3 */

void RotateRMatrix3(RMatrix3Ptr theMatrix, float x, float y, float z, float theta)
{
	RMatrix3 newMatrix;
	float t = SQR(x) + SQR(y) + SQR(z);

	if (t > 0.0) {
		float s = (float)SIN(theta);
		float c = (float)COS(theta);
		float tx, ty, txy, txz, tyz, sx, sy, sz;

		t = 1.0f / (float)SQRT(t);
		x *= t;
		y *= t;
		z *= t;

		t = 1.0f - c;
		tx = t * x;
		ty = t * y;
		
		txy = tx * y;
		txz = tx * z;
		tyz = ty * z;
		
		sx = s * x;
		sy = s * y;
		sz = s * z;
		
		newMatrix.m.m00 = tx * x + c;
		newMatrix.m.m01 = txy + sz;
		newMatrix.m.m02 = txz - sy;
		newMatrix.m.m10 = txy - sz;
		newMatrix.m.m11 = ty * y + c;
		newMatrix.m.m12 = tyz + sx;
		newMatrix.m.m20 = txz + sy;
		newMatrix.m.m21 = tyz - sx;
		newMatrix.m.m22 = t * z * z + c;
		
		AppendRMatrix3(theMatrix, &newMatrix);
	}
}  /* RotateRMatrix3 */

void ScaleMatrix4(Matrix4Ptr theMatrix, double sx, double sy, double sz)
{
	if (FABS(sx) < matrix_minScale)
		sx = (sx >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sy) < matrix_minScale)
		sy = (sy >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sz) < matrix_minScale)
		sz = (sz >= 0.0 ? matrix_minScale : -matrix_minScale);
	theMatrix->m.m00 *= sx;
	theMatrix->m.m01 *= sy;
	theMatrix->m.m02 *= sz;
	theMatrix->m.m10 *= sx;
	theMatrix->m.m11 *= sy;
	theMatrix->m.m12 *= sz;
	theMatrix->m.m20 *= sx;
	theMatrix->m.m21 *= sy;
	theMatrix->m.m22 *= sz;
	theMatrix->m.m30 *= sx;
	theMatrix->m.m31 *= sy;
	theMatrix->m.m32 *= sz;
}  /* ScaleMatrix4 */

void ScaleRMatrix4(RMatrix4Ptr theMatrix, float sx, float sy, float sz)
{
	if (FABS(sx) < matrix_minScale)
		sx = (float)(sx >= 0.0f ? matrix_minScale : -matrix_minScale);
	if (FABS(sy) < matrix_minScale)
		sy = (float)(sy >= 0.0f ? matrix_minScale : -matrix_minScale);
	if (FABS(sz) < matrix_minScale)
		sz = (float)(sz >= 0.0f ? matrix_minScale : -matrix_minScale);
	theMatrix->m.m00 *= sx;
	theMatrix->m.m01 *= sy;
	theMatrix->m.m02 *= sz;
	theMatrix->m.m10 *= sx;
	theMatrix->m.m11 *= sy;
	theMatrix->m.m12 *= sz;
	theMatrix->m.m20 *= sx;
	theMatrix->m.m21 *= sy;
	theMatrix->m.m22 *= sz;
	theMatrix->m.m30 *= sx;
	theMatrix->m.m31 *= sy;
	theMatrix->m.m32 *= sz;
}  /* ScaleRMatrix4 */

void ScaleMatrix3(Matrix3Ptr theMatrix, double sx, double sy, double sz)
{
	if (FABS(sx) < matrix_minScale)
		sx = (sx >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sy) < matrix_minScale)
		sy = (sy >= 0.0 ? matrix_minScale : -matrix_minScale);
	if (FABS(sz) < matrix_minScale)
		sz = (sz >= 0.0 ? matrix_minScale : -matrix_minScale);
	theMatrix->m.m00 *= sx;
	theMatrix->m.m01 *= sy;
	theMatrix->m.m02 *= sz;
	theMatrix->m.m10 *= sx;
	theMatrix->m.m11 *= sy;
	theMatrix->m.m12 *= sz;
	theMatrix->m.m20 *= sx;
	theMatrix->m.m21 *= sy;
	theMatrix->m.m22 *= sz;
}  /* ScaleMatrix3 */

void ScaleRMatrix3(RMatrix3Ptr theMatrix, float sx, float sy, float sz)
{
	if (FABS(sx) < matrix_minScale)
		sx = (float)(sx >= 0.0f ? matrix_minScale : -matrix_minScale);
	if (FABS(sy) < matrix_minScale)
		sy = (float)(sy >= 0.0f ? matrix_minScale : -matrix_minScale);
	if (FABS(sz) < matrix_minScale)
		sz = (float)(sz >= 0.0f ? matrix_minScale : -matrix_minScale);
	theMatrix->m.m00 *= sx;
	theMatrix->m.m01 *= sy;
	theMatrix->m.m02 *= sz;
	theMatrix->m.m10 *= sx;
	theMatrix->m.m11 *= sy;
	theMatrix->m.m12 *= sz;
	theMatrix->m.m20 *= sx;
	theMatrix->m.m21 *= sy;
	theMatrix->m.m22 *= sz;
}  /* ScaleRMatrix3 */

void NegateXMatrix4(Matrix4Ptr theMatrix)
{
	theMatrix->m.m00 = -theMatrix->m.m00;
	theMatrix->m.m10 = -theMatrix->m.m10;
	theMatrix->m.m20 = -theMatrix->m.m20;
	theMatrix->m.m30 = -theMatrix->m.m30;
}  /* NegateXMatrix4 */

void NegateXRMatrix4(RMatrix4Ptr theMatrix)
{
	theMatrix->m.m00 = -theMatrix->m.m00;
	theMatrix->m.m10 = -theMatrix->m.m10;
	theMatrix->m.m20 = -theMatrix->m.m20;
	theMatrix->m.m30 = -theMatrix->m.m30;
}  /* NegateXRMatrix4 */

void NegateXMatrix3(Matrix3Ptr theMatrix)
{
	theMatrix->m.m00 = -theMatrix->m.m00;
	theMatrix->m.m10 = -theMatrix->m.m10;
	theMatrix->m.m20 = -theMatrix->m.m20;
}  /* NegateXMatrix3 */

void NegateXRMatrix3(RMatrix3Ptr theMatrix)
{
	theMatrix->m.m00 = -theMatrix->m.m00;
	theMatrix->m.m10 = -theMatrix->m.m10;
	theMatrix->m.m20 = -theMatrix->m.m20;
}  /* NegateXRMatrix3 */

void NegateYMatrix4(Matrix4Ptr theMatrix)
{
	theMatrix->m.m01 = -theMatrix->m.m01;
	theMatrix->m.m11 = -theMatrix->m.m11;
	theMatrix->m.m21 = -theMatrix->m.m21;
	theMatrix->m.m31 = -theMatrix->m.m31;
}  /* NegateYMatrix4 */

void NegateYRMatrix4(RMatrix4Ptr theMatrix)
{
	theMatrix->m.m01 = -theMatrix->m.m01;
	theMatrix->m.m11 = -theMatrix->m.m11;
	theMatrix->m.m21 = -theMatrix->m.m21;
	theMatrix->m.m31 = -theMatrix->m.m31;
}  /* NegateYRMatrix4 */

void NegateYMatrix3(Matrix3Ptr theMatrix)
{
	theMatrix->m.m01 = -theMatrix->m.m01;
	theMatrix->m.m11 = -theMatrix->m.m11;
	theMatrix->m.m21 = -theMatrix->m.m21;
}  /* NegateYMatrix3 */

void NegateYRMatrix3(RMatrix3Ptr theMatrix)
{
	theMatrix->m.m01 = -theMatrix->m.m01;
	theMatrix->m.m11 = -theMatrix->m.m11;
	theMatrix->m.m21 = -theMatrix->m.m21;
}  /* NegateYRMatrix3 */

void NegateZMatrix4(Matrix4Ptr theMatrix)
{
	theMatrix->m.m02 = -theMatrix->m.m02;
	theMatrix->m.m12 = -theMatrix->m.m12;
	theMatrix->m.m22 = -theMatrix->m.m22;
	theMatrix->m.m32 = -theMatrix->m.m32;
}  /* NegateZMatrix4 */

void NegateZRMatrix4(RMatrix4Ptr theMatrix)
{
	theMatrix->m.m02 = -theMatrix->m.m02;
	theMatrix->m.m12 = -theMatrix->m.m12;
	theMatrix->m.m22 = -theMatrix->m.m22;
	theMatrix->m.m32 = -theMatrix->m.m32;
}  /* NegateZRMatrix4 */

void NegateZMatrix3(Matrix3Ptr theMatrix)
{
	theMatrix->m.m02 = -theMatrix->m.m02;
	theMatrix->m.m12 = -theMatrix->m.m12;
	theMatrix->m.m22 = -theMatrix->m.m22;
}  /* NegateZMatrix3 */

void NegateZRMatrix3(RMatrix3Ptr theMatrix)
{
	theMatrix->m.m02 = -theMatrix->m.m02;
	theMatrix->m.m12 = -theMatrix->m.m12;
	theMatrix->m.m22 = -theMatrix->m.m22;
}  /* NegateZRMatrix3 */

void SkewMatrix4(Matrix4Ptr theMatrix, double skew)
{
	double temp;
	
	temp = skew * theMatrix->m.m02;
	theMatrix->m.m00 += temp;
	theMatrix->m.m01 += temp;
	
	temp = skew * theMatrix->m.m12;
	theMatrix->m.m10 += temp;
	theMatrix->m.m11 += temp;
	
	temp = skew * theMatrix->m.m22;
	theMatrix->m.m20 += temp;
	theMatrix->m.m21 += temp;
	
	temp = skew * theMatrix->m.m32;
	theMatrix->m.m30 += temp;
	theMatrix->m.m31 += temp;
}  /* SkewMatrix4 */

void SkewRMatrix4(RMatrix4Ptr theMatrix, float skew)
{
	float temp;
	
	temp = skew * theMatrix->m.m02;
	theMatrix->m.m00 += temp;
	theMatrix->m.m01 += temp;
	
	temp = skew * theMatrix->m.m12;
	theMatrix->m.m10 += temp;
	theMatrix->m.m11 += temp;
	
	temp = skew * theMatrix->m.m22;
	theMatrix->m.m20 += temp;
	theMatrix->m.m21 += temp;
	
	temp = skew * theMatrix->m.m32;
	theMatrix->m.m30 += temp;
	theMatrix->m.m31 += temp;
}  /* SkewRMatrix4 */

void SkewMatrix3(Matrix3Ptr theMatrix, double skew)
{
	double temp;
	
	temp = skew * theMatrix->m.m02;
	theMatrix->m.m00 += temp;
	theMatrix->m.m01 += temp;
	
	temp = skew * theMatrix->m.m12;
	theMatrix->m.m10 += temp;
	theMatrix->m.m11 += temp;
	
	temp = skew * theMatrix->m.m22;
	theMatrix->m.m20 += temp;
	theMatrix->m.m21 += temp;
}  /* SkewMatrix3 */

void SkewRMatrix3(RMatrix3Ptr theMatrix, float skew)
{
	float temp;
	
	temp = skew * theMatrix->m.m02;
	theMatrix->m.m00 += temp;
	theMatrix->m.m01 += temp;
	
	temp = skew * theMatrix->m.m12;
	theMatrix->m.m10 += temp;
	theMatrix->m.m11 += temp;
	
	temp = skew * theMatrix->m.m22;
	theMatrix->m.m20 += temp;
	theMatrix->m.m21 += temp;
}  /* SkewRMatrix3 */

void ECSMatrix4(Matrix4Ptr theMatrix, double dx, double dy, double dz)
{
	if ((dx != 0.0) || (dy != 0.0) || (dz != 1.0)) {
		double z1 = dx;
		double z2 = dy;
		double z3 = dz;

		Normalize(&z1, &z2, &z3);
		if ((z1 != 0.0) || (z2 != 0.0) || (z3 != 1.0)) {
			Matrix4 e;
			double x1, y1, x2, y2, x3, y3;
			
			if ((FABS(z1) < 0.015625) && (FABS(z2) < 0.015625)) {
				x1 = z3;
				x2 = 0.0;
				x3 = -z1;	/* VectorProduct(0.0, 1.0, 0.0, z1, z2, z3, &x1, &x2, &x3); */
			} else {
				x1 = -z2;
				x2 = z1;
				x3 = 0.0;	/* VectorProduct(0.0, 0.0, 1.0, z1, z2, z3, &x1, &x2, &x3); */
			}
			Normalize(&x1, &x2, &x3);

			VectorProduct(z1, z2, z3, x1, x2, x3, &y1, &y2, &y3);
			Normalize(&y1, &y2, &y3);

			e.m.m00 = x1;
			e.m.m01 = x2;
			e.m.m02 = x3;
			e.m.m03 = 0.0;
			
			e.m.m10 = y1;
			e.m.m11 = y2;
			e.m.m12 = y3;
			e.m.m13 = 0.0;
			
			e.m.m20 = z1;
			e.m.m21 = z2;
			e.m.m22 = z3;
			e.m.m23 = 0.0;
			
			e.m.m30 = 0.0;
			e.m.m31 = 0.0;
			e.m.m32 = 0.0;
			e.m.m33 = 1.0;
			AppendMatrix4(theMatrix, &e);
		}
	}
}  /* ECSMatrix4 */

void ECSRMatrix4(RMatrix4Ptr theMatrix, float dx, float dy, float dz)
{
	if ((dx != 0.0) || (dy != 0.0) || (dz != 1.0)) {
		float z1 = dx;
		float z2 = dy;
		float z3 = dz;

		NormalizeR(&z1, &z2, &z3);
		if ((z1 != 0.0) || (z2 != 0.0) || (z3 != 1.0)) {
			RMatrix4 e;
			float x1, y1, x2, y2, x3, y3;
			
			if ((FABS(z1) < 0.015625) && (FABS(z2) < 0.015625)) {
				x1 = z3;
				x2 = 0.0f;
				x3 = -z1;	/* VectorProduct(0.0, 1.0, 0.0, z1, z2, z3, &x1, &x2, &x3); */
			} else {
				x1 = -z2;
				x2 = z1;
				x3 = 0.0f;	/* VectorProduct(0.0, 0.0, 1.0, z1, z2, z3, &x1, &x2, &x3); */
			}
			NormalizeR(&x1, &x2, &x3);

			VectorProductR(z1, z2, z3, x1, x2, x3, &y1, &y2, &y3);
			NormalizeR(&y1, &y2, &y3);

			e.m.m00 = x1;
			e.m.m01 = x2;
			e.m.m02 = x3;
			e.m.m03 = 0.0f;
			
			e.m.m10 = y1;
			e.m.m11 = y2;
			e.m.m12 = y3;
			e.m.m13 = 0.0f;
			
			e.m.m20 = z1;
			e.m.m21 = z2;
			e.m.m22 = z3;
			e.m.m23 = 0.0f;
			
			e.m.m30 = 0.0f;
			e.m.m31 = 0.0f;
			e.m.m32 = 0.0f;
			e.m.m33 = 1.0f;
			AppendRMatrix4(theMatrix, &e);
		}
	}
}  /* ECSRMatrix4 */

void ECSMatrix3(Matrix3Ptr theMatrix, double dx, double dy, double dz)
{
	if ((dx != 0.0) || (dy != 0.0) || (dz != 1.0)) {
		double z1 = dx;
		double z2 = dy;
		double z3 = dz;

		Normalize(&z1, &z2, &z3);
		if ((z1 != 0.0) || (z2 != 0.0) || (z3 != 1.0)) {
			Matrix3 e;
			double x1, y1, x2, y2, x3, y3;
			
			if ((FABS(z1) < 0.015625) && (FABS(z2) < 0.015625)) {
				x1 = z3;
				x2 = 0.0;
				x3 = -z1;	/* VectorProduct(0.0, 1.0, 0.0, z1, z2, z3, &x1, &x2, &x3); */
			} else {
				x1 = -z2;
				x2 = z1;
				x3 = 0.0;	/* VectorProduct(0.0, 0.0, 1.0, z1, z2, z3, &x1, &x2, &x3); */
			}
			Normalize(&x1, &x2, &x3);

			VectorProduct(z1, z2, z3, x1, x2, x3, &y1, &y2, &y3);
			Normalize(&y1, &y2, &y3);

			e.m.m00 = x1;
			e.m.m01 = x2;
			e.m.m02 = x3;
			
			e.m.m10 = y1;
			e.m.m11 = y2;
			e.m.m12 = y3;
			
			e.m.m20 = z1;
			e.m.m21 = z2;
			e.m.m22 = z3;

			AppendMatrix3(theMatrix, &e);
		}
	}
}  /* ECSMatrix3 */

void ECSRMatrix3(RMatrix3Ptr theMatrix, float dx, float dy, float dz)
{
	if ((dx != 0.0) || (dy != 0.0) || (dz != 1.0)) {
		float z1 = dx;
		float z2 = dy;
		float z3 = dz;

		NormalizeR(&z1, &z2, &z3);
		if ((z1 != 0.0) || (z2 != 0.0) || (z3 != 1.0)) {
			RMatrix3 e;
			float x1, y1, x2, y2, x3, y3;
			
			if ((FABS(z1) < 0.015625) && (FABS(z2) < 0.015625)) {
				x1 = z3;
				x2 = 0.0f;
				x3 = -z1;	/* VectorProduct(0.0, 1.0, 0.0, z1, z2, z3, &x1, &x2, &x3); */
			} else {
				x1 = -z2;
				x2 = z1;
				x3 = 0.0f;	/* VectorProduct(0.0, 0.0, 1.0, z1, z2, z3, &x1, &x2, &x3); */
			}
			NormalizeR(&x1, &x2, &x3);

			VectorProductR(z1, z2, z3, x1, x2, x3, &y1, &y2, &y3);
			NormalizeR(&y1, &y2, &y3);

			e.m.m00 = x1;
			e.m.m01 = x2;
			e.m.m02 = x3;
			
			e.m.m10 = y1;
			e.m.m11 = y2;
			e.m.m12 = y3;
			
			e.m.m20 = z1;
			e.m.m21 = z2;
			e.m.m22 = z3;

			AppendRMatrix3(theMatrix, &e);
		}
	}
}  /* ECSRMatrix3 */

void TransformVector(double* a, double* b, double* c, Matrix4Ptr theMatrix)
{
	double ap = *a, bp = *b, cp = *c;
	*a = ap * theMatrix->m.m00 + bp * theMatrix->m.m10 + cp * theMatrix->m.m20;
	*b = ap * theMatrix->m.m01 + bp * theMatrix->m.m11 + cp * theMatrix->m.m21;
	*c = ap * theMatrix->m.m02 + bp * theMatrix->m.m12 + cp * theMatrix->m.m22;
}  /* TransformVector */

void TransformNormal(double* a, double* b, double* c, Matrix4Ptr theMatrix)
{
	double ap = *a, bp = *b, cp = *c;
	double aNew = ap * theMatrix->m.m00 + bp * theMatrix->m.m10 + cp * theMatrix->m.m20;
	double bNew = ap * theMatrix->m.m01 + bp * theMatrix->m.m11 + cp * theMatrix->m.m21;
	double cNew = ap * theMatrix->m.m02 + bp * theMatrix->m.m12 + cp * theMatrix->m.m22;
	double distance = SQRT(SQR(aNew) + SQR(bNew) + SQR(cNew));

	if (distance == 0.0) {
		*a = 0.0;
		*b = 0.0;
		*c = 1.0;
	} else {
		distance = 1.0 / distance;
		*a = aNew * distance;
		*b = bNew * distance;
		*c = cNew * distance;
	}
}  /* TransformNormal */

void TransformPoint(double* x, double* y, double* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;

	*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
	*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
	*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
}  /* TransformPoint */

void TransformPointR(double* x, double* y, double* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;

	*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20;
	*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21;
	*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22;
}  /* TransformPointR */

void TransformPointH(double* x, double* y, double* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;
	double h = xp * theMatrix->m.m03 + yp * theMatrix->m.m13 + zp * theMatrix->m.m23 + theMatrix->m.m33;

	if (h == 1.0) {
		*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
		*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
		*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
	} else if (h == 0.0) {
		*x = 0.0;
		*y = 0.0;
		*z = 0.0;
	} else {
		h = 1.0 / h;
		*x = (xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30) * h;
		*y = (xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31) * h;
		*z = (xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32) * h;
	}
}  /* TransformPointH */

void TransformXPoint(double* x, double* y, double* z, XMatrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;

	*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
	*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
	*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
}  /* TransformXPoint */

void TransformRPoint(float* x, float* y, float* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;
	
	*x = (float)(xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30);
	*y = (float)(xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31);
	*z = (float)(xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32);
}  /* TransformRPoint */

void TransformRPointH(float* x, float* y, float* z, Matrix4Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;
	double h = xp * theMatrix->m.m03 + yp * theMatrix->m.m13 + zp * theMatrix->m.m23 + theMatrix->m.m33;

	if (h == 1.0) {
		*x = (float)(xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30);
		*y = (float)(xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31);
		*z = (float)(xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32);
	} else if (h == 0.0) {
		*x = 0.0f;
		*y = 0.0f;
		*z = 0.0f;
	} else {
		h = 1.0 / h;
		*x = (float)((xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30) * h);
		*y = (float)((xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31) * h);
		*z = (float)((xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32) * h);
	}
}  /* TransformRPointH */

void TransformCoordinate(CoordPtr theCoordinate, Matrix4Ptr theMatrix)
{
	double xp = theCoordinate->x, yp = theCoordinate->y, zp = theCoordinate->z;

	theCoordinate->x = (float)(xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30);
	theCoordinate->y = (float)(xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31);
	theCoordinate->z = (float)(xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32);
}  /* TransformCoordinate */

void TransformCoordinateH(CoordPtr theCoordinate, Matrix4Ptr theMatrix)
{
	double xp = theCoordinate->x, yp = theCoordinate->y, zp = theCoordinate->z;
	double h = xp * theMatrix->m.m03 + yp * theMatrix->m.m13 + zp * theMatrix->m.m23 + theMatrix->m.m33;

	if (h == 1.0) {
		theCoordinate->x = (float)(xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30);
		theCoordinate->y = (float)(xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31);
		theCoordinate->z = (float)(xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32);
	} else if (h == 0.0) {
		theCoordinate->x = 0.0f;
		theCoordinate->y = 0.0f;
		theCoordinate->z = 0.0f;
	} else {
		h = 1.0 / h;
		theCoordinate->x = (float)((xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30) * h);
		theCoordinate->y = (float)((xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31) * h);
		theCoordinate->z = (float)((xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32) * h);
	}
}  /* TransformCoordinateH */

void TransformDCoordinate(DCoordPtr theCoordinate, Matrix4Ptr theMatrix)
{
	double xp = theCoordinate->x, yp = theCoordinate->y, zp = theCoordinate->z;

	theCoordinate->x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
	theCoordinate->y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
	theCoordinate->z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
}  /* TransformDCoordinate */

void TransformDCoordinateH(DCoordPtr theCoordinate, Matrix4Ptr theMatrix)
{
	double xp = theCoordinate->x, yp = theCoordinate->y, zp = theCoordinate->z;
	double h = xp * theMatrix->m.m03 + yp * theMatrix->m.m13 + zp * theMatrix->m.m23 + theMatrix->m.m33;

	if (h == 1.0) {
		theCoordinate->x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30;
		theCoordinate->y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31;
		theCoordinate->z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32;
	} else if (h == 0.0) {
		theCoordinate->x = 0.0;
		theCoordinate->y = 0.0;
		theCoordinate->z = 0.0;
	} else {
		h = 1.0 / h;
		theCoordinate->x = (xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20 + theMatrix->m.m30) * h;
		theCoordinate->y = (xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21 + theMatrix->m.m31) * h;
		theCoordinate->z = (xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22 + theMatrix->m.m32) * h;
	}
}  /* TransformDCoordinateH */

void TransformPoint3(double* x, double* y, double* z, Matrix3Ptr theMatrix)
{
	double xp = *x, yp = *y, zp = *z;
	
	*x = xp * theMatrix->m.m00 + yp * theMatrix->m.m10 + zp * theMatrix->m.m20;
	*y = xp * theMatrix->m.m01 + yp * theMatrix->m.m11 + zp * theMatrix->m.m21;
	*z = xp * theMatrix->m.m02 + yp * theMatrix->m.m12 + zp * theMatrix->m.m22;
}  /* TransformPoint */

void CopyMatrix4To3(Matrix3Ptr theMatrix, Matrix4Ptr b)
{
	theMatrix->v.mv0[0] = b->v.mv0[0];
	theMatrix->v.mv0[1] = b->v.mv0[1];
	theMatrix->v.mv0[2] = b->v.mv0[2];
	theMatrix->v.mv1[0] = b->v.mv1[0];
	theMatrix->v.mv1[1] = b->v.mv1[1];
	theMatrix->v.mv1[2] = b->v.mv1[2];
	theMatrix->v.mv2[0] = b->v.mv2[0];
	theMatrix->v.mv2[1] = b->v.mv2[1];
	theMatrix->v.mv2[2] = b->v.mv2[2];
}  /* CopyMatrix4To3 */

void CopyRMatrix4To3(RMatrix3Ptr theMatrix, RMatrix4Ptr b)
{
	theMatrix->v.mv0[0] = b->v.mv0[0];
	theMatrix->v.mv0[1] = b->v.mv0[1];
	theMatrix->v.mv0[2] = b->v.mv0[2];
	theMatrix->v.mv1[0] = b->v.mv1[0];
	theMatrix->v.mv1[1] = b->v.mv1[1];
	theMatrix->v.mv1[2] = b->v.mv1[2];
	theMatrix->v.mv2[0] = b->v.mv2[0];
	theMatrix->v.mv2[1] = b->v.mv2[1];
	theMatrix->v.mv2[2] = b->v.mv2[2];
}  /* CopyRMatrix4To3 */

void CopyMatrix3To4(Matrix4Ptr theMatrix, Matrix3Ptr b)
{
	theMatrix->v.mv0[0] = b->v.mv0[0];
	theMatrix->v.mv0[1] = b->v.mv0[1];
	theMatrix->v.mv0[2] = b->v.mv0[2];
	theMatrix->v.mv1[0] = b->v.mv1[0];
	theMatrix->v.mv1[1] = b->v.mv1[1];
	theMatrix->v.mv1[2] = b->v.mv1[2];
	theMatrix->v.mv2[0] = b->v.mv2[0];
	theMatrix->v.mv2[1] = b->v.mv2[1];
	theMatrix->v.mv2[2] = b->v.mv2[2];
	theMatrix->m.m03 = 0.0;
	theMatrix->m.m13 = 0.0;
	theMatrix->m.m23 = 0.0;
	theMatrix->m.m30 = 0.0;
	theMatrix->m.m31 = 0.0;
	theMatrix->m.m32 = 0.0;
	theMatrix->m.m33 = 1.0;
}  /* CopyMatrix3To4 */

void CopyRMatrix3To4(RMatrix4Ptr theMatrix, RMatrix3Ptr b)
{
	theMatrix->v.mv0[0] = b->v.mv0[0];
	theMatrix->v.mv0[1] = b->v.mv0[1];
	theMatrix->v.mv0[2] = b->v.mv0[2];
	theMatrix->v.mv1[0] = b->v.mv1[0];
	theMatrix->v.mv1[1] = b->v.mv1[1];
	theMatrix->v.mv1[2] = b->v.mv1[2];
	theMatrix->v.mv2[0] = b->v.mv2[0];
	theMatrix->v.mv2[1] = b->v.mv2[1];
	theMatrix->v.mv2[2] = b->v.mv2[2];
	theMatrix->m.m03 = 0.0f;
	theMatrix->m.m13 = 0.0f;
	theMatrix->m.m23 = 0.0f;
	theMatrix->m.m30 = 0.0f;
	theMatrix->m.m31 = 0.0f;
	theMatrix->m.m32 = 0.0f;
	theMatrix->m.m33 = 1.0f;
}  /* CopyMatrix3To4 */

/* Quaternion Routines --- Q=A*i+B*j+C*k+R=q0*i+q1*j+q2*k+q3 */

void AddQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	r->x = p->x + q->x;
	r->y = p->y + q->y;
	r->z = p->z + q->z;
	r->w = p->w + q->w;
}  /* AddQuaternion */

void SubtractQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	r->x = p->x - q->x;
	r->y = p->y - q->y;
	r->z = p->z - q->z;
	r->w = p->w - q->w;
}  /* SubtractQuaternion */

void MultiplyQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	double x = p->w * q->x + p->x * q->w + p->y * q->z - p->z * q->y;
	double y = p->w * q->y + p->y * q->w + p->z * q->x - p->x * q->z;
	double z = p->w * q->z + p->z * q->w + p->x * q->y - p->y * q->x;
	double w = p->w * q->w - p->x * q->x - p->y * q->y - p->z * q->z;

	r->x = x;
	r->y = y;
	r->z = z;
	r->w = w;
}  /* MultiplyQuaternion */

void DivideQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	double a = -1.0 / (SQR(q->w) - SQR(q->x) - SQR(q->y) - SQR(q->z));

	MultiplyQuaternion(p, q, r);
	r->x *= a;
	r->y *= a;
	r->z *= a;
	r->w *= a;
}  /* DivideQuaternion */

void ConjugateQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	r->x = -q->x;
	r->y = -q->y;
	r->z = -q->z;
	r->w = q->w;
}  /* ConjugateQuaternion */

void InvertQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double normInv = 1.0 / (SQR(q->x) + SQR(q->y) + SQR(q->z) + SQR(q->w));

	r->x = -q->x * normInv;
	r->y = -q->y * normInv;
	r->z = -q->z * normInv;
	r->w = q->w * normInv;
}  /* InvertQuaternion */

void ExponentiateQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double scale = SQRT(SQR(q->x) + SQR(q->y) + SQR(q->z));

	r->w = COS(scale);
	if (scale > matrix_epsilon)
		scale = SIN(scale) / scale;
	else
		scale = 1.0;
	r->x = scale * q->x;
	r->y = scale * q->y;
	r->z = scale * q->z;
}  /* ExponentiateQuaternion */

void LogarithmQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double scale = SQRT(SQR(q->x) + SQR(q->y) + SQR(q->z));

	if (scale > matrix_epsilon)
		scale = ATAN2(scale, q->w) / scale;
	r->x = scale * q->x;
	r->y = scale * q->y;
	r->z = scale * q->z;
	r->w = 0.0;
}  /* LogarithmQuaternion */

void SquareRouteQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double len = SQRT(SQR(q->w) + SQR(q->x) + SQR(q->y));
	double l = 1.0 / len;
	double w = q->w * l;
	double x = q->x * l;
	double y = q->y * l;
	double m = 1.0 / SQRT(SQR(w) + SQR(x));
	double a = SQRT((l + x) * 0.5);
	double b = SQRT((l - y) * 0.5);

	len = SQRT(len);
	r->w = len * b * w * m;
	r->x = len * b * x * m;
	r->y = len * a;
	r->z = q->z;
}  /* SquareRootQuaternion */

void SquareQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double a = q->w + q->w;

	r->w = SQR(q->w) - SQR(q->x) - SQR(q->z) - SQR(q->w);
	r->x = a * q->x;
	r->y = a * q->y;
	r->z = a * q->z;
}  /* SquareQuaternion */

void ScaleQuaternion(double c, QuaternionPtr q, QuaternionPtr r)
{
	r->x = q->x * c;
	r->y = q->y * c;
	r->z = q->z * c;
	r->w = q->w * c;
}  /* ScaleQuaternion */

double DotQuaternion(QuaternionPtr p, QuaternionPtr q)
{
	return p->x * q->x + p->y * q->y + p->z * q->z + p->w * q->w;
}  /* DotQuaternion */

void DoubleQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	Quaternion s;
	double d = DotQuaternion(p, q);
	
	ScaleQuaternion(d + d, q, &s);
	SubtractQuaternion(&s, p, r);
}  /* DoubleQuaternion */

void BisectQuaternion(QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	Quaternion s;
	double size;

	AddQuaternion(p, q, &s);
	size = SQR(s.x) + SQR(s.y) + SQR(s.z) + SQR(s.w);
	if (size > matrix_epsilon)
		ScaleQuaternion(1.0 / SQRT(size), &s, r);
	else {
		r->x = -p->y;
		r->y = p->x;
		r->z = -p->w;
		r->w = p->z;
	}
}  /* BisectQuaternion */

void NormalizeQuaternion(QuaternionPtr q, QuaternionPtr r)
{
	double size = SQR(q->x) + SQR(q->y) + SQR(q->z) + SQR(q->w);

	if (size > matrix_epsilon)
		ScaleQuaternion(1.0 / SQRT(size), q, r);
	else {
		r->x = 0.0;
		r->y = 0.0;
		r->z = 0.0;
		r->w = 1.0;
	}
}  /* NormalizeQuaternion */

void LinearInterpolateQuaternion(double t, QuaternionPtr p, QuaternionPtr q, QuaternionPtr r)
{
	if (t <= 0.0)  /* As t goes from 0 to 1, r goes from p to q */
		*r = *p;
	else if (t >= 1.0)
		*r = *q;
	else {
		double cosom = DotQuaternion(p, q);

		if (1.0 + cosom > matrix_epsilon)
			if (1.0 - cosom > matrix_epsilon) {
				double omega = ACOS(cosom);
				double sinom = SIN(omega);
				double sclp = SIN((1.0 - t) * omega) / sinom;
				double sclq = SIN(t * omega) / sinom;

				r->x = sclp * p->x + sclq * q->x;
				r->y = sclp * p->y + sclq * q->y;
				r->z = sclp * p->z + sclq * q->z;
				r->w = sclp * p->w + sclq * q->w;
			} else {	/* Use linear interpolation if angle is small */
				double sclp = 1.0 - t;
				double sclq = t;

				r->x = sclp * p->x + sclq * q->x;
				r->y = sclp * p->y + sclq * q->y;
				r->z = sclp * p->z + sclq * q->z;
				r->w = sclp * p->w + sclq * q->w;
			} else {	/* Interpolate with mid orientation if the angle is nearly 180 degrees */
				double sclp = SIN((0.5 - t) * PI);
				double sclq = SIN(t * PI);
				double x = sclp * p->x - sclq * p->y;
				double y = sclp * p->y + sclq * p->x;
				double z = sclp * p->z - sclq * p->w;
				double w = sclp * p->w + sclq * p->z;

				r->x = x;
				r->y = y;
				r->z = z;
				r->w = w;
			}
	}
}  /* LinearInterpolateQuaternion */

void BezierInteroplateQuaternion(double t, QuaternionPtr q0, QuaternionPtr q1, QuaternionPtr q2, QuaternionPtr q3, QuaternionPtr r)
{
	Quaternion a, b, m, n, o;
	
	/* Insure that the quaternion interpolation takes the shortest path */
	
	if (DotQuaternion(q1, q2) < 0.0)
		ConjugateQuaternion(q2, &m);
	else
		m = *q2;

	/* If q0 does not exist, then find it with DoubleQuaternion(q2,q1,q0) else */

	if (DotQuaternion(q1, q0) < 0.0)
		ConjugateQuaternion(q0, &n);
	else
		n = *q0;

	/* If q3 does not exist, then find it with DoubleQuaternion(q1,q2,q3) else */

	if (DotQuaternion(q2, q3) < 0.0)
		ConjugateQuaternion(q3, &o);
	else
		o = *q3;
	
	/* Alternatively, a could be set to q1 when q0 does not exist and b could be set */
	/* to q2 when q3 does not exist */
	
	/* Find the forwards tangent at q1 */
	
	DoubleQuaternion(&n, q1, &a);
	BisectQuaternion(&a, &m, &a);
	
	/* Find the backwards tangent at q2 */
	
	DoubleQuaternion(&o, &m, &b);
	BisectQuaternion(&b, q1, &b);
	
	/* Recursively interpolate the Bezier curve */
	
	LinearInterpolateQuaternion(t, q1, &a, &m);
	LinearInterpolateQuaternion(t, &a, &b, &n);
	LinearInterpolateQuaternion(t, &b, q2, &o);
	LinearInterpolateQuaternion(t, &m, &n, &m);
	LinearInterpolateQuaternion(t, &n, &o, &n);
	LinearInterpolateQuaternion(t, &m, &n, r);
	
	/* Cord length is Sqrt(Sqr(q.px-p.px)+Sqr(q.py-p.py)+Sqr(q.pz-p.pz)+Sqr(q.pw-p.pw)) */
	/* Arc length is Dot(p,q) */
	/* Find length of the curve by calculating coord lengths and sub-dividing until the difference in total length */
	/* becomes small enough */
}  /* BezierInterpolateQuaternion */

void Quaternion2Matrix4(QuaternionPtr theQuaternion, Matrix4Ptr theMatrix)
{
	double doubleX = theQuaternion->x + theQuaternion->x;
	double doubleY = theQuaternion->y + theQuaternion->y;
	double doubleZ = theQuaternion->z + theQuaternion->z;
	double angleX = doubleX * theQuaternion->w;
	double angleY = doubleY * theQuaternion->w;
	double angleZ = doubleZ * theQuaternion->w;
	double xMultX = doubleX * theQuaternion->x;
	double xMultY = doubleX * theQuaternion->y;
	double xMultZ = doubleX * theQuaternion->z;
	double yMultY = doubleY * theQuaternion->y;
	double yMultZ = doubleY * theQuaternion->z;
	double zMultZ = doubleZ * theQuaternion->z;
	
	theMatrix->m.m00 = 1.0 - yMultY - zMultZ;
	theMatrix->m.m01 = xMultY + angleZ;
	theMatrix->m.m02 = xMultZ - angleY;
	theMatrix->m.m03 = 0.0;
	theMatrix->m.m10 = xMultY - angleZ;
	theMatrix->m.m11 = 1.0 - xMultX - zMultZ;
	theMatrix->m.m12 = yMultZ + angleX;
	theMatrix->m.m13 = 0.0;
	theMatrix->m.m20 = xMultZ + angleY;
	theMatrix->m.m21 = yMultZ - angleX;
	theMatrix->m.m22 = 1.0 - xMultX - yMultY;
	theMatrix->m.m33 = 0.0;
	theMatrix->m.m30 = 0.0;
	theMatrix->m.m31 = 0.0;
	theMatrix->m.m32 = 0.0;
	theMatrix->m.m33 = 1.0;
}  /* Quaternion2Matrix4 */

void Matrix42Quaternion(Matrix4Ptr theMatrix, QuaternionPtr theQuaternion)
{
	theQuaternion->w = 0.25 * (1.0 + theMatrix->m.m00 + theMatrix->m.m11 + theMatrix->m.m22);
	if (theQuaternion->w > matrix_nearZero) {
		theQuaternion->w = SQRT(theQuaternion->w);
		theQuaternion->x = (theMatrix->m.m12 - theMatrix->m.m21) / (4.0 * theQuaternion->w);
		theQuaternion->y = (theMatrix->m.m20 - theMatrix->m.m02) / (4.0 * theQuaternion->w);
		theQuaternion->z = (theMatrix->m.m01 - theMatrix->m.m10) / (4.0 * theQuaternion->w);
	} else {
		theQuaternion->w = 0.0;
		theQuaternion->x = -0.5 * (theMatrix->m.m11 + theMatrix->m.m22);
		if (theQuaternion->x > matrix_nearZero) {
			double n = theQuaternion->x;
			
			theQuaternion->x = SQRT(n);
			n = 1.0 / (n + n);
			theQuaternion->y = theMatrix->m.m01 * n;
			theQuaternion->z = theMatrix->m.m02 * n;
		} else {
			theQuaternion->x = 0.0;
			theQuaternion->y = 0.5 * (1.0 - theMatrix->m.m22);
			if (theQuaternion->y > matrix_nearZero) {
				double n = theQuaternion->y;
				
				theQuaternion->y = SQRT(theQuaternion->y);
				theQuaternion->z = theMatrix->m.m12 / (n + n);
			} else {
				theQuaternion->y = 0.0;
				theQuaternion->z = 1.0;
			}
		}
	}
}  /* Matrix42Quaternion */

void QuaternionToMatrix4(QuaternionPtr theQuaternion, Matrix4Ptr theMatrix)
{
	double s = SQR(theQuaternion->x) + SQR(theQuaternion->y) + SQR(theQuaternion->z) + SQR(theQuaternion->w);

	if (s > 0.0) {
		double xs, ys, zs, wx, wy, wz, xx, xy, xz, yy, yz, zz;

		s = 2.0 / s;
		xs = theQuaternion->x * s;
		ys = theQuaternion->y * s;
		zs = theQuaternion->z * s;
		wx = theQuaternion->w * xs;
		wy = theQuaternion->w * ys;
		wz = theQuaternion->w * zs;
		xx = theQuaternion->x * xs;
		xy = theQuaternion->x * ys;
		xz = theQuaternion->x * zs;
		yy = theQuaternion->y * ys;
		yz = theQuaternion->y * zs;
		zz = theQuaternion->z * zs;
		
		theMatrix->m.m00 = 1.0 - (yy + zz);
		theMatrix->m.m01 = xy - wz;
		theMatrix->m.m02 = xz + wy;
		theMatrix->m.m03 = 0.0;
		theMatrix->m.m10 = xy + wz;
		theMatrix->m.m11 = 1.0 - (xx + zz);
		theMatrix->m.m12 = yz - wx;
		theMatrix->m.m13 = 0.0;
		theMatrix->m.m20 = xz - wy;
		theMatrix->m.m21 = yz + wx;
		theMatrix->m.m22 = 1.0 - (xx + yy);
		theMatrix->m.m23 = 0.0;
		theMatrix->m.m30 = 0.0;
		theMatrix->m.m31 = 0.0;
		theMatrix->m.m32 = 0.0;
		theMatrix->m.m33 = 1.0;
	} else
		IdentityMatrix4(theMatrix);
}  /* QuaternionToMatrix */

void Matrix4ToQuaternion(Matrix4Ptr theMatrix, QuaternionPtr theQuaternion)
{
	double s;
	double tr = theMatrix->m.m00 + theMatrix->m.m11 + theMatrix->m.m22;

	if (tr > 0.0) {
		s = SQRT(tr + 1.0);
		theQuaternion->w = s * 0.5;
		s = 0.5 / s;
		theQuaternion->x = (theMatrix->m.m21 - theMatrix->m.m12) * s;
		theQuaternion->y = (theMatrix->m.m02 - theMatrix->m.m21) * s;
		theQuaternion->z = (theMatrix->m.m10 - theMatrix->m.m01) * s;
	} else if (theMatrix->m.m11 > theMatrix->m.m00)
		if (theMatrix->m.m22 > theMatrix->m.m11) {
			s = SQRT((theMatrix->matrix[1][1] - (theMatrix->matrix[2][1] + theMatrix->matrix[0][0])) + 1.0);
			theQuaternion->y = s * 0.5;
			theQuaternion->z = (theMatrix->matrix[2][1] - theMatrix->matrix[1][2]) * s;
			theQuaternion->x = (theMatrix->matrix[0][1] - theMatrix->matrix[1][0]) * s;
			theQuaternion->w = (theMatrix->matrix[0][2] - theMatrix->matrix[2][0]) * s;
		} else {
			s = SQRT((theMatrix->matrix[2][2] - (theMatrix->matrix[0][2] + theMatrix->matrix[1][1])) + 1.0);
			theQuaternion->z = s * 0.5;
			theQuaternion->x = (theMatrix->matrix[0][2] - theMatrix->matrix[2][0]) * s;
			theQuaternion->y = (theMatrix->matrix[1][2] - theMatrix->matrix[2][1]) * s;
			theQuaternion->w = (theMatrix->matrix[1][0] - theMatrix->matrix[0][1]) * s;
		}
	else if (theMatrix->m.m22 > theMatrix->m.m00) {
		s = SQRT((theMatrix->matrix[2][2] - (theMatrix->matrix[0][2] + theMatrix->matrix[1][1])) + 1.0);
		theQuaternion->z = s * 0.5;
		theQuaternion->x = (theMatrix->matrix[0][2] - theMatrix->matrix[2][0]) * s;
		theQuaternion->y = (theMatrix->matrix[1][2] - theMatrix->matrix[2][1]) * s;
		theQuaternion->w = (theMatrix->matrix[1][0] - theMatrix->matrix[0][1]) * s;
	} else {
		s = SQRT((theMatrix->matrix[0][0] - (theMatrix->matrix[1][0] + theMatrix->matrix[2][2])) + 1.0);
		theQuaternion->x = s * 0.5;
		theQuaternion->y = (theMatrix->matrix[1][0] - theMatrix->matrix[0][1]) * s;
		theQuaternion->z = (theMatrix->matrix[2][0] - theMatrix->matrix[0][2]) * s;
		theQuaternion->w = (theMatrix->matrix[2][1] - theMatrix->matrix[1][2]) * s;
	}
}  /* Matrix4ToQuaternion */

void Matrix4ToAngles(Matrix4Ptr theMatrix, double *x, double *y, double *z)
{
	double sinY = -theMatrix->m.m02;
	double cosY = SQRT(1.0 - SQR(sinY));

	if (cosY > matrix_epsilon) {
		double temp = 1.0 / cosY;

		*x = ATAN2(theMatrix->m.m12 * temp, theMatrix->m.m22 * temp);
		*z = ATAN2(theMatrix->m.m01 * temp, theMatrix->m.m00 * temp);
	} else {
		*x = ATAN2(-theMatrix->m.m21, theMatrix->m.m11);
		*z = ATAN2(0.0, 1.0);
	}
	*y = ATAN2(sinY, cosY);
}  /* Matrix4ToAngles */

void RMatrix4ToAngles(RMatrix4Ptr theMatrix, double *x, double *y, double *z)
{
	double sinY = -theMatrix->m.m02;
	double cosY = SQRT(1.0 - SQR(sinY));

	if (cosY > matrix_epsilon) {
		double temp = 1.0 / cosY;

		*x = ATAN2(theMatrix->m.m12 * temp, theMatrix->m.m22 * temp);
		*z = ATAN2(theMatrix->m.m01 * temp, theMatrix->m.m00 * temp);
	} else {
		*x = ATAN2(-theMatrix->m.m21, theMatrix->m.m11);
		*z = ATAN2(0.0, 1.0);
	}
	*y = ATAN2(sinY, cosY);
}  /* RMatrix4ToAngles */
